/**
 * @file callbacks.c misc UI stuff
 *
 * Most of the GUI code is distributed over the ui_*.c
 * files but what didn't fit somewhere else stayed here.
 * 
 * Copyright (C) 2003, 2004 Lars Lindner <lars.lindner@gmx.net>
 * Copyright (C) 2004 Christophe Barbe <christophe.barbe@ufies.org>    
 * Copyright (C) 2004 Nathan J. Conrad <t98502@users.sourceforge.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <gdk/gdkkeysyms.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <osso-ic.h>
#include <hildon-widgets/gtk-infoprint.h>
#include <hildon-widgets/hildon-caption.h>
#include <hildon-widgets/hildon-time-editor.h>

#include <hildon-widgets/hildon-file-details-dialog.h>
#include <hildon-fm/hildon-widgets/hildon-file-chooser-dialog.h>

#include <osso-log.h>
#include <string.h>
#include <glib.h>

#include "debug.h"
#include "interface.h"
#include "support.h"
#include "folder.h"
#include "feed.h"
#include "item.h"
#include "conf.h"
#include "export.h"
#include "htmlview.h"
#include "common.h"
#include "callbacks.h"
#include "ui_mainwindow.h"
#include "ui_folder.h"
#include "ui_feedlist.h"
#include "ui_itemlist.h"
#include "ui_queue.h"
#include "ui_notification.h"
#include "ui_mainwindow.h"
#include "state_save.h"
#include "update.h"
#include "dbus.h"
#include "main.h"
#include "gtkhtml/gtkhtml_plugin.h"
#include <gtkhtml/gtkhtml.h>

extern AppData *app_data;

extern time_t rss_updated;
extern long rss_unread;

extern GHashTable *feedHandler;

extern GtkWidget *newsbox;
extern GtkWidget *feedlist_container;
extern feedPtr searchFeed;
extern GtkWidget * gtkhtml_htmlview;

gchar *prev_find_text = NULL;
/** prev_text_has_a_match
    will be set to TRUE if prev_find_text appears in the page
    will be reset to FALSE if pre_find_text is also reset
*/
gboolean prev_text_has_a_match; 
/* all used icons */
GdkPixbuf *icons[MAX_ICONS];

/* icon names */
static gchar *iconNames[] = { "qgn_list_rss_feed",    /* default feed favicon */
    "qgn_list_rss_feed_invalid", /* invalid feed*/
    "qgn_list_gene_fldr_cls",    /* Folder closed */
    "qgn_list_gene_fldr_opn",    /* Folder opened */
    "qgn_list_gene_fldr_clp",  /*  closed folder icon - plus sign */
    "qgn_list_gene_fldr_exp",   /*  expanded folder icon - minus sign */
    NULL
};

/*------------------------------------------------------------------------------*/
/* generic GUI functions                            */
/*------------------------------------------------------------------------------*/

/************************************************************************/
/*                        PRIVATE FUNCTIONS                             */
/************************************************************************/

/** Filters the tree model so that only folders are shown
  *
  * @param model the tree model to filter
  * @param iter a tree iter
  * @param data not used
  * @return TRUE if the node pointed at by the iterator is a folder, FALSE otherwise
  */
static gboolean filter_show_folders_only(GtkTreeModel * model,
                     GtkTreeIter * iter, gpointer data)
{

    nodePtr ptr = NULL;

    if (model == NULL) {
    ULOG_WARN("filter_show_folders_only: WARNING: model is NULL.");
    return TRUE;
    }

    gtk_tree_model_get(model, iter, FS_PTR, &ptr, -1);
    if (ptr == NULL) {
    return FALSE;
    }

    if (ptr->type == FST_FOLDER) {
    return TRUE;
    } else {
    return FALSE;
    }

}

/* free the icons[] array
*/
void free_icons()
{
    gint i;
    ULOG_DEBUG("freeing the icons array");
    for (i = 0; i < MAX_ICONS ; i++)
        if (icons[i] != NULL)
            {
                g_object_unref(icons[i]);
                icons[i] = NULL;
            }                

}
void ui_init(gboolean startIconified, AppUIData * app_ui_data)
{
    gchar *path;
    //int i = 0;

    //g_assert(app_ui_data != NULL);
    //g_assert(app_ui_data->main_view != NULL);
    
    mainwindow = GTK_WIDGET(app_ui_data->main_view);
    create_mainwindow(app_ui_data);

    /* order important !!! */
    ui_feedlist_init(feedlist);
    //getItemStore();
    ULOG_DEBUG("ui_init: Finished ui_feedlist_init");
    ui_folder_set_expansion(ui_feedlist_get_root_folder(), TRUE);
    //TODO: REMOVE THIS
    ui_mainwindow_set_mode(itemlist_mode);
    /*
    for (i = 0; i < MAX_ICONS -2 ; i++)
        icons[i] = create_pixbuf(iconNames[i]);
    */

    /* TODO: will Those icons will be unref'ed by the gtktree view ?*/
    /* create rss feed icons */
    icons[0] = create_pixbuf(iconNames[0]);
    /* create rss invalid feed icons*/
    icons[1] = create_pixbuf(iconNames[1]);
    /* create rss folder closed icon . This is sometimes used in dialog lable*/
    icons[2] = create_pixbuf(iconNames[2]);
    
    /* create composite folder closed with emblem */
    icons[3] = create_composite_pixbuf(iconNames[2], iconNames[4]);
    /* create composite folder opened with emblem */
    icons[4] = create_composite_pixbuf(iconNames[3], iconNames[5]);

    app_ui_data->search_mode = SFM_NORMAL;
    app_ui_data->feed_displayed = FALSE;
        
    /* set up icons that are build from stock */
    ui_dnd_setup_URL_receiver(mainwindow);    /* setup URL dropping support */
    conf_load_subscriptions();
    /* Set the interval for automatic updates */
    if(getBooleanConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES))
        set_update_timer(FALSE);
    ui_mainwindow_finish();    /* Display the initial status info */
    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_data->app_ui_data->unsubscribe_button),
                 FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_data->app_ui_data->
                  feed_properties_button), FALSE);

    if(getBooleanConfValue(DISABLE_TOOLBAR))
    {    
        GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_normal)->active = FALSE;
    gtk_widget_hide(app_ui_data->main_toolbar);
    }
    else
        GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_normal)->active = TRUE;
                  
    if(getBooleanConfValue(DISABLE_TOOLBAR_FULLSCREEN))
        GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_full)->active = FALSE;
    else
        GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_full)->active = TRUE;    

    if(getBooleanConfValue(ENABLE_FIND_ON_PAGE))
    {
        GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_find_on_page)->active = TRUE;
        gtk_widget_show_all(GTK_WIDGET(app_ui_data->find_toolbar));
    }
    else
    {
        GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_find_on_page)->active = FALSE;
        gtk_widget_hide_all(GTK_WIDGET(app_ui_data->find_toolbar));
    }
 
    if (checkConfValueExist(RSS_SETTINGS_REUSE_BROWSER_WINDOW) == FALSE)
        setBooleanConfValue(RSS_SETTINGS_REUSE_BROWSER_WINDOW, TRUE);

    if (checkConfValueExist(RSS_SETTINGS_NEWEST_POSTS_FIRST)== FALSE)
        setBooleanConfValue(RSS_SETTINGS_NEWEST_POSTS_FIRST, TRUE);
           
    if (checkConfValueExist(RSS_SETTINGS_AUTOMATIC_UPDATES_INTERVAL)== FALSE)
        setNumericConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES_INTERVAL, DEFAULT_INTERVAL);
    
    path = g_strdup_printf("%s" G_DIR_SEPARATOR_S ".osso_rss_feed_reader/cache/images",
	                        g_get_home_dir());
    app_ui_data->img_cache = file_cache_new(path, 2*1024*1024);

    if (startIconified)
        gtk_widget_hide(GTK_WIDGET(app_ui_data->app));
    else
        gtk_widget_show(GTK_WIDGET(app_ui_data->app));
}

/*------------------------------------------------------------------------------*/
/* simple callbacks which don't belong to item or feed list             */
/*------------------------------------------------------------------------------*/

/**tvh: Icon render for the Folder tree view in the dialog mode
 * (manage folder dialog and change folder dialog )
 * Must use the icon, just for the folder icons being changed
 * as the dialogs use a filtered, but different model, and treeviews

*/
void
   dialog_folder_icon_render_func (GtkTreeViewColumn *col,
                           GtkCellRenderer   *renderer,
                           GtkTreeModel      *model,
                           GtkTreeIter       *iter,
                           gpointer           user_data)
{
    folderPtr ptr = NULL;
    GtkTreeIter child;
    gtk_tree_model_get(model, iter,
               FS_PTR, &ptr, -1);
    gboolean has_sub_folder = gtk_tree_model_iter_children(model, &child, iter);

    if (ptr->type == FST_FOLDER )
    {
        if (has_sub_folder == FALSE)
             g_object_set(renderer, "pixbuf", icons[ICON_FOLDER_CLOSED], NULL);
        else
             g_object_set(renderer, "pixbuf", ui_folder_select_icon(ptr->dialog_mode_expanded), NULL);
    }
}


/** tvh: Populate the folders for a filtered tree model used in dialog tree views
 * Model will be fitered. And the folder_view will connect to its own
 * collapse, expand callbacks
*/
void populate_folders(GtkWidget * dialog)
{
    GtkWidget *folder_view = lookup_widget(dialog, "folder_view");

    GtkTreeModel *model = NULL;
    GtkCellRenderer *textRenderer = NULL;
    GtkCellRenderer *iconRenderer = NULL;
    GtkTreeViewColumn *column = NULL;
    GtkTreeModel *filter = NULL;
    folderPtr rootfolder ;
    ULOG_DEBUG("..........populate_folders: starting...");

    filter = (GtkTreeModel *) gtk_tree_model_filter_new(GTK_TREE_MODEL(feedstore), NULL);
    
    gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(filter),
                       filter_show_folders_only,
                       NULL, NULL);
    
    model = GTK_TREE_MODEL(filter);

/*    model = GTK_TREE_MODEL(feedstore);*/
    gtk_tree_view_set_model(GTK_TREE_VIEW(folder_view), GTK_TREE_MODEL(model) );

    /* Show all the folders. */
    gtk_tree_view_collapse_all(GTK_TREE_VIEW(folder_view));
    /* Collapse the root folder */
    rootfolder = (folderPtr) ui_feedlist_get_root_folder();
    rootfolder->dialog_mode_expanded = FALSE;

    /* folder_view should have its own handlers */
    g_signal_connect(G_OBJECT(folder_view), "row-expanded",
             G_CALLBACK(ui_folder_mfdialog_collapse_expand_cb), NULL);
    g_signal_connect(G_OBJECT(folder_view), "row-collapsed",
             G_CALLBACK(ui_folder_mfdialog_collapse_expand_cb), NULL);
    
    /* we only render the state and title */
    iconRenderer = gtk_cell_renderer_pixbuf_new();
    textRenderer = gtk_cell_renderer_text_new();

    g_object_set(G_OBJECT(textRenderer), "ellipsize", PANGO_ELLIPSIZE_MIDDLE, NULL);

    gtk_cell_renderer_set_fixed_size(textRenderer, CELL_RENDERER_WIDTH, -1);

    column = gtk_tree_view_column_new();

    gtk_tree_view_column_pack_start(column, iconRenderer, FALSE);
    gtk_tree_view_column_pack_start(column, textRenderer, TRUE);

    /* Must use the func to render correctly as we don't want to modify the
    original model just because of expanding/collapsing the folders in dialogs
    */
    gtk_tree_view_column_set_cell_data_func(column, iconRenderer, dialog_folder_icon_render_func, NULL, NULL);
    /*
    gtk_tree_view_column_add_attribute(column, iconRenderer, "pixbuf",
                       FS_ICON);
    */

    gtk_tree_view_column_add_attribute(column, textRenderer, "markup",
                       FS_LABEL);

    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(GTK_TREE_VIEW(folder_view), column);
}

/** Update the progress bar if needed
 * This is called in two cases , for now
    1. Refreshing
    2. Receiving images
    3. TODO: searching
*/
void update_progress_bar(gint mode)
{
    
    ULOG_DEBUG("zzzzzzzzzzzzzUPDATE_PROGRESS_BARzzzzzzzzzzzzzzzzzzzzzz");
    switch (mode){
    // TODO: This actually includes PROGRESSBAR_RECEIVING_MODE TOO
    case PROGRESSBAR_REFRESH_MODE: 
        if (app_data->app_ui_data->feed_directory_loading)
            return;
              
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR
                      (app_data->app_ui_data->
                       progress_bar),
                      (float) get_counter(RSS_CURRENT)/(float) get_counter(RSS_MAX));
        break;
    case PROGRESSBAR_IMAGE_MODE:
    /*  It is assurred before calling this mode that 
        receiving image is the current progressbar mode
    */
    {
        gchar * text = NULL;
        /*
        text = g_strdup_printf(_("rss_pb_toolbar_receiving"), 
                app_data->app_ui_data->images_current_download,
                      app_data->app_ui_data->images_max_download);
        */
        text = g_strdup_printf(_("rss_pb_toolbar_receiving"), 
                        get_counter(IMG_CURRENT), get_counter(IMG_MAX));
        ULOG_DEBUG("update_progress_bar: img mode = %s", text);
        
        gtk_progress_bar_set_text(GTK_PROGRESS_BAR
                    (app_data->app_ui_data->progress_bar), (const gchar *) text);
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR
                      (app_data->app_ui_data->
                       progress_bar),
                      (float) get_counter(IMG_CURRENT)/(float) get_counter(IMG_MAX));
                    
        //TODO: free like this?
        g_free(text);
      
        break;
    }
    default:
        break;
    }
    return;
}
/** Hide progressbar again if needed
  * Sometimes when downloading either images or xml files (feeds)
  * The progressbar has to be shown if it is hidden
*/
void hide_progressbar_if_needed(void)
{

    AppUIData *app_ui_data = app_data->app_ui_data;
    //g_assert(app_ui_data != NULL);

    if( (getBooleanConfValue(DISABLE_TOOLBAR_FULLSCREEN) &&
        ( TRUE == hildon_appview_get_fullscreen
                    (hildon_app_get_appview(app_ui_data->app)))
            ) ||
            (
            getBooleanConfValue(DISABLE_TOOLBAR) &&
                (FALSE == hildon_appview_get_fullscreen
                    (hildon_app_get_appview(app_ui_data->app)))
            )
          ) // then hide the toolbar
        gtk_widget_hide(GTK_WIDGET(app_ui_data->main_toolbar));
    return;
}
/* tvh 2006May10 : made this so that applet can quit if refreshing from applet
*/

gboolean after_refresh_idle_quit(gpointer data)
{
    
    ULOG_DEBUG("after_refresh_idle_quit");
    //gdk_threads_enter();
    ULOG_DEBUG("after_refresh_idle_quit: Calling on_quit");
    on_quit(TRUE);
    ULOG_DEBUG("after_refresh_idle_quit: Finished on_quit");
    //gdk_threads_leave();
    //destroy the event source
    return FALSE;
}

void on_refresh_finished(gboolean disable_banner,
             gboolean do_not_signal_applet)
{
    AppUIData *app_ui_data;
    int search_mode = 0;
        
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    //g_assert(app_ui_data != NULL);

    search_mode = app_ui_data->search_mode;
    if (!app_ui_data->downloading_images)
        app_ui_data->search_mode = SFM_NORMAL;
    //FEED_MODE or FD_MODE is not true anymore. Give way for other    
   
    app_ui_data->progressbar_mode = PROGRESSBAR_NORMAL_MODE;
    //app_ui_data->receiving_feeds = FALSE;

    if (SFM_INTERRUPTED == search_mode) {
        ULOG_DEBUG("Refresh interrupted");
    
        if(app_ui_data->add_feed_dialog != NULL)
            newdialog_return_to_normal();
    
        if (!do_not_signal_applet &&
                app_ui_data->start_on_background &&
                !app_ui_data->network_disconnected)
            rss_dbus_signal_applet("refresh_interrupted");
        else if((!do_not_signal_applet &&
                !app_ui_data->start_on_background) ||
                (!do_not_signal_applet &&
                app_ui_data->start_on_background &&
                app_ui_data->network_disconnected))
            rss_dbus_signal_applet("refresh_finished");

        if (!disable_banner)
        {
            //only popup the banner if not on background
            if  (!app_ui_data->start_on_background)
                gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
                         _("rss_ib_refresh_interrupted"));
        
        }
    } else if (!do_not_signal_applet) {
        ULOG_DEBUG("Refresh completed");
        rss_dbus_refresh_completed();
    }

    ULOG_DEBUG("search mode: %d", search_mode);
    
    if (SFM_DOWNLOAD == search_mode ||
        SFM_INTERRUPTED == search_mode) {
        /* This happens if there are errors in refreshing, for example
        the feed is unavailable. This is set in ui_feed:ui_feed_process_update_result */
        if(app_ui_data->errors_in_refreshing) {   
            /* If the user has interrupted the refresh we dont
            display banners */
            if(search_mode != SFM_INTERRUPTED) {
                if(app_ui_data->add_feed_dialog == NULL) {
                    if(app_ui_data->start_on_background)
                        rss_dbus_signal_applet(OSSO_RSS_FEED_READER_REFRESH_PREVENTED_SIGNAL);
                    else if(app_ui_data->rss_max_download > 1)
                    {
                        //only popup if not running on background
                        if (!app_ui_data->start_on_background)
                            gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), 
                                         _("rss_ni_error_refreshing_all_feeds"));
                    }
                    else if(app_ui_data->rss_max_download == 1)
                       gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
                                         _("rss_ni_error_resfreshing_one_feed"));                      
                } else {
                    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_adding_failed"));
                    newdialog_return_to_normal();
                }
           }
           app_ui_data->errors_in_refreshing = FALSE;                 
        }
    //if (!app_ui_data->downloading_images)
    //    restore_search_button();
    //download_remove_dequeue_source();
    }
    /*tvh: Hide toolbar again if needed,
       according to gconf value and which fullscreen mode we're in*/
    //Only when we're not downloading images
    /* moved to hide_progressbar_if_needed and on_download_finished() */
    /*
    if (!app_ui_data->downloading_images)
        if( (
            getBooleanConfValue(DISABLE_TOOLBAR_FULLSCREEN) &&
                ( TRUE == hildon_appview_get_fullscreen
                    (hildon_app_get_appview(app_ui_data->app)))
            ) ||
            (
            getBooleanConfValue(DISABLE_TOOLBAR) &&
                (FALSE == hildon_appview_get_fullscreen
                    (hildon_app_get_appview(app_ui_data->app)))
            )
          ) // then hide the toolbar
            gtk_widget_hide(GTK_WIDGET(app_ui_data->main_toolbar));

    */
    // TODO: MUTEX??
    app_data->app_ui_data->rss_max_download = 0;
    app_data->app_ui_data->rss_current_download = 0;           
    //app_data->app_ui_data->images_max_download = 0;
    //app_data->app_ui_data->images_current_download = 0;           
    //app_data->app_ui_data->downloading_images = FALSE;

    if(!app_ui_data->start_on_background &&
       !do_not_signal_applet)
        set_update_timer(TRUE);
    
    if(app_ui_data->start_on_background)
    //    on_quit(TRUE);
     {
        ULOG_DEBUG("on_refresh_finished: Adding the g_idle for refresh quit owhen on background");
        //g_idle_add_full(G_PRIORITY_HIGH_IDLE, after_refresh_idle_quit, NULL, NULL);
        //g_idle_add( after_refresh_idle_quit, NULL);    
        on_close(NULL,NULL,NULL);
     }
     
     else
     {
        ULOG_DEBUG("on_refresh_finished: App didn't start on background");
     
     }
     return;
}

void on_refreshbtn_clicked_cb(nodePtr ptr)
{
    feed_schedule_update((feedPtr) ptr, FEED_REQ_DOWNLOAD_FAVICON);
}

void show_refresh_progressbar_button(void)
{
    AppUIData *app_ui_data = NULL;
     
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
      app_ui_data = app_data->app_ui_data;
    
    ULOG_DEBUG("....show_refresh_progressbar_button...mode = %d",app_ui_data->search_mode);
    //Shouldn't deal with the Loading... bar. It was set by feed_directory.c
    // because it's only one file. And it needs to be shown before downloading

    if (app_ui_data->feed_directory_loading)
    {
        ULOG_DEBUG("...............WTF..............fffffffffffffff........");
          app_ui_data->search_mode = SFM_DOWNLOAD;
        return;
    }
    if ((app_ui_data->search_mode == SFM_DOWNLOAD) &&
        ( app_ui_data->progressbar_mode == PROGRESSBAR_REFRESH_MODE))
        //already showing it. just return
        return;

    // now we really must update it!
    //only show if need to update
    /*
    if ((app_ui_data->search_mode != SFM_DOWNLOAD) &&
        ( app_ui_data->progressbar_mode != PROGRESSBAR_REFRESH_MODE))
    */
    {
         if (app_ui_data->progressbar_mode == PROGRESSBAR_SEARCH_MODE)
        {
            ULOG_DEBUG("oooooooooooooooERRORS!!!!!!!!!!!!!!oooooooooooooooooooo");
            ULOG_DEBUG("why in search mode, yet displaying refresh bar???");
            return;
         }
        ULOG_DEBUG("....SHOWING _refresh_progressbar_button...");
        //g_assert(NULL != app_ui_data->button_image_stop);
       g_return_if_fail (NULL != app_ui_data->button_image_stop);
       
       gtk_widget_set_sensitive(GTK_WIDGET
                   (app_ui_data->menuitem_refresh_feeds), FALSE);
       gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->refresh_button),
                FALSE);
       gtk_widget_set_sensitive(GTK_WIDGET
                   (app_ui_data->menuitem_add), FALSE);
       gtk_widget_set_sensitive(GTK_WIDGET
                   (app_ui_data->menuitem_importopml), FALSE);
       gtk_widget_set_sensitive(GTK_WIDGET
                   (app_ui_data->menuitem_feed_directory), FALSE);
                   
       gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->new_feed_button),
                   FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_unsubscribe),
                FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->unsubscribe_button),
                FALSE);

        //remove search field if nec. This is in fact *ALMOST* always the case
        // it's only wrong if progressbar_mode = IMAGE_MODE where that means
        // image requests are in front of feed requests. I wonder if it's ever the
        //case that image requests are being process. Then comes the feeds?
        //yes if we refresh while displaying a page with many images

        if ( (app_ui_data->search_field)->parent == app_ui_data->var_field)
        //if (app_ui_data->progressbar_mode == PROGRESSBAR_NORMAL_MODE)
        {
                  
          gtk_container_remove(GTK_CONTAINER(app_ui_data->var_field),
                   app_ui_data->search_field);
          gtk_container_add(GTK_CONTAINER(app_ui_data->var_field),
               app_ui_data->progress_bar);
            gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON
                       (app_ui_data->search_button),
                       app_ui_data->button_image_stop);

        }
        
        gtk_progress_bar_set_text(GTK_PROGRESS_BAR(app_ui_data->progress_bar),
                _("rss_ia_refreshing"));

       gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR
                       (app_ui_data->progress_bar), 0);


       /*tvh: show the toolbar if it's hidden 
        Then in on_refresh_finished, hide it again if necessary 
        */
       gtk_widget_show (GTK_WIDGET(app_ui_data->main_toolbar));

       app_ui_data->search_mode = SFM_DOWNLOAD;
       app_ui_data->progressbar_mode = PROGRESSBAR_REFRESH_MODE;

       rss_dbus_signal_applet(OSSO_RSS_FEED_READER_REFRESH_STARTED);
    }
}
/* Show the receiving xxx/yyy progressbar with a stop button
    This will only take effect if we're not in
    1. SFM_SEARCH
    2. SFM_DOWNLOAD for feeds or feed directory
    In those cases, FALSE will be returned
    Otherwise, TRUE is returned, meaning the xxx/yyy is shown
    
    We need true/false in this case, different from show_refresh_progress_button
    as REFRESH mode has higher priority
*/
gboolean show_receiving_images_progressbar_button(void)
{

    AppUIData *app_ui_data = NULL;
    gchar * text = NULL;

    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    ULOG_DEBUG("show_receiving_images_progressbar_button");
    if (!app_data->app_ui_data->downloading_images)
        app_data->app_ui_data->downloading_images = TRUE;

    //TODO: searching is taking place. Images are never downloaded before search is done
    //anyway. so this is wrong. It can never happen!!! But ... might be the case
    // that the download_done cb in gtkhtml is activated before the search has finished finalizing??
    if (app_data->app_ui_data->search_mode == SFM_SEARCH )
    {
        ULOG_DEBUG("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nyour fucked");
        return FALSE;
        
    }
    //refreshing is taking place
    if ( (app_data->app_ui_data->search_mode == SFM_DOWNLOAD) && 
         (app_data->app_ui_data->progressbar_mode == PROGRESSBAR_REFRESH_MODE))
        /* || 
            (app_data->app_ui_data->progressbar_mode == PROGRESSBAR_RECEIVING_MODE)
        ))*/
        return FALSE;
    // if already in the mode: return TRUE
    if ( (app_data->app_ui_data->search_mode == SFM_DOWNLOAD) && 
         (app_data->app_ui_data->progressbar_mode == PROGRESSBAR_IMAGE_MODE))
        return TRUE;
        
    //Now it must be (SET to) receiving image progress bar & button
    /*if ((app_data->app_ui_data->search_mode != SFM_DOWNLOAD) || 
        (app_data->app_ui_data->progressbar_mode != PROGRESSBAR_IMAGE_MODE) )
        */
    else 
    {
        app_ui_data = app_data->app_ui_data;
       
       //g_assert(NULL != app_ui_data->button_image_stop);
       g_return_val_if_fail((NULL != app_ui_data->button_image_stop), FALSE);
       
       //if (app_ui_data->progressbar_mode == PROGRESSBAR_NORMAL_MODE)
           //maybe previous mode was FEED_MODE OR FD_MODE.
           //so let's do a check first

        if ( (app_ui_data->search_field)->parent == app_ui_data->var_field)
       {
           ULOG_DEBUG("show_receiving_images_progressbar_button ....1");
    
              gtk_container_remove(GTK_CONTAINER(app_ui_data->var_field),
                           app_ui_data->search_field);
             gtk_container_add(GTK_CONTAINER(app_ui_data->var_field),
                          app_ui_data->progress_bar);            
              gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON
                       (app_ui_data->search_button),
                       app_ui_data->button_image_stop);
        }

            ULOG_DEBUG("show_receiving_images_progressbar_button....2");
        // might have "receiving 0/0" here....       
        /*
        text = g_strdup_printf(_("rss_pb_toolbar_receiving"), 
                app_data->app_ui_data->images_current_download,
                      app_data->app_ui_data->images_max_download);
        */
        text = g_strdup_printf(_("rss_pb_toolbar_receiving"), 
                        get_counter(IMG_CURRENT), get_counter(IMG_MAX));
        
        gtk_progress_bar_set_text(GTK_PROGRESS_BAR
                    (app_data->app_ui_data->progress_bar), (const gchar *) text);
        
       
        ULOG_DEBUG("show_receiving_images_progressbar_button....3");
         gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR
                       (app_ui_data->progress_bar), 0);



       //TODO: uhm......how to free this 
       g_free(text);
       /*tvh: show the toolbar if it's hidden 
        Then in on_download_finished, hide it again if necessary 
        */
       gtk_widget_show (GTK_WIDGET(app_ui_data->main_toolbar));

       app_ui_data->search_mode = SFM_DOWNLOAD;
       app_data->app_ui_data->progressbar_mode = PROGRESSBAR_IMAGE_MODE;
    }
    return TRUE;
}


/* on receiving image  finished, there's nothing we have to do
    except resetting the progress bar.
    the task of showing the seach_button is done in on_download_finished
*/
void on_receiving_images_finished(void)
{
    
    app_data->app_ui_data->images_max_download = 0;
    app_data->app_ui_data->images_current_download = 0;           
    app_data->app_ui_data->downloading_images = FALSE;
    //this condition should always be TRUE actually. 
    // It must have been assured before calling this as images have the lowest priority 
    // out of three queues
    if (app_data->app_ui_data->progressbar_mode == PROGRESSBAR_IMAGE_MODE)
        app_data->app_ui_data->progressbar_mode = PROGRESSBAR_NORMAL_MODE;
    return;
}

void restore_search_button(void)
{
    AppUIData *app_ui_data = NULL;
     
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    app_data->app_ui_data->search_mode = SFM_NORMAL;
    app_data->app_ui_data->progressbar_mode = PROGRESSBAR_NORMAL_MODE;
    app_data->app_ui_data->downloading_images = FALSE;

    ULOG_DEBUG("restore_search_button");
        
    app_ui_data = app_data->app_ui_data;

    //g_assert(NULL != app_ui_data->button_image_refresh);
    g_return_if_fail(NULL != app_ui_data->button_image_refresh);

    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_ui_data->menuitem_refresh_feeds), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->refresh_button),
                 TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_add), TRUE);

    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_importopml), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_feed_directory), TRUE);

    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->new_feed_button),
                TRUE);            
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_unsubscribe),
                TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->unsubscribe_button),
                TRUE);                                                     
    gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON
                    (app_ui_data->search_button),
                    app_ui_data->button_image_search);
    if ( (app_ui_data->progress_bar)->parent == app_ui_data->var_field)
    {
        gtk_container_remove(GTK_CONTAINER(app_ui_data->var_field),
             app_ui_data->progress_bar);
        gtk_container_add(GTK_CONTAINER(app_ui_data->var_field),
              app_ui_data->search_field);
    }
}

void on_refresh_action_response(GtkWidget * refresh_dialog,
                gint responseId, gpointer user_data)
{
    gboolean res = FALSE;

    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);

    if (responseId == GTK_RESPONSE_CANCEL ||
    responseId == GTK_RESPONSE_DELETE_EVENT) {
    gtk_widget_destroy(GTK_WIDGET(refresh_dialog));
        app_data->app_ui_data->refresh_dialog = NULL;
    ULOG_INFO("Refresh was canceled");
    return;
    }

    GtkWidget *selected_feed_widget =
    lookup_widget(refresh_dialog, "radio_button_selected_feed");
    gboolean selectedFeedOnly =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
                     (selected_feed_widget));
    
    ULOG_INFO("Refresh invoked...");

    if (selectedFeedOnly) {
    ULOG_INFO("Selected feed only...");
    nodePtr ptr = ui_feedlist_get_selected();

    if (ptr != NULL && ptr->type == FST_FEED) {
            ULOG_DEBUG("Updating the selected feed");
        feed_schedule_update((feedPtr) ptr, FEED_REQ_DOWNLOAD_FAVICON);
    }
    } 
    else {
    ULOG_DEBUG("Updating all feeds");
    res = ui_feedlist_do_for_all_with_int(NULL,
                          ACTION_FILTER_FEED,
                          update_node);
    }
    gtk_widget_destroy(GTK_WIDGET(refresh_dialog));
    app_data->app_ui_data->refresh_dialog = NULL;
}

void on_refresh_action(AppUIData * app_ui_data)
{
    GtkWidget *refreshdialog = NULL;

    //g_assert(NULL != app_ui_data);

    /* A refresh dialog has already been created so lets not create
       another one. This might happen when several iap requests are
       sent in succession */
    
    if(app_ui_data->refresh_dialog == NULL)
    {
        app_ui_data->refresh_dialog =
            refreshdialog = create_refresh_dialog();
        g_signal_connect(G_OBJECT(refreshdialog), "response",
                 G_CALLBACK(on_refresh_action_response), NULL);
        gtk_widget_show(refreshdialog);
    }
}

void on_new_folder_action_response(GtkWidget * new_folder_dialog,
                   gint responseId, gpointer user_data)
{
    GtkWidget *dialog = NULL;
    GtkWidget *folder_view = NULL;

    ULOG_DEBUG(__FUNCTION__);

    dialog = (GtkWidget *) user_data;
    if (dialog != NULL) {
    folder_view = lookup_widget(dialog, "folder_view");
    //g_assert(folder_view != NULL);
    g_return_if_fail (folder_view != NULL);
    ULOG_DEBUG("attempting to grab focus");
    gtk_widget_grab_focus(folder_view);
    }
}

gboolean update_node(nodePtr ptr)
{
    //g_assert(app_data != NULL);
    g_return_val_if_fail((app_data != NULL), FALSE);
    //g_assert(app_data->app_ui_data != NULL);
    g_return_val_if_fail( (app_data->app_ui_data != NULL), FALSE);
    feed_schedule_update((feedPtr) ptr, FEED_REQ_RESET_TITLE |
                    FEED_REQ_RESET_UPDATE_INT | 
                    FEED_REQ_AUTO_DISCOVER |
                    FEED_REQ_DOWNLOAD_FAVICON);
    /*ULOG_DEBUG("update_node: search_mode = %d",
           app_data->app_ui_data->search_mode);*/
    return app_data->app_ui_data->search_mode != SFM_INTERRUPTED;
}

void display_new_folder_dialog(GtkWidget * widget, GtkWindow *parent_window)
{
    GtkTreeView *folder_view = NULL;
    GtkTreeViewColumn *column = NULL;
    GtkTreePath *path = NULL;
    GtkTreeIter iter;
    nodePtr node = NULL;
    folderPtr folder = NULL;
    GValue value;

    ULOG_DEBUG(__FUNCTION__);

    memset(&value, 0, sizeof(value));
    if (widget != NULL) {
    ULOG_DEBUG("widget was set on call");
    folder_view = GTK_TREE_VIEW(lookup_widget(widget, "folder_view"));
    //g_assert(folder_view != NULL && GTK_IS_TREE_VIEW(folder_view));
    g_return_if_fail (folder_view != NULL && GTK_IS_TREE_VIEW(folder_view));
    gtk_tree_view_get_cursor(folder_view, &path, &column);
    if (path != NULL) {
        ULOG_DEBUG("and we got path to cursor");
        gtk_tree_model_get_iter(gtk_tree_view_get_model
                    (folder_view), &iter, path);
        gtk_tree_model_get_value(gtk_tree_view_get_model
                     (folder_view), &iter, FS_PTR, &value);
        node = (nodePtr) g_value_peek_pointer(&value);
        if (node != NULL && node->type == FST_FOLDER) {
        folder = (folderPtr) node;
        }
        gtk_tree_path_free(path);
        path = NULL;
    }
    }
    GtkWidget *nfdialog = create_newfolderdialog(folder, folder_view, parent_window);
    g_value_unset(&value);
    g_signal_connect(G_OBJECT(nfdialog), "response",
             G_CALLBACK(on_new_folder_action_response), widget);
    gtk_widget_show(nfdialog);
}



/*TODO: this will be removed later as SAVE & SAVEAS is not used
 * Actually it can be used for exporting opml file
 */
enum e_file_chooser {
    FILE_CHOOSER_LOAD,
    FILE_CHOOSER_SAVE,
    FILE_CHOOSER_SAVEAS
};

static
HildonFileSystemModel *get_file_system_model(AppUIData *app_ui, GtkWidget *ref_widget)
{
  g_return_val_if_fail(app_ui != NULL, NULL);
  
  if(!app_ui->model) {
    app_ui->model = g_object_new (HILDON_TYPE_FILE_SYSTEM_MODEL,
                  "ref_widget", ref_widget,
                  NULL);
  }

  return app_ui->model;
}

/**
  Show a Hildon file chooser dialog
   
  @param app_ui Application UI data
  @param save #FILE_CHOOSER_LOAD, #FILE_CHOOSER_SAVE, or
              #FILE_CHOOSER_SAVEAS
  @return filename (g_free'd by the caller), or NULL if problems
*/
static
const gchar *file_chooser( AppUIData *app_ui, enum e_file_chooser mode )
{
    //g_assert(app_ui);
    
    HildonFileSystemModel *model= NULL;
    GtkWidget *fs= NULL;
    model= get_file_system_model(app_ui, GTK_WIDGET(app_ui->app));
    
    if (!model)
      return NULL;

    fs = hildon_file_chooser_dialog_new_with_properties(
                  GTK_WINDOW(app_ui->app),
                  "file_system_model", model,
                  "action", GTK_FILE_CHOOSER_ACTION_OPEN,
                  NULL);
    gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(fs), FALSE);
    gtk_window_set_modal (GTK_WINDOW(fs), TRUE);

        if (!fs) {
        ULOG_ERR( "Failed creating HildonFileChooserDialog" );
        return NULL;
    }

    /* RSS News Reader-specific dialog texts
    */
    gtk_window_set_title( GTK_WINDOW (fs), _("rss_ti_select_opml_file") );

    /*what's this? 
    g_object_set( G_OBJECT (fs), 
                  "empty-text", _("sfil_va_select_object_no_objects_sketches"),
                  NULL );
    */

    // Filtering .xml or .opml files only
    GtkFileFilter *filter= gtk_file_filter_new();

    gtk_file_filter_add_pattern( filter, "*."OPML_FILE_EXTENSION_1 );//.xml
    gtk_file_filter_add_pattern( filter, "*."OPML_FILE_EXTENSION_2 );//.opml

    gtk_file_chooser_add_filter( GTK_FILE_CHOOSER (fs), filter );
    gtk_file_chooser_set_filter( GTK_FILE_CHOOSER (fs), filter );

    /* Folder that the dialog opens in 
    */
    const gchar *uri= "/home/user/MyDocs";
    //g_assert(uri);
    gtk_file_chooser_set_current_folder_uri (GTK_FILE_CHOOSER (fs), uri);

    /*
     ossohelp_dialog_help_enable(GTK_DIALOG(fs),
                                SKETCH_TOPIC_OPENFILE, 
                                app_ui->osso);
    */
    /* Run the dialog
    */
    if ( gtk_dialog_run( GTK_DIALOG (fs) ) != GTK_RESPONSE_OK ) {
        /* dialog cancelled, exit */
        gtk_widget_destroy(fs);
        return NULL;
    }

    /* check that user did not just click OK
        without selecting anything
     */
    const gchar *fn= gtk_file_chooser_get_uri( GTK_FILE_CHOOSER (fs) );
    if (!fn) {
        gtk_widget_destroy(fs);
        return NULL;
    }

    /* "Houston, we have selection.."
    */
    const gchar *uri_new = gtk_file_chooser_get_current_folder_uri(GTK_FILE_CHOOSER(fs));
     if (!uri_new) {
     }
     /*
     else if ( strcmp( app_ui->uri_, uri_new ) != 0 ) {        // changed the folder
        g_free(app_ui->uri_);
        app_ui->uri_= uri_new;        
        uri_new=NULL;
        const gchar *cwd = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(fs));
        if (!cwd) {
         } else if (strcmp(app_ui->cwd_,cwd)!=0){
        g_free(app_ui->cwd_);
        app_ui->cwd_=cwd;
        g_free(cwd), cwd=NULL; 
       } 
      }
    */ 
    gtk_widget_destroy(fs), fs=NULL;

    // append opml file extension if necessary 
    // How could this happen? Didn't the pattern filter work?
    
    /*
    if (!g_str_has_suffix( fn, "." OPML_FILE_EXTENSION_1 )) {
        gchar *tmp= g_strconcat( fn, "." OPML_FILE_EXTENSION_1, NULL );
        g_free(fn);
        fn= tmp;
    }
    else if (!g_str_has_suffix( fn, "." OPML_FILE_EXTENSION_2 )) {
        gchar *tmp= g_strconcat( fn, "." OPML_FILE_EXTENSION_2, NULL );
        g_free(fn);
        fn= tmp;
    }
    */
    return fn;
}


void display_file_chooser_dialog(GtkWidget *widget, gpointer user_data)
{
    ULOG_DEBUG("display_file_chooser_dialog ...........\n");
    AppUIData *app_ui_data;
    
   // g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    const gchar *filename = file_chooser( app_data->app_ui_data, FILE_CHOOSER_LOAD);
    ULOG_DEBUG("display_file_chooser_dialog: Browsed opml file name = %s \n", filename);

    //now update the opmlAdressEntry field
    GtkWidget *opmlAdressEntry = lookup_widget (GTK_WIDGET (app_ui_data->dialog),"opmlAddressEntry");
    gtk_entry_set_text(GTK_ENTRY(opmlAdressEntry),filename);
     
}



void on_manage_folders_action_response(GtkWidget * manage_folders_dialog,
                       gint responseId, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
     
   // g_assert(app_data != NULL);
   // g_assert(app_data->app_ui_data != NULL);
    

    app_ui_data = app_data->app_ui_data;

    GtkTreeView *treeview =
    GTK_TREE_VIEW(lookup_widget(manage_folders_dialog, "folder_view"));
    nodePtr node = ui_feedlist_get_selected_from_tree(treeview);

    ULOG_DEBUG(__FUNCTION__);

    switch (responseId) {
    case RSS_RESPONSE_NEW_FOLDER:
    display_new_folder_dialog(manage_folders_dialog, GTK_WINDOW(user_data));
    break;
    case RSS_RESPONSE_RENAME_FOLDER:
    ULOG_INFO("Rename folder");

    if ((node != NULL) && (node->type == FST_FOLDER)) {
        folderPtr folder = (folderPtr) node;
        ULOG_INFO("Selected folder for rename: %s", folder->title);
        if (folder_get_nonremovable(folder))
        /*
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
                  _("rss_ib_renaming_folder_not_allowed"));
        */
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
                  dgettext("hildon-common-strings", "ckdg_ib_unable_to_rename"));

        else
        create_folder_rename_dialog(folder);
    } else {
        ULOG_INFO("No selection, or selection is not a folder!");
    }

    break;

    case RSS_RESPONSE_DELETE_FOLDER:
    ULOG_INFO("Delete folder");

    if ((node != NULL) && (node->type == FST_FOLDER)) {
        folderPtr folder = (folderPtr) node;
        ui_feedlist_confirm_and_delete_folder(folder);
    } else {
        ULOG_INFO("No selection, or selection is not a folder!");
    }

    break;
    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_DELETE_EVENT:
    app_ui_data->dialog = NULL;
    gtk_widget_destroy(GTK_WIDGET(manage_folders_dialog));
    break;
    }
}

void on_change_folder_action_response(GtkWidget * change_folder_dialog,
                      gint responseId, gpointer user_data)
{
    ULOG_DEBUG(__FUNCTION__);

    struct fp_prop_ui_data *ui_data = (struct fp_prop_ui_data *) user_data;

    switch (responseId) {
    case RSS_RESPONSE_NEW_FOLDER:
    display_new_folder_dialog(change_folder_dialog, GTK_WINDOW(ui_data->parent_window));
    break;
    case GTK_RESPONSE_OK:
    ULOG_INFO("change folder");
    GtkTreeView *treeview =
        GTK_TREE_VIEW(lookup_widget
              (change_folder_dialog, "folder_view"));
    nodePtr node = ui_feedlist_get_selected_from_tree(treeview);
    if ((node != NULL) && (node->type == FST_FOLDER)) {
        folderPtr folder = (folderPtr) node;
        ULOG_INFO("Selected folder to change to: %s", folder->title);

        ui_data->parent_folder = folder;
        gtk_label_set_text(GTK_LABEL(ui_data->label_widget),
                   folder->title);
    } else {
        ULOG_INFO
        ("No selection, or selection is not a folder! Cannot change.");
    }
    gtk_widget_destroy(GTK_WIDGET(change_folder_dialog));
    //ui_data is destroyed by the parent dialog

    break;
    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_DELETE_EVENT:
    gtk_widget_destroy(GTK_WIDGET(change_folder_dialog));
    break;
    }
}


void display_change_folder_dialog(GtkWidget * widget, gpointer user_data)
{
    GtkTreeView *folder_view = NULL;
    GtkTreeModel *folder_model = NULL;
    folderPtr current_folder = NULL;
    GtkTreePath *current_path = NULL;
    struct fp_prop_ui_data *data = NULL;

    ULOG_DEBUG(__FUNCTION__);

    data = (struct fp_prop_ui_data *) user_data;
    if (data->parent_folder != NULL) {
    current_folder = data->parent_folder;
    } else {
    current_folder = ui_feedlist_get_root_folder();
    }
   // g_assert(current_folder != NULL);
    g_return_if_fail(current_folder != NULL);

    GtkWidget *cfdialog = data->parent_window = create_change_folder_dialog(GTK_WINDOW(data->dialog));
    populate_folders(cfdialog);

    /* Select the root folder. */
    folder_view = GTK_TREE_VIEW(lookup_widget(cfdialog, "folder_view"));
    //g_assert(folder_view != NULL);
     g_return_if_fail(current_folder != NULL);
    folder_model = gtk_tree_view_get_model(folder_view);
    //g_assert(folder_model != NULL);
    g_return_if_fail (folder_model != NULL);
    ui_feedlist_search_model_by_ptr(folder_model,
                    (nodePtr) current_folder,
                    &current_path, NULL);
    gtk_tree_view_set_cursor(folder_view, current_path, NULL, FALSE);
    gtk_tree_path_free(current_path);
    current_path = NULL;

    g_signal_connect(G_OBJECT(cfdialog), "response",
             G_CALLBACK(on_change_folder_action_response),
             user_data);
    ULOG_INFO("Change folder dialog created, showing ...\n");
    gtk_widget_show(cfdialog);
    ULOG_INFO("Change folder dialog created, on screen.\n");
}


void on_managefoldersbtn_clicked(GtkButton * button, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
     
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    ULOG_DEBUG(__FUNCTION__);

    if (SFM_SEARCH == app_ui_data->search_mode){
        ULOG_DEBUG("rss_ib_unable_managefolders_searching");
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_managefolders_searching"));
        return;
    }
    else {
        ULOG_DEBUG("wtf");
    }

    ULOG_DEBUG("on_managefoldersbtn_clicked: creating mng folder");
    GtkWidget *mfdialog = app_ui_data->dialog =
    create_manage_folders_dialog();

    ULOG_DEBUG("on_managefoldersbtn_clicked: populating mng folder");
    populate_folders(mfdialog);
    ULOG_DEBUG("on_managefoldersbtn_clicked: DONE populating mng folder");

    g_signal_connect(G_OBJECT(mfdialog), "response",
             G_CALLBACK(on_manage_folders_action_response), mfdialog);

    ULOG_DEBUG("on_managefoldersbtn_clicked: signal connected");
    gtk_widget_show(mfdialog);
}

void on_prefdialog_response(GtkWidget * prefdialog, gint responseId,
                gpointer user_data)
{
    guint hours = 0;
    guint minutes = 0;
    guint seconds = 0;
    gboolean close = TRUE;
    gboolean old_sortOrder ;

    if (responseId == GTK_RESPONSE_CANCEL ||
    responseId == GTK_RESPONSE_DELETE_EVENT) {
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    gtk_widget_destroy(GTK_WIDGET(prefdialog));
    //restore_time_locale ( (gchar*)user_data );
    ULOG_INFO("Settings dialog was canceled");
    return;
    }

    ULOG_INFO("Settings processing begins");

    GtkWidget *openLinksPulldown =
        lookup_widget(prefdialog, "openLinksPulldown");
    GtkWidget *sortOrderPulldown =
        lookup_widget(prefdialog, "sortOrderPulldown");
    GtkWidget *radio_button_automatically =
        lookup_widget(prefdialog, "radio_button_automatically");
    GtkWidget *updateIntervalWidget =
        lookup_widget(prefdialog, "update_interval_widget");

    /*gboolean manually =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
                     (radio_button_manually));
    ULOG_DEBUG("manually=%s", manually ? "true" : "false");*/

    gboolean automatically =
        gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
                     (radio_button_automatically));
    ULOG_DEBUG("automatically=%s", automatically ? "true" : "false");

    gboolean openLinksSelection =
        (gtk_combo_box_get_active(GTK_COMBO_BOX(openLinksPulldown)) == 0);
    ULOG_DEBUG("open links: %s",
           openLinksSelection ? "reuse window" : "new window");

    gboolean sortOrderSelection =
        (gtk_combo_box_get_active(GTK_COMBO_BOX(sortOrderPulldown)) == 0);
    ULOG_DEBUG("sort order: %s",
           sortOrderSelection ? "newest first" : "oldest first");

    hildon_time_editor_get_time(HILDON_TIME_EDITOR(updateIntervalWidget),
                &hours, &minutes, &seconds);
    int updateInterval = hours * 60 + minutes;
    if (updateInterval < MIN_INTERVAL)
    {
        //gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_interval_cannot_be_under_15min"));
        gchar *tmp =  g_strdup_printf (dgettext("hildon-common-strings", "ckct_ib_minimum_value"),(int) MIN_INTERVAL);
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),tmp);
        g_free(tmp);
        updateInterval = MIN_INTERVAL;    //minutes
    
    hildon_time_editor_set_time(HILDON_TIME_EDITOR(updateIntervalWidget),
                updateInterval / 60,
                updateInterval % 60, 0);

    close = FALSE;
    }
    
    setBooleanConfValue(RSS_SETTINGS_REUSE_BROWSER_WINDOW,
            openLinksSelection);
    old_sortOrder = getBooleanConfValue(RSS_SETTINGS_NEWEST_POSTS_FIRST);
    if (old_sortOrder != sortOrderSelection)
    {
        setBooleanConfValue(RSS_SETTINGS_NEWEST_POSTS_FIRST,
            sortOrderSelection);
        if (displayed_node != NULL)
        //display the current feed page again if not searchFeed ??
        //    if (displayed_node->type != FST_VFOLDER)
        {
            /*
            gtkhtml_stop();    
            ui_htmlview_write(app_data->app_ui_data->html_scrollpane,NULL,NULL, displayed_node);
            */
            ui_feedlist_reload_selected(displayed_node);
        }
    }
    gboolean wasAutomatic = getBooleanConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES);
                        
    setBooleanConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES, automatically);
    setNumericConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES_INTERVAL,
            updateInterval);

    if(close)
    {
    gtk_widget_destroy(GTK_WIDGET(prefdialog));
    
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    if(wasAutomatic)
    {
        ULOG_DEBUG("Removing automatic update timer");
        if(!g_source_remove(app_data->app_ui_data->automatic_update_source))
            ULOG_DEBUG("Source for automatic update timer not found!");
        
    }
    
    if(automatically)
            set_update_timer(FALSE);    
    
    //restore_time_locale ( (gchar*)user_data );
    }
      
    ULOG_INFO("Settings processing ends");
}

void radio_button_manually_toggled_cb (GtkToggleButton *togglebutton,
                                       gpointer         user_data)
{
    GtkWidget *prefdialog = (GtkWidget *) user_data;
    GtkWidget *updateIntervalWidget =
        lookup_widget(prefdialog, "update_interval_widget");

    if (gtk_toggle_button_get_active(togglebutton) == TRUE)
    {
        gtk_widget_set_sensitive(updateIntervalWidget, FALSE);
        ULOG_DEBUG("\n\n\n\n should insensitivize the time_editor ehre ");

    }
    else
        gtk_widget_set_sensitive(updateIntervalWidget, TRUE);

}
/* This function is not needed as the "toggled" signal 
   will be activated for radio_button_manual in any case 
   as it is either the selected button, or previously selected button
*/
/*
void radio_button_automatically_toggled_cb (GtkToggleButton *togglebutton,
                                       gpointer         user_data)
{
    ULOG_DEBUG("\n\n\n\nnot needed ");
}
*/


void on_refreshbtn_clicked(GtkButton * button, gpointer user_data)
{
    on_refresh_menuitem_selected(NULL,user_data);             
}

void on_refresh_menuitem_selected(GtkWidget * menuitem, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
    nodePtr ptr = NULL;
    
    ULOG_DEBUG("***refresh menuitem selected");
    app_ui_data = (AppUIData *) user_data;
    //g_assert(app_ui_data != NULL);

    if (!GTK_WIDGET_SENSITIVE(app_data->app_ui_data->menuitem_refresh_feeds)) {
        if (SFM_DOWNLOAD == app_ui_data->search_mode) 
            gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
                 _("rss_ib_refresh_already_started"));
        else if (SFM_SEARCH == app_ui_data->search_mode) 
            gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_cannot_start_refresh"));
        else
            ULOG_DEBUG("oooooooooooooooooooooERRORooooooooooooooooooooooooo");
        return;
    }
    if (app_ui_data->low_mem)
    {
        gtk_infoprint( GTK_WINDOW(app_data->app_ui_data->app), 
                    dgettext("ke-recv", "memr_ib_operation_disabled") );
        return;                    
    }
    ptr = ui_feedlist_get_selected();
    
    /* 2006June06: tvh: no reason to call this while refreshing */
    //ui_search_clear_search_results();
    
    if(displayed_node == (nodePtr)searchFeed)
        ui_mainwindow_show_statistics();

    if(ptr == NULL || ptr->type != FST_FEED)
        app_ui_data->iap_action = OSSO_IAP_REFRESH_ALL_NO_DIALOG;
    else //pop up the refresh dialog
        app_ui_data->iap_action = OSSO_IAP_POPUP_REFRESH_DIALOG;

    ULOG_DEBUG("refresh clicked:...requesting connection");
    
    osso_iap_connect(OSSO_IAP_ANY, OSSO_IAP_REQUESTED_CONNECT,
            app_ui_data);

}

void toggle_folder_list(GtkWidget * button, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
     
    gboolean state = FALSE;

    debug_enter("toggle_folder_list");
    
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    state = !(GTK_WIDGET_VISIBLE(feedlist_container) != 0);


    gboolean is_active =
        gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON (app_ui_data->folders_list_toggle_button));
    gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON (app_ui_data->folders_list_toggle_button),
                                             !is_active);

    ui_set_folder_list_state(app_ui_data, state, FALSE);

    /* Action for the folder_list_button */

    debug_exit("toggle_folder_list");
}


void on_folder_list_button_clicked(GtkWidget * button, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
     
    gboolean state = FALSE;

    debug_enter("on_folder_list_button_clicked");
    
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    
    state = !(GTK_WIDGET_VISIBLE(feedlist_container) != 0);

    ui_set_folder_list_state(app_ui_data, state, TRUE);

    debug_exit("on_folder_list_button_clicked");
}


void toggle_toolbar(GtkWidget * widget, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
    GtkWidget *toolbar;
    
    ULOG_DEBUG(__FUNCTION__);
    
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
     
    app_ui_data = app_data->app_ui_data;
    
    //g_assert(app_ui_data->main_toolbar != NULL);
    
    toolbar = app_ui_data->main_toolbar;
        
    gboolean fullscreen =
    hildon_appview_get_fullscreen(hildon_app_get_appview
                      (app_ui_data->app)) != 0;
     
    setBooleanConfValue(DISABLE_TOOLBAR_FULLSCREEN, !GTK_CHECK_MENU_ITEM(
                        app_ui_data->menuitem_full)->active);
    setBooleanConfValue(DISABLE_TOOLBAR, !GTK_CHECK_MENU_ITEM(
                        app_ui_data->menuitem_normal)->active);        
    
    if ((!fullscreen && widget == GTK_WIDGET(app_ui_data->menuitem_full))
    || (fullscreen
        && widget == GTK_WIDGET(app_ui_data->menuitem_normal)))
    return;
    
    if (GTK_WIDGET_VISIBLE(toolbar)) {
    gtk_widget_hide(toolbar);
    } else {
        gtk_widget_show(toolbar);
    }
}


void on_toggle_fullscreen_mode(GtkWidget * widget, gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
    gboolean state = FALSE;

    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    //g_assert(app_ui_data != NULL);

    state = !(hildon_appview_get_fullscreen
          (hildon_app_get_appview(app_ui_data->app))
          != 0);
    ui_set_fullscreen_state(app_ui_data, state, FALSE);
}


void toggle_fullscreen_and_update_menu(GtkWidget * widget,
                       gpointer user_data)
{
    AppUIData *app_ui_data = NULL;
    gboolean state = FALSE;

    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    //g_assert(app_ui_data != NULL);

    state = !(hildon_appview_get_fullscreen
          (hildon_app_get_appview(app_ui_data->app))
          != 0);
    ui_set_fullscreen_state(app_ui_data, state, TRUE);
    
}

void on_cut(GtkAction * action, gpointer data)
{
    GtkWidget *focus = NULL;
    AppUIData *app_ui_data = NULL;

    app_ui_data = (AppUIData *) data;
    //g_assert(app_ui_data != NULL);

    if (GTK_WIDGET_SENSITIVE(app_ui_data->menuitem_cut)) {
    if (NULL != app_ui_data->dialog)
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->dialog));
    else
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->app));

    if (GTK_IS_EDITABLE(focus))
        gtk_editable_cut_clipboard(GTK_EDITABLE(focus));
    return;
    }
    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_to_cut"));
}

void on_copy(GtkAction * action, gpointer data)
{

    GtkWidget *focus = NULL;
    AppUIData *app_ui_data = NULL;

    app_ui_data = (AppUIData *) data;
    //g_assert(app_ui_data != NULL);

    if (GTK_WIDGET_SENSITIVE(app_ui_data->menuitem_copy)) {
        if (NULL != app_ui_data->dialog)
            focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->dialog));
        else
            focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->app));

        if (GTK_IS_EDITABLE(focus)) //like dialog entry or stuff
        {
            gtk_editable_copy_clipboard(GTK_EDITABLE(focus));
            gtk_infoprint(GTK_WINDOW(app_ui_data->app),
                      _("ecoc_ib_edwin_copied"));
              return;
        }
        else { //check if gtkhtml widget has focus and then
               //make sure it has selection first 
        //not gtkeditable, not feedlist, it should be gtkhtml
        //if ( (focus != feedlist)
        if ((GTK_WIDGET(focus) != GTK_WIDGET(feedlist))        
                     && htmlwidget_has_selection (app_data->app_ui_data->html_scrollpane))
            {
                htmlwidget_copy_selection (GTK_WIDGET(app_data->app_ui_data->html_scrollpane));
                gtk_infoprint(GTK_WINDOW(app_ui_data->app),
                              _("ecoc_ib_edwin_copied"));
                  return;
            }
            else
            {
                ULOG_DEBUG ("on_copy: ***ERRORS****\nSomething must have gone wrong here.\
                            htmlwidget has no selection. Dialog is not focused. \
                            Nothing is selected \n \
                            Yet copy is SENSITIVE. Recheck the code \n ");                          
            }
        }

    }
    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_to_copy"));


    //delete the gtktextview stuff use gtkhtml in stead
/*
    GtkWidget *focus = NULL;
    AppUIData *app_ui_data = NULL;

    app_ui_data = (AppUIData *) data;
    g_assert(app_ui_data != NULL);

    if (GTK_WIDGET_SENSITIVE(app_ui_data->menuitem_copy)) {
    if (NULL != app_ui_data->dialog)
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->dialog));
    else
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->app));

    if (GTK_IS_EDITABLE(focus))
    {
        gtk_editable_copy_clipboard(GTK_EDITABLE(focus));
        gtk_infoprint(GTK_WINDOW(app_ui_data->app),
                  _("ecoc_ib_edwin_copied"));
    }
    return;
    }
    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_to_copy"));
*/
}

void on_paste(GtkAction * action, gpointer data)
{
    GtkWidget *focus = NULL;
    AppUIData *app_ui_data = NULL;

    app_ui_data = (AppUIData *) data;
    //g_assert(app_ui_data != NULL);

    GtkClipboard *clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
    if (gtk_clipboard_wait_is_text_available(clipboard)) {

    if (GTK_WIDGET_SENSITIVE(app_ui_data->menuitem_paste)) {
        if (NULL != app_ui_data->dialog)
        focus =
            gtk_window_get_focus(GTK_WINDOW(app_ui_data->dialog));
        else
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->app));

        if (GTK_IS_EDITABLE(focus))
        gtk_editable_paste_clipboard(GTK_EDITABLE(focus));
        return;
    }
    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_to_paste"));
    } else
    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_nothing_to_paste"));
}

void on_selectall(GtkAction * action, gpointer data)
{
    GtkWidget *focus = NULL;
    AppUIData *app_ui_data = NULL;

    app_ui_data = (AppUIData *) data;
    //g_assert(app_ui_data != NULL);

    ULOG_DEBUG("on_selectall");
    if (GTK_WIDGET_SENSITIVE(app_ui_data->menuitem_selectall)) {
        if (NULL != app_ui_data->dialog)
            focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->dialog));
        else
            focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->app));

        if (GTK_IS_EDITABLE(focus)) //like dialog entry or stuff
        {
            gtk_editable_select_region(GTK_EDITABLE(focus), 0, -1);
            ULOG_DEBUG("gtk editable SELECTED ALL");
              return;
        }
        else { //must be the htmlwidget that has selection 
                //make sure it has selection first 
                //not gtkeditable, not feedlist, it should be gtkhtml
            ULOG_DEBUG("on_selectall: checking if focus is not feedlist...");
            if (GTK_WIDGET(focus) != GTK_WIDGET(feedlist))
            //if (focus != feedlist)
            {
                gtkhtml_selectall();
                ULOG_DEBUG("GKTHML selected all");
                  return;
            }
            else
            {
                ULOG_DEBUG ("on_copy: ***ERRORS****\nSomething must have gone wrong here.\
                            htmlwidget has no selection. Dialog is not focused. \
                            gtkeditable is empty \n \
                            Yet select all is SENSITIVE. Recheck the code \n ");                          
                 return;
                           
            }
        }

    }
    else 
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
             dgettext("hildon-common-strings", "ckct_ib_nothing_to_select"));

    return;

}

void on_menuitem_edit(GtkWidget * widget, gpointer user_data)
{
    AppUIData *app_ui_data = (AppUIData *) user_data;
    GtkWidget *focus = NULL;
    GtkClipboard *clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);

    //g_assert(NULL != user_data);
    g_return_if_fail (NULL != user_data);
    //g_assert(NULL != clipboard);
    g_return_if_fail (NULL != clipboard);
    /* first, pretend everything is not click-able */
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_cut), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_copy),
                 FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_paste),
                 FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_selectall),
                 FALSE);
                 
    if (NULL != app_ui_data->dialog)
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->dialog));
    else
        focus = gtk_window_get_focus(GTK_WINDOW(app_ui_data->app));

    if (GTK_IS_EDITABLE(focus)) {
        if (gtk_editable_get_selection_bounds
           (GTK_EDITABLE(focus), NULL, NULL)) {
                   gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_cut),
                      TRUE);
                   gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_copy), TRUE);
           }
       if (gtk_clipboard_wait_is_text_available(clipboard)) {
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_paste), TRUE);
       }
       gchar * tmp = gtk_editable_get_chars(GTK_EDITABLE(focus), 0, -1);
       if (strlen(tmp) != 0) //there is something to be select-all'ed
          gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_selectall),
                 TRUE);
       g_free(tmp);
    }

    /* htmlwidget checking... */
    else {
        ULOG_DEBUG("on_menuitem_edit: gtkeditable is not the focus");
        //if ((focus != feedlist) &&
        if ( (GTK_WIDGET(focus) != GTK_WIDGET(feedlist)) &&
           htmlwidget_has_selection (app_data->app_ui_data->html_scrollpane))
        {
            ULOG_DEBUG("on_menuitem_edit: htmlwidget has selection.");
            //can set to COPY sensitive.
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_copy), TRUE);
          
        }
        //TODO: This will be safe.        
        //gboolean has_focus;
        //g_object_get(G_OBJECT(app_ui_data->htmlwidget), "focus", &has_focus, NULL);

        //not gtkeditable, not feedlist, it should be gtkhtml

        //if ( focus != feedlist)
        if ( GTK_WIDGET(focus) != GTK_WIDGET(feedlist))
            gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_selectall),
                                    TRUE);
    }

}

/* Insensitiveize feed directory if neeeded */
void on_menuitem_tools(GtkWidget * widget, gpointer user_data)
{

    gchar *feed_dir_source = getStringConfValue(FEED_DIRECTORY_SOURCE);
    if (strlen(feed_dir_source) == 0 )
    {
        ULOG_DEBUG("on_menuitem_tools: setting feed dir to insensitive");
        gtk_widget_set_sensitive(GTK_WIDGET
                     (app_data->app_ui_data->menuitem_feed_directory), FALSE);
    }
    else if (app_data->app_ui_data->search_mode == SFM_NORMAL)
        gtk_widget_set_sensitive(GTK_WIDGET
                     (app_data->app_ui_data->menuitem_feed_directory), TRUE);

    g_free(feed_dir_source);
    return;
}
void toggle_find_on_page(GtkWidget *widget, gpointer user_data)
{
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL); 

    setBooleanConfValue(ENABLE_FIND_ON_PAGE, GTK_CHECK_MENU_ITEM(widget)->active);
    
    if(GTK_CHECK_MENU_ITEM(widget)->active)
        gtk_widget_show_all(GTK_WIDGET(app_data->app_ui_data->find_toolbar));
    else
        gtk_widget_hide_all(GTK_WIDGET(app_data->app_ui_data->find_toolbar));
}

void free_find_string()
{
    AppUIData *app_ui_data = NULL;
    
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    
    if(app_ui_data->find_string != NULL &&
       strcmp(app_ui_data->find_string,""))
    g_free(app_ui_data->find_string);

    app_ui_data->find_string = "";
}

void find_toolbar_search(GtkWidget * widget, gpointer data)
{
    gchar *find_text = NULL;
    gchar *infotext = NULL;
    
    ULOG_DEBUG("find_toolbar_search");

    //g_assert(widget != NULL);
        g_return_if_fail (NULL != widget);
    g_object_get(G_OBJECT(widget), "prefix", &find_text, NULL);
    ULOG_DEBUG("prefix: %s", find_text);

    if (( NULL == find_text ) || (strcmp (find_text, "") == 0) )
    {    
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_enter_text_to_be_found"));
        return;    
    }    

    gboolean rewind = FALSE;
    gboolean matched = FALSE;
    if (prev_find_text == NULL)
    {
        prev_find_text = g_strdup(find_text);
        prev_text_has_a_match = FALSE;
        rewind = TRUE; //start from beginning if starting a new search
    }
    else if (strcmp(find_text,prev_find_text) != 0)
    {
        if ( NULL != prev_find_text)
            g_free(prev_find_text);
        prev_find_text = g_strdup(find_text);
        prev_text_has_a_match = FALSE;
        rewind = TRUE; //start from beginning if starting a new search
    }
    else if ((strcmp(find_text,prev_find_text) == 0) && (prev_text_has_a_match == FALSE))
        rewind = TRUE;
 

    gtkhtml_find_cb( GTK_WIDGET(app_data->app_ui_data ->html_scrollpane),
        find_text,rewind, &matched );

    if (matched)
        prev_text_has_a_match = TRUE;

    if (matched == FALSE)
    {
        if ( rewind == TRUE )
        { 
            infotext =  g_strdup_printf ( _("ecoc_ib_edwin_cannot_find" ),find_text);
            gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), infotext);
            g_free(infotext);
        }
        else 
            gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_search_completed"));

    }
    g_free(find_text);
    find_text = NULL;

}

void find_toolbar_close(GtkWidget * widget, gpointer data)
{
    ULOG_DEBUG("find toolbar close");
    AppUIData * app_ui_data = NULL;
     
    //g_assert(app_data != NULL);
     
    app_ui_data = app_data->app_ui_data;

    //g_assert(app_ui_data != NULL);
    
    if(app_ui_data->find_toolbar == NULL)
        return;
    
    GTK_CHECK_MENU_ITEM(app_ui_data->menuitem_find_on_page)->active=0;

    setBooleanConfValue(ENABLE_FIND_ON_PAGE, FALSE);
    
    gtk_widget_hide_all(GTK_WIDGET(app_ui_data->find_toolbar));
}

/*------------------------------------------------------------------------------*/
/* status bar callback, error box function, GUI update function            */
/*------------------------------------------------------------------------------*/

void ui_update(void)
{
    if (!lifereaStarted)
    return;
    while (gtk_events_pending())
        gtk_main_iteration();
}

void ui_show_error_box(const char *format, ...)
{
    GtkWidget *dialog = NULL;
    va_list args;
    gchar *msg = NULL;

    if(format == NULL)
        return;

    va_start(args, format);
    msg = g_strdup_vprintf(format, args);
    va_end(args);

    dialog = gtk_message_dialog_new(GTK_WINDOW(mainwindow),
                    GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_ERROR,
                    GTK_BUTTONS_CLOSE, "%s", msg);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
    g_free(msg);
}

void ui_show_info_box(const char *format, ...)
{
    GtkWidget *dialog = NULL;
    va_list args;
    gchar *msg = NULL;

    if(format == NULL)
        return;
    
    va_start(args, format);
    msg = g_strdup_vprintf(format, args);
    va_end(args);

    dialog = gtk_message_dialog_new(GTK_WINDOW(mainwindow),
                    GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_INFO,
                    GTK_BUTTONS_CLOSE, "%s", msg);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
    g_free(msg);
}

/*------------------------------------------------------------------------------*/
/* exit handler                                    */
/*------------------------------------------------------------------------------*/

void save_state(AppData *app_data)
{
    AppState *app_state = NULL;

    //g_assert(app_data != NULL);
    
    reset_state_file(app_data->osso);
    
    app_state = create_app_state();
    construct_app_state(app_state);
    reader_save_state(app_data, app_state);
    destroy_app_state(app_state);
    app_state = NULL;
}

void restore_state(AppData *app_data)
{
    AppState *app_state = NULL;
    StateSaveResultCode stres = ST_SUCCESS;

    //g_assert(app_data != NULL);

    app_state = create_app_state();
    stres = reader_load_state(app_data, app_state);
    
    ULOG_DEBUG("State save result code: %d", stres);
    if (stres == ST_SUCCESS) {
        ULOG_DEBUG("Updating app state");
    update_app_state(app_state);
    }
    destroy_app_state(app_state);
    app_state = NULL;
}

void on_untop(GtkWidget * widget, gpointer user_data)
{
    //AppData *app_data = NULL;

    //app_data = (AppData *) user_data;

    ULOG_DEBUG("**************************on_untop**************************");

    //g_assert(app_data != NULL);
    
    //g_assert(app_data->app_ui_data != NULL);

    app_data->app_ui_data->application_topped = FALSE;

    hildon_app_set_killable(app_data->app_ui_data->app, TRUE);

}

void abort_quit();

void on_top(GtkWidget * widget, gpointer user_data)
{   
    //AppData *app_data = NULL;

    //app_data = (AppData *) user_data;

    ULOG_DEBUG("************************on_top*****************************");
      
    //g_assert(app_data != NULL);

    abort_quit();
    
    ui_mainwindow_empty_search_field();
                
    //g_assert(app_data->app_ui_data != NULL);
    
    app_data->app_ui_data->start_on_background = FALSE;
    app_data->app_ui_data->application_topped = TRUE;
    ULOG_DEBUG("on_top: showing window");
    /* Main-window is hidden if we were started for refresh in background */
    //gdk_threads_enter();
    gtk_widget_show(GTK_WIDGET(app_data->app_ui_data->app));
    //gdk_threads_leave();
    ULOG_DEBUG("on_top: showed window");    
    hildon_app_set_killable(app_data->app_ui_data->app, FALSE);
}

gboolean on_key_press(GtkWidget * widget, GdkEventKey * event,
              gpointer user_data)
{
    AppData *app_data = (AppData *) user_data;

    switch (event->keyval) {
    case GDK_F3:        /* For debugging only. */
    ULOG_DEBUG("emulate task away\n");
    on_untop(NULL, app_data);
    return TRUE;
    default:
    break;
    }
    return FALSE;
}

gboolean key_press_for_cancelling_dialog(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
    (void) data;

    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);

    if(event == NULL)
        return FALSE;

    switch (event->keyval) {
        case GDK_Escape:
            app_data->app_ui_data->dialog_closed = TRUE;
            return FALSE;
        default:
            return FALSE;
    }
    return FALSE;
}

/* Update the toolbar and menu*/
void    ui_on_search_finished(void)
{
    
    AppUIData *app_ui_data;
     
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    //g_assert(app_ui_data != NULL);
    
    //g_assert(NULL != app_ui_data->button_image_search);
         g_return_if_fail (NULL != app_ui_data->button_image_search);
    gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON
                    (app_ui_data->search_button),
                    app_ui_data->button_image_search);
    gtk_container_remove(GTK_CONTAINER(app_ui_data->var_field),
             app_ui_data->progress_bar);
    gtk_container_add(GTK_CONTAINER(app_ui_data->var_field),
              app_ui_data->search_field);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->refresh_button),
                 TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_ui_data->menuitem_refresh_feeds), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_add), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->new_feed_button),
                TRUE);            
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_unsubscribe),
                TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->unsubscribe_button),
                TRUE);                                                     
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_properties), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->feed_properties_button),
                TRUE);            
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_importopml), TRUE);                
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_manage_folders),
                TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->manage_folders_button),
                TRUE);                    
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_feed_directory), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_ui_data->menuitem_search_results), TRUE);                
                                 
    //g_assert(SFM_SEARCH == app_ui_data->search_mode
    //     || SFM_INTERRUPTED == app_ui_data->search_mode);
    g_return_if_fail (SFM_SEARCH == app_ui_data->search_mode
         || SFM_INTERRUPTED == app_ui_data->search_mode);
    if (SFM_SEARCH == app_ui_data->search_mode) {
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_search_completed"));
    } else if (SFM_INTERRUPTED == app_ui_data->search_mode) {
       gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_search_interrupted"));
    }
    
    app_ui_data->search_mode = SFM_NORMAL;
    app_ui_data->progressbar_mode = PROGRESSBAR_NORMAL_MODE;
    
    ui_mainwindow_empty_search_field();
    //update toolbar to be hidden or not. Because user might have switched 
    //to different Fullscreen mode, and we wanted to show them the progress
    //bar even if it's in toolbar-hidden mode.
    hide_progressbar_if_needed();


}
/* handle for users pressing stop while searching
*/
void stop_search()
{
    ULOG_DEBUG("vvvvvvvvvvvvv....stop_search.....vvvvvvvvvvv......");
    //this is enough to stop the vfolder_apply_rules() for one feed
    app_data->app_ui_data->search_mode = SFM_INTERRUPTED;
    //Now we need to stop the html page too. TODO
    gtkhtml_finish_search_feed();
    ui_on_search_finished();
    return;    
}

/* ui_on_search_start: update toolbar and menuitem stuff
*/
void ui_on_search_started(void)
{

    AppUIData *app_ui_data;
     
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    //g_assert(app_ui_data != NULL);
    
    //g_assert(NULL != app_ui_data->search_button);
    //g_assert(NULL != app_ui_data->button_image_stop);

    gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON
                    (app_ui_data->search_button),
                    app_ui_data->button_image_stop);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(app_ui_data->progress_bar),
                  _("rss_ia_searching"));
    ULOG_DEBUG("\n\n\n\n\n on_search_entry_activate: Setting fraction\n\n\n\n\n");
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR
                  (app_ui_data->progress_bar), 0);
    gtk_container_remove(GTK_CONTAINER(app_ui_data->var_field),
             app_ui_data->search_field);
    gtk_container_add(GTK_CONTAINER(app_ui_data->var_field),
              app_ui_data->progress_bar);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->refresh_button),
                 FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_ui_data->menuitem_refresh_feeds), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET
                 (app_ui_data->menuitem_search_results), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_add), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->new_feed_button),
                FALSE);            
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_unsubscribe),
                FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->unsubscribe_button),
                FALSE);                                                     
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_properties), FALSE);
    ///
    //gtk_widget_set_sensitive(GTK_WIDGET
    //            (app_ui_data->menuitem_search_results), FALSE);
    //
    // No adding feed while searching 
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_feed_directory), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET
                (app_ui_data->menuitem_importopml), FALSE);

    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->feed_properties_button),
                FALSE);            
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_manage_folders),
                FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->manage_folders_button),
                FALSE);                                                     

    app_ui_data->search_mode = SFM_SEARCH;
    app_ui_data->progressbar_mode = PROGRESSBAR_SEARCH_MODE;
}
/** Find on page
  *
  * @param iter the search is done forward from the given iter
  * @param find_text the text to be found
  * @param textbuffer the textbuffer to look for the given string
  * @param new_search TRUE if a new search is being conducted, FALSE
  *        if the search string or textbuffer has not been changed, which means
  *        that the user is still looking for more instances of the same string

  * tvh: Not used by gtkhtml. So commented out
  */
/*
static void find(GtkTextIter iter, gchar *find_text, GtkTextBuffer *textbuffer, gboolean new_search)
{

    gboolean found = FALSE;
    GtkTextIter match_start, match_end;
    GtkTextIter start_iter, end_iter, find_iter, lowercase_match_start, lowercase_match_end;
    GtkTextMark *tmp = NULL;
    GtkTextBuffer *lowercase_textbuffer = NULL;
    AppUIData *app_ui_data = NULL;
    int start = 0;
    int lowercase_start = 0;
    int lowercase_end = 0;   
    gchar *slice = NULL;
    gchar *lowercase_slice = NULL;
    gchar *lfind_text = NULL;

    g_assert(app_data != NULL);
    g_assert(app_data->app_ui_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    
    / The following piece of code is needed for implementing a case insensitive search.
       It would be MUCH easier, if gtk_text_iter_forward_search would implement this,
       but until that happens, this ugly solution is needed /
    start = gtk_text_iter_get_offset(&iter);
    gtk_text_buffer_get_start_iter(textbuffer, &start_iter);
    gtk_text_buffer_get_end_iter(textbuffer, &end_iter);
    slice = gtk_text_buffer_get_slice(textbuffer, &start_iter, &end_iter, TRUE);
    lowercase_slice = g_utf8_strdown(slice, -1);
    lfind_text = g_utf8_strdown(find_text, -1);

    lowercase_textbuffer = gtk_text_buffer_new(gtk_text_buffer_get_tag_table(textbuffer));
    gtk_text_buffer_set_text(lowercase_textbuffer, lowercase_slice, -1);

    gtk_text_buffer_get_iter_at_offset(lowercase_textbuffer, &find_iter, start);

    found = gtk_text_iter_forward_search(&find_iter, lfind_text,
                        GTK_TEXT_SEARCH_VISIBLE_ONLY,
                        &lowercase_match_start, &lowercase_match_end,
                        NULL);

    if(found)
    {
        lowercase_start = gtk_text_iter_get_offset(&lowercase_match_start);
        lowercase_end = gtk_text_iter_get_offset(&lowercase_match_end);

        gtk_text_buffer_get_iter_at_offset(textbuffer, &match_start, lowercase_start);
        gtk_text_buffer_get_iter_at_offset(textbuffer, &match_end, lowercase_end);
    }

    g_free(slice);
    g_free(lowercase_slice);
    g_free(lfind_text);
    / Case insensitive search ends /

    /found = gtk_text_iter_forward_search(&iter, find_text,
                        GTK_TEXT_SEARCH_VISIBLE_ONLY,
                        &match_start, &match_end,
                        NULL);/
        
    if(found)
    {
    / Cursor is moved to the end of match. /
    gtk_text_buffer_select_range(textbuffer,
                &match_end,
                &match_start);
    
    tmp = gtk_text_buffer_get_mark( GTK_TEXT_BUFFER( textbuffer ),
                    "insert" );
    / not used anymore .Gtkhtml_find is better. ui_htmlview_get_buffer therefor not needed
    gtk_text_view_scroll_to_mark( GTK_TEXT_VIEW(ui_htmlview_get_buffer()),
                tmp, 0.0f, FALSE, 0.5f, 0.5f );
    /
    } 
    else
    {
            if(new_search)
        {
            gtk_infoprintf(GTK_WINDOW(app_ui_data->app),
                     _("ecoc_ib_edwin_cannot_find"),
                 find_text);
                free_find_string();                
        }
        else
            gtk_infoprint(GTK_WINDOW(app_ui_data->app),
                      _("rss_ib_search_completed"));
    }
}
*/
