/**
 * @file html.c HTML file handling / feed auto discovery
 * 
 * Copyright (C) 2004 ahmed el-helw <ahmedre@cc.gatech.edu>
 * Copyright (C) 2004 Lars Lindner <lars.lindner@gmx.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include "support.h"
#include "callbacks.h"
#include "debug.h"
#include "html.h"

#include <osso-log.h>

/************************************************************************/
/*                        PRIVATE FUNCTIONS                             */
/************************************************************************/

static gchar *checkLinkRef(const gchar * str)
{
    gchar *res = NULL;
    gchar *tmp = NULL, *tmp2 = NULL;

    /*debug1(DEBUG_PARSING, "checking link %s", str); */
    tmp = strstr(str, "href=");
    if (NULL == tmp)
    tmp = strstr(str, "HREF=");
    if (NULL == tmp)
    return NULL;
    /* FIXME: single quotes support */
    tmp2 = strchr(tmp, '\"');
    if (NULL == tmp2)
    return NULL;
    tmp = strchr(tmp2 + 1, '\"');
    if (tmp == NULL)
        return NULL;
    *tmp = '\0';
    res = g_strdup(tmp2 + 1);
    *tmp = '\"';

    if ((strstr(str, "alternate") != NULL) &&
    ((strstr(str, "text/xml") != NULL) ||
     (strstr(str, "rss+xml") != NULL) ||
     (strstr(str, "rdf+xml") != NULL) ||
     (strstr(str, "atom+xml") != NULL)))
    return res;
    g_free(res);
    return NULL;
}

static gchar *checkLinkRefForFavicon(const gchar * str)
{
    gchar *res = NULL;
    gchar *tmp = NULL, *tmp2 = NULL;

    /*debug1(DEBUG_PARSING, "checking link %s", str); */
    tmp = strstr(str, "href=");
    if (NULL == tmp)
    tmp = strstr(str, "HREF=");
    if (NULL == tmp)
    return NULL;

    /* FIXME: single quotes support */
    tmp2 = strchr(tmp, '\"');
    if (NULL == tmp2)
    return NULL;

    tmp = strchr(tmp2 + 1, '\"');
    if (tmp == NULL)
        return NULL;
    *tmp = '\0';
    res = g_strdup(tmp2 + 1);
    *tmp = '\"';
    
    if ((strstr(str, "shortcut icon") != NULL ||
     strstr(str, "icon") != NULL))/* &&
    ((strstr(str, "image/x-icon") != NULL)))*/
    return res;
    g_free(res);

    return NULL;
}

static gchar *checkNormalLink(const gchar * str)
{
    gchar *res = NULL, *tmp = NULL, *tmp2 = NULL;

    debug1(DEBUG_PARSING, "checking link %s", str);
    tmp = strstr(str, "href=");
    if (NULL == tmp)
    tmp = strstr(str, "HREF=");
    if (NULL == tmp)
    return NULL;
    /* FIXME: single quotes support */
    tmp2 = strchr(tmp, '\"');
    if (NULL == tmp2)
    return NULL;
    tmp = strchr(tmp2 + 1, '\"');
    /* 28 June 2006 Added this check*/
    if (tmp == NULL)
        return NULL;
    *tmp = '\0';
    res = g_strdup(tmp2 + 1);
    *tmp = '\"';

    if ((strstr(res, "rdf")) || (strstr(res, "xml")) ||
    (strstr(res, "rss")))
    return res;
    g_free(res);
    return NULL;
}

static gchar *search_links(const gchar * data, int type, gboolean favicon)
{
    gchar *ptr = NULL;
    const gchar *tmp = data;
    gchar *result = NULL;
    gchar *res = NULL;
    gchar *tstr = NULL;
    gchar *endptr = NULL;

    while (1) {
    ptr = strstr(tmp, ((type == 0) ? "<link " : "<a "));
    if (NULL == ptr)
        ptr = strstr(tmp, ((type == 0) ? "<LINK " : "<A "));
    if (NULL == ptr)
        break;

    endptr = strchr(ptr, '>');
    if (endptr == NULL)
        return NULL;
    *endptr = '\0';
    tstr = g_strdup(ptr);
    *endptr = '>';
    ULOG_DEBUG("tstr: %s", tstr);
    
    res = ((type == 0) ?
           (favicon ? checkLinkRefForFavicon(tstr) :
        checkLinkRef(tstr))
           : checkNormalLink(tstr));
    g_free(tstr);
    if (res != NULL) {
        result = res;
        break;
    }
    tmp = endptr;
    }
    return result;
}

/************************************************************************/
/*                        PUBLIC FUNCTIONS                              */
/************************************************************************/

gchar *html_auto_discover_feed(const gchar * data, const gchar * baseUri)
{
    gchar *res = NULL, *tmp = NULL;

    debug0(DEBUG_UPDATE, "searching through link tags");
    res = search_links(data, 0, FALSE);
    debug1(DEBUG_UPDATE, "search result: %s", res ? res : "none found");
    if (res == NULL) {
    debug0(DEBUG_UPDATE, "searching through href tags");
    res = search_links(data, 1, FALSE);
    debug1(DEBUG_UPDATE, "search result: %s",
           res ? res : "none found");
    }

    if (res == NULL) {

    } else {
    /* turn relative URIs into absolute URIs */
    tmp = res;
    res = common_build_url(res, baseUri);
    g_free(tmp);
    }

    return res;
}

gchar *html_discover_favicon(const gchar * data, const gchar * baseUri)
{
    gchar *res = NULL, *tmp = NULL;

    debug0(DEBUG_UPDATE, "searching through link tags");
    res = search_links(data, 0, TRUE);
    debug1(DEBUG_UPDATE, "search result: %s", res ? res : "none found");

    if (res != NULL) {
    /* turn relative URIs into absolute URIs */
    tmp = res;
    res = common_build_url(res, baseUri);
    g_free(tmp);
    }

    return res;
}

void find_image_links(gchar *buff, HtmlImageCallback callback, gpointer user_data)
{
    gchar *p = NULL;
    gchar *n = NULL;
    gchar *imgptr = NULL;
    gchar *link = NULL;
    gchar *lp = NULL;
    gint size;
    gchar oldchar;

    g_return_if_fail(buff);
    while ((p = strcasestr(buff, "http")))
    {
        if (((strncmp(p+4, "://", 3) == 0) ||
            (strncasecmp(p+4, "s://", 4) == 0)) &&
            (n = strpbrk(p+7, "\"'")) &&
            (imgptr = memrchr(buff, '<', p - buff)))
        {
            oldchar = *p;
            *p = 0;
            imgptr = strcasestr(imgptr, "img");
            *p = oldchar;

            if (imgptr)
            {
                size = n-p;
                link = g_new(char, size+1);
                memcpy(link, p, size);
                link[size] = 0;

/*                if ((lp = memrchr(link, '.', size)) &&
                    (size - (lp - link) < 7)) //maximum extension length
                {*/
                    callback(link, user_data);
/*                }*/

                g_free(link);
            }
        }

        buff = p + 1;
    }
}
