/**
 * @file main.c Liferea main program
 *
 * Copyright (C) 2003 Lars Lindner <lars.lindner@gmx.net>
 * 
 * Some code like the command line handling was inspired by 
 *
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "main.h"

AppData *app_data;
extern GtkWidget *newsbox;
extern time_t rss_updated;

GThread *mainThread = NULL;
gboolean lifereaStarted = FALSE;

/************************************************************************/
/*                        PRIVATE FUNCTIONS                             */
/************************************************************************/

/** Shows the command line parameters supported by the program
  */
static void show_help(void)
{
    GString *str = g_string_new(NULL);

    g_string_append_c(str, '\n');
    g_string_append_printf(str, "Liferea %s\n\n", VERSION);
    g_string_append_printf(str, "%s\n",
               "  --version        Prints Liferea's version number");
    g_string_append_printf(str, "%s\n",
               "  --help           Prints this message and exits");
    g_string_append_printf(str, "%s\n",
               "  --iconify        Starts the program iconified");
    g_string_append_c(str, '\n');
    g_string_append_printf(str, "%s\n",
               "  --debug-cache    Print debugging messages for the cache handling");
    g_string_append_printf(str, "%s\n",
               "  --debug-conf     Print debugging messages of the configuration handling");
    g_string_append_printf(str, "%s\n",
               "  --debug-update   Print debugging messages of the feed update processing");
    g_string_append_printf(str, "%s\n",
               "  --debug-parsing  Print debugging messages of all parsing functions");
    g_string_append_printf(str, "%s\n",
               "  --debug-gui      Print debugging messages of all GUI functions");
    g_string_append_printf(str, "%s\n",
               "  --debug-trace    Print debugging messages when entering/leaving functions");
    g_string_append_printf(str, "%s\n",
               "  --debug-all      Print debugging messages of all types");
    g_string_append_printf(str, "%s\n",
               "  --debug-verbose  Print verbose debugging messages");

    g_string_append_c(str, '\n');
    ULOG_INFO("%s", str->str);
    g_string_free(str, TRUE);
}

/** Removes the lock file
  */
static void main_unlock()
{
    gchar *filename = NULL;

    filename =
    g_strdup_printf("%s" G_DIR_SEPARATOR_S "lock",
            common_get_cache_path());
    unlink(filename);
    g_free(filename);
}

static void initialize_app_ui_data()
{
    app_data->app_ui_data->start_on_background = FALSE;
    app_data->app_ui_data->find_string = "";
    app_data->app_ui_data->rss_current_download = 0;
    app_data->app_ui_data->rss_max_download = 0;
    app_data->app_ui_data->rss_search_found = 0;
    app_data->app_ui_data->automatic_update_source = 0;
    app_data->app_ui_data->download_dequeue_source = 0;
    //app_data->app_ui_data->iap_action = OSSO_IAP_NORMAL;
    app_data->app_ui_data->iap_action = OSSO_IAP_NO_ACTION;
    app_data->app_ui_data->errors_in_refreshing = FALSE;
    app_data->app_ui_data->add_feed_dialog = NULL;
    app_data->app_ui_data->ui_data = NULL;
    app_data->app_ui_data->source = NULL;
    app_data->app_ui_data->filter = NULL;
    app_data->app_ui_data->draggednode = NULL;
    app_data->app_ui_data->feed_directory_cancelled = FALSE;
    app_data->app_ui_data->network_disconnected = FALSE;
    app_data->app_ui_data->application_topped = TRUE;
    app_data->app_ui_data->refresh_dialog = NULL;
    app_data->app_ui_data->dialog_closed = FALSE;
    app_data->gconf_client = NULL;
    app_data->app_ui_data->search_mode = SFM_NORMAL;
    app_data->app_ui_data->html_displayed = TRUE;
    
}

/** Back to sigterm approach. Its not a good idea to do a state restore
  * every time the application is topped so a full state save is not done
  * on untopping. That would require clearing the item list. A part
  * of the state is saved on untopping the program and the rest is done here */
static void sigterm_cb(int signum)
{
    (void) signum;
    ULOG_DEBUG("SIGTERM............");
    //g_assert(app_data != NULL);
    g_return_if_fail(app_data!=NULL);
    save_state(app_data);

    on_quit(FALSE);
    ULOG_DEBUG("calling on_exit");
    exit(0);
}

/************************************************************************/
/*                        PUBLIC FUNCTIONS                              */
/************************************************************************/

int main(int argc, char *argv[])
{
    gulong debug_flags = 0;
    gboolean startIconified = FALSE;
    gboolean headline_clicked = FALSE;
    const char *arg = NULL;
    gchar *filename = NULL;
    gint i = 0;
    
    /* Allocate application data */
    app_data = g_new0(AppData, 1);
    app_data->app_ui_data = g_new0(AppUIData, 1);

    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);

    initialize_app_ui_data();
    
    ULOG_DEBUG("Application starting");

    signal( SIGTERM, sigterm_cb );

    init_osso(app_data);

    filename =
    g_strdup_printf("%s" G_DIR_SEPARATOR_S ".osso_rss_feed_reader"
            G_DIR_SEPARATOR_S "background", g_get_home_dir());
    if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
    startIconified = TRUE;
    app_data->app_ui_data->start_on_background = TRUE;
    g_unlink(filename);
    }
    g_free(filename);

    filename =
    g_strdup_printf("%s" G_DIR_SEPARATOR_S ".osso_rss_feed_reader"
            G_DIR_SEPARATOR_S "headline_clicked", g_get_home_dir());
    if (g_file_test(filename, G_FILE_TEST_EXISTS)) {
    headline_clicked = TRUE;
    g_unlink(filename);
    }
    g_free(filename);
    
#ifdef USE_SM
    gchar *opt_session_arg = NULL, *opt_config_dir_arg = NULL;
#endif

#ifdef ENABLE_NLS
    setlocale(LC_ALL, "");
    bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
    textdomain(GETTEXT_PACKAGE);
#endif

    /* parse arguments  */
    debug_flags = 0;
/*    debug_flags = DEBUG_TRACE|DEBUG_CACHE|DEBUG_CONF|DEBUG_UPDATE|DEBUG_PARSING|DEBUG_GUI; */
    for (i = 1; i < argc; ++i) {
    arg = argv[i];

    if (!strcmp(arg, "--debug-cache"))
        debug_flags |= DEBUG_CACHE;
    else if (!strcmp(arg, "--debug-conf"))
        debug_flags |= DEBUG_CONF;
    else if (!strcmp(arg, "--debug-update"))
        debug_flags |= DEBUG_UPDATE;
    else if (!strcmp(arg, "--debug-parsing"))
        debug_flags |= DEBUG_PARSING;
    else if (!strcmp(arg, "--debug-gui"))
        debug_flags |= DEBUG_GUI;
    else if (!strcmp(arg, "--debug-trace"))
        debug_flags |= DEBUG_TRACE;
    else if (!strcmp(arg, "--debug-all"))
        debug_flags |=
        DEBUG_TRACE | DEBUG_CACHE | DEBUG_CONF | DEBUG_UPDATE |
        DEBUG_PARSING | DEBUG_GUI;
    else if (!strcmp(arg, "--debug-verbose"))
        debug_flags |= DEBUG_VERBOSE;
    else if (!strcmp(arg, "--version") || !strcmp(arg, "-v")) {
        ULOG_INFO("liferea %s\n", VERSION);
        return 0;
    } else if (!strcmp(arg, "--help") || !strcmp(arg, "-h")) {
        show_help();
        return 0;
    } else if (!strcmp(arg, "--iconify")) {
        startIconified = TRUE;
#ifdef USE_SM
    } else if (!strcmp(arg, "--session")) {
        i++;
        if (i < argc) {
        opt_session_arg = g_strdup(argv[i]);
        } else
        ULOG_DEBUG
            ("The --session argument must be given a parameter.\n");
#endif
    } else {
        ULOG_DEBUG("Liferea encountered an unknown argument: %s\n",
               arg);
    }
    }
    set_debug_level(debug_flags);

/* "the program name is set in gdk_init(), which is called by gtk_init()
 * "Note that for thread-safety reasons this function can only be called once.
 *    g_set_prgname("liferea"); */
    gtk_set_locale();
    /* threads might have been init'ed by other apps like dbus, gnomevfs..*/
    //g_thread_init(NULL);
    /* Use this safety patch from bug 25738 */
    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }

    gdk_threads_init();
    gtk_init(&argc, &argv);

    mainThread = g_thread_self();    /* we need to now this when locking in ui_queue.c */

    /* order is important! */
    conf_init();        /* initialize gconf */
#ifdef USE_SM
    session_init(BIN_DIR G_DIR_SEPARATOR_S "liferea", opt_session_arg);
    session_set_cmd(NULL, startIconified);
#endif
/* We don't use plugin architecture because of performance concerns */

    download_init();        /* Initialize the download subsystem */

    metadata_init();

    feed_init();        /* register feed types */

    general_load();        /* retrieve stored general data */

    /* Create application UI */
    ui_create_main_window(app_data);

    ui_mainwindow_load_search_entries();
            
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    ui_init(startIconified, app_data->app_ui_data);    /* setup mainwindow and initialize gconf configured GUI behaviour */
    ULOG_DEBUG("tvh main.c: ui_init done ");
    /* tvh: commented this out. This registers 'windows-1252 encoding handler
    This hangs with feeds http://gravyboy.com/fanboyalmanac/FanboyAlmanacFeed.xml
     */
    //register_encoding_handlers();
    ULOG_DEBUG("tvh main.c: encode handler registered ");

    //g_assert(app_data != NULL);
    ULOG_DEBUG("tvh main.c: app is not NULL ");

    /* We were not started by D-BUS refresh request, GConf has currently
     * active IAP, last refresh was over 15 minutes ago, no search or
     * refresh is going on right now and automatic updates => start refresh.
     */
     
    if (!startIconified &&
    (time(NULL) - rss_updated) > 15 * 60 &&
    SFM_NORMAL == app_data->app_ui_data->search_mode &&
    getBooleanConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES)) 
    {
#ifdef WITHOUT_CONNECTIVITY
        ui_feedlist_do_for_all(NULL, ACTION_FILTER_FEED,
                       on_refreshbtn_clicked_cb);
#else   
        gchar *icd_gconf_current = getStringConfValue(ICD_GCONF_CURRENT);

        //if(*icd_gconf_current)
        if (strlen(icd_gconf_current) > 0)
        ui_feedlist_do_for_all(NULL, ACTION_FILTER_FEED,
                       on_refreshbtn_clicked_cb);
        g_free(icd_gconf_current);                   
#endif                   
    }
    

    if(!headline_clicked)
    {
        /* NOTE: Spec changed 2006June05: No default feed directory used
         * So the whole of start first time is not relavent any more. Thus commented out
         */
        /*
        gchar *start_first_time = getStringConfValue(START_FIRST_TIME);
        // Program run for the first time 
        if( !startIconified && 
               ((strlen(start_first_time) == 0 ) || 
                strcmp(start_first_time,COMPILE_TIME)) )
        {
            ULOG_DEBUG("\n\n\n\n\n**************\nmain: program runs for the first time\
                            \n**************\n\n\n\n\n\n");
            setStringConfValue(START_FIRST_TIME,COMPILE_TIME);
            on_menuitem_feed_directory_clicked(NULL, NULL);
        }
        g_free(start_first_time);
        */
        if(!startIconified)
        {
            ui_feedlist_delete_read_items(NULL);    
            /* Refresh the applet since the feed contents might have been changed */    
            rss_dbus_refresh_completed();
            restore_state(app_data);
        }
    }

    ui_folder_handle_empty_root();
    ULOG_DEBUG("tvh main.c: ui_folder_handle empty root done ");

    gdk_threads_enter();
    lifereaStarted = TRUE;
    ULOG_DEBUG("\n\n\n****\nGtk_main now \n****\n\n");
    gtk_main();
    gdk_threads_leave();
    
    ULOG_DEBUG("Clean exit");
    
    return 0;
}

/* Function to keep the "new" items, and delete the old items*/
void filter_on_exit(feedPtr fp)
{
    ui_feed_set_new_items_hidden(fp);
    ui_feed_delete_old_items(fp);
    ui_feed_remove_hidden_attribute(fp);
    feed_save(fp);
}

void save_application_data()
{
    conf_feedlist_save();    /* should save feedlist and folder states */
    
    /*Delete the old items, keep the news*/
    ui_feedlist_do_for_all(NULL,
               ACTION_FILTER_FEED | ACTION_FILTER_DIRECTORY,
               filter_on_exit);

    /* should save all feeds still in memory */
    ui_feedlist_do_for_all(NULL,
               ACTION_FILTER_FEED | ACTION_FILTER_DIRECTORY,
               feed_unload);

    /* save itemlist properties */
    setBooleanConfValue(LAST_ITEMLIST_MODE, !itemlist_mode);
    
    /*save gutter size in general , together with rss_unread and such */
    general_save();

}

/* timed out cb for on_quit
    on_quit needs this timedout as it needs to make sure gtkhtml is not currently
    active, or any search/refreshing...is active....
*/
static
gboolean on_quit_timedout_cb(gpointer data)
{
    // wait until things are settled down for the gtkhtml.
    // perhaps html_displayed is not needed 
    //if ((app_data->app_ui_data->search_mode != SFM_NORMAL) || !app_data->app_ui_data->html_displayed) 
    if (app_data->app_ui_data->search_mode != SFM_NORMAL)
    {
        ULOG_DEBUG("on_quit_timedout_cb: INTO ANOTHER ROUND");
        if (app_data->app_ui_data->search_mode != SFM_NORMAL) 
            ULOG_DEBUG("not SFM_NORMAL");
        else if (!app_data->app_ui_data->html_displayed) 
            ULOG_DEBUG("not html_displayed");
        //return TRUE;
    }
    ULOG_DEBUG("on_quit_timedout_cb: really exit now.");
    gboolean * reset_state = (gboolean * )data;

    //this free is probably not needed...
    free_icons();
    free_find_string();

    ui_mainwindow_save_search_entries();
    
    
    ui_itemlist_clear();    /* to free all item list related data */
    
    save_application_data();

    //download_deinit();
    
    if(reset_state)
        reset_state_file(app_data->osso);
    //try to return right here without waiting for anything like download or such
    //we have saved data, so why we need any other complicated stuff
    //not even download_deinit()
    //return FALSE;
    /* Hildon should kill those
    if(app_data != NULL)
        if(app_data->app_ui_data != NULL) {
        g_object_unref(app_data->app_ui_data->button_image_search);
            g_object_unref(app_data->app_ui_data->button_image_stop);
        g_object_unref(app_data->app_ui_data->search_field);
        g_object_unref(app_data->app_ui_data->button_image_refresh);
        g_object_unref(app_data->app_ui_data->progress_bar);
    }
    */    
    //free the image cache
    if(app_data->app_ui_data != NULL)
        g_object_unref(G_OBJECT(app_data->app_ui_data->img_cache));

    ULOG_DEBUG("on_quit_timedout_cb: destroying main window now");

    gtk_widget_destroy(mainwindow);
        ULOG_DEBUG("on_quit_timedout_cb: destroying main window now 2");
    //g_assert(app_data != NULL);
    ULOG_DEBUG("on_quit_timeout_cb: 1");
    gtk_main_quit();    
    ULOG_DEBUG("on_quit_timeout_cb: 2");
     
    deinit_osso(app_data);
    ULOG_DEBUG("on_quit_timeout_cb: 3");

    return FALSE;
    /* This should be done by HIldon lib?
    if(app_data != NULL)
    {
        if(app_data->app_ui_data != NULL)
        {
          g_free(app_data->app_ui_data);
      app_data->app_ui_data = NULL;
        }
    g_free(app_data);
    app_data = NULL;
    }
     */
//    main_unlock();

}

/* On the X close button clicked. Or process killed 
*/
/* Somehow refrshing from applet sometimes calls this func more than once
 * a dirty hack to have 'count' here
 * TODO: fix it properly. why called more than once???
 
 * This variable is also appropriate as the X button _can_ be clicked many times
 */
gboolean on_quit_cb(gboolean reset_state)
{
    ULOG_DEBUG("on_quit");

    // stop displaying or searching first 
    if ((app_data->app_ui_data->search_mode != SFM_NORMAL) || !app_data->app_ui_data->html_displayed) 
    {
        ULOG_DEBUG("\n****\n Exit while still not in SFM_NORMAL mode or page not yet finished displaying\
         This suckssss. \n\n*****\nFIX IT \n****\n");
        // stop the gtkhtml now
        if (app_data->app_ui_data->search_mode == SFM_DOWNLOAD)
        {
            handle_download_stopped_by_user();
            gtkhtml_stop();
            //2006June18 here we should signal applet.
            rss_dbus_refresh_completed();
        }
        else if (app_data->app_ui_data->search_mode == SFM_SEARCH)
           stop_search();

        else 
            gtkhtml_stop();
        // add a timeout here and then make sure things are quit        
        //g_timeout_add(100, on_quit_timedout_cb, &reset_state);        

    }
    //2006June18: just call it. 
    // just call it right away. It's safe now
    //else 
        on_quit_timedout_cb(&reset_state);
    // moved this to a timeout
    //tell on_close that it will be handled by main loop. thus return TRUE
    return FALSE;
}

static gint quit_count = 0;
static gint quit_source = 0;

gboolean on_quit(gboolean reset_state)
{
    ULOG_DEBUG("@@@ On quit called @@@");
    if (quit_count == 0)
    {
        quit_count++;
        quit_source = g_idle_add((GSourceFunc) on_quit_cb, (gpointer) reset_state);
        ULOG_DEBUG("@@@ Quit source: %d @@@", quit_source);
    }
    
    return TRUE;
}

void abort_quit()
{
    ULOG_DEBUG("@@@ Check to aborting quit (%d) @@@", quit_source);
    if (quit_source) {
	ULOG_DEBUG("@@@ Aborting quit @@@");
        g_source_remove(quit_source);
        quit_count = 0;
        quit_source = 0;
    }
}
