/**
 * @file ui_feed.c    UI actions concerning a single feed
 *
 * Copyright (C) 2004 Lars Lindner <lars.lindner@gmx.net>
 * Copyright (C) 2004 Nathan J. Conrad <t98502@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <libxml/uri.h>
#include <hildon-widgets/gtk-infoprint.h>
#include <strings.h>
#include "debug.h"
#include "support.h"
#include "feed.h"
#include "conf.h"
#include "callbacks.h"
#include "update.h"
#include "htmlview.h"
#include "interface.h"
#include "ui_feed.h"
#include "ui_queue.h"
#include "favicon.h"
#include "ui_notification.h"
#include "ui_feed_directory.h"

#include <osso-log.h>

extern guint rss_max_feed;
extern AppData *app_data;


static void on_newdialog_response(GtkDialog * dialog, gint response_id,
                  gpointer user_data);
static void on_importopmldialog_response(GtkDialog * dialog, gint response_id,
                  gpointer user_data);
static void on_propdialog_response(GtkDialog * dialog, gint response_id,
                   gpointer user_data);
/************************************************************************/
/*                        PRIVATE FUNCTIONS                             */
/************************************************************************/                  
                  
/**
 * Creates a new error description according to the passed
 * HTTP status and the feeds parser errors. If the HTTP
 * status is a success status and no parser errors occured
 * no error messages is created. The created error message 
 * can be queried with feed_get_error_description().
 *
 * @param fp        feed
 * @param httpstatus    HTTP status
 * @param resultcode the update code's return code (see update.h)
 */
static void ui_feed_set_error_description(feedPtr fp, gint httpstatus,
                      gint resultcode)
{
    gchar *tmp1 = NULL, *tmp2 = NULL, *buffer = NULL;
    gboolean errorFound = FALSE;

    //g_assert(NULL != fp);
    g_return_if_fail(NULL != fp);
    g_free(fp->errorDescription);
    fp->errorDescription = NULL;

    if (((httpstatus >= 200) && (httpstatus < 400)) &&    /* HTTP codes starting with 2 and 3 mean no error */
    (NULL == fp->parseErrors))
    return;

    if ((200 != httpstatus) || (resultcode != NET_ERR_OK)) {
    /* first specific codes */
    switch (httpstatus) {
    case 401:
        tmp2 =
        g_strdup
        ("You are unauthorized to download this feed. Please update your username and "
         "password in the feed properties dialog box.");
        break;
    case 402:
        tmp2 = g_strdup("Payment Required");
        break;
    case 403:
        tmp2 = g_strdup("Access Forbidden");
        break;
    case 404:
        tmp2 = g_strdup("Resource Not Found");
        break;
    case 405:
        tmp2 = g_strdup("Method Not Allowed");
        break;
    case 406:
        tmp2 = g_strdup("Not Acceptable");
        break;
    case 407:
        tmp2 = g_strdup("Proxy Authentication Required");
        break;
    case 408:
        tmp2 = g_strdup("Request Time-Out");
        break;
    case 410:
        tmp2 =
        g_strdup
        ("Gone. Resource doesn't exist. Please unsubscribe!");
        break;
    }
    /* Then, netio errors */
    if (tmp2 == NULL) {
        switch (resultcode) {
        case NET_ERR_URL_INVALID:
        tmp2 = g_strdup("URL is invalid");
        break;
        case NET_ERR_UNKNOWN:
        case NET_ERR_CONN_FAILED:
        case NET_ERR_SOCK_ERR:
        tmp2 = g_strdup("Error connecting to remote host");
        break;
        case NET_ERR_HOST_NOT_FOUND:
        tmp2 = g_strdup("Hostname could not be found");
        break;
        case NET_ERR_CONN_REFUSED:
        tmp2 =
            g_strdup
            ("Network connection was refused by the remote host");
        break;
        case NET_ERR_TIMEOUT:
        tmp2 = g_strdup("Remote host did not finish sending data");
        break;
        /* Transfer errors */
        case NET_ERR_REDIRECT_COUNT_ERR:
        tmp2 =
            g_strdup("Too many HTTP redirects were encountered");
        break;
        case NET_ERR_REDIRECT_ERR:
        case NET_ERR_HTTP_PROTO_ERR:
        case NET_ERR_GZIP_ERR:
        tmp2 = g_strdup("Remote host sent an invalid response");
        break;
        /* These are handled above      
           case NET_ERR_HTTP_410:
           case NET_ERR_HTTP_404:
           case NET_ERR_HTTP_NON_200:
         */
        case NET_ERR_AUTH_FAILED:
        case NET_ERR_AUTH_NO_AUTHINFO:
        tmp2 = g_strdup("Authentication failed");
        break;
        case NET_ERR_AUTH_GEN_AUTH_ERR:
        case NET_ERR_AUTH_UNSUPPORTED:
        tmp2 =
            g_strdup
            ("Webserver's authentication method incompatible with Liferea");
        break;
        }
    }
    /* And generic messages in the unlikely event that the above didn't work */
    if (NULL == tmp2) {
        switch (httpstatus / 100) {
        case 3:
        tmp2 =
            g_strdup
            ("Feed not available: Server requested unsupported redirection!");
        break;
        case 4:
        tmp2 = g_strdup("Client Error");
        break;
        case 5:
        tmp2 = g_strdup("Server Error");
        break;
        default:
        tmp2 = g_strdup("(unknown networking error happened)");
        break;
        }
    }
    errorFound = TRUE;
    tmp1 = g_strdup_printf(HTTP_ERROR_TEXT, httpstatus, tmp2);
    addToHTMLBuffer(&buffer, tmp1);
    g_free(tmp1);
    g_free(tmp2);
    }

    /* add parsing error messages */
    if (NULL != fp->parseErrors) {
    errorFound = TRUE;
    }

    /* if none of the above error descriptions matched... */
    if (!errorFound) {
    }

    fp->errorDescription = buffer;
}

/** Decodes source
  *
  * @param ui_data pointer to an ui data structure
  * @return the decoded source
  */
static gchar *ui_feed_dialog_decode_source(struct fp_prop_ui_data *ui_data)
{
    gchar *source = NULL;

    gchar *str = NULL, *tmp2 = NULL;
    /* First, strip leading and trailing whitespace */
    str =
    g_strstrip(g_strdup
           (gtk_entry_get_text(GTK_ENTRY(ui_data->sourceEntry))));

    /* Add http:// if needed */
    if (strstr(str, "://") == NULL) {
    tmp2 = g_strdup_printf("http://%s", str);
    g_free(str);
    str = tmp2;
    }

    /* Add trailing / if needed */
    if (strstr(strstr(str, "://") + 3, "/") == NULL) {
    tmp2 = g_strdup_printf("%s/", str);
    g_free(str);
    str = tmp2;
    }

    /* Use the values in the textboxes if also specified in the URL! */
    if ((NULL != ui_data->authcheckbox) &&
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
                     (ui_data->authcheckbox))) {
    xmlURIPtr uri = xmlParseURI(BAD_CAST str);
    if (uri != NULL) {
        xmlChar *sourceUrl;
        xmlFree(uri->user);
        uri->user = g_strdup_printf("%s:%s",
                    gtk_entry_get_text(GTK_ENTRY
                               (ui_data->
                                username)),
                    gtk_entry_get_text(GTK_ENTRY
                               (ui_data->
                                password)));
        sourceUrl = xmlSaveUri(uri);
        source = g_strdup(BAD_CAST sourceUrl);
        g_free(uri->user);
        uri->user = NULL;
        xmlFree(sourceUrl);
        xmlFreeURI(uri);
    } else
        source = g_strdup(str);
    } else {
    source = g_strdup(str);
    }
    g_free(str);

    return source;
}

/** After closing the add new feed dialog this function is called
  *
  * @param dialog the new feed dialog
  * @param response_id the id sent by hitting one of the dialog buttons
  * @param user_data pointer to an ui data structure
  */
static void on_newdialog_response(GtkDialog * dialog, gint response_id,
                  gpointer user_data)
{
    struct fp_prop_ui_data *ui_data = (struct fp_prop_ui_data *) user_data;
    AppUIData *app_ui_data;
    
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    gchar *source = NULL;

    //g_assert(app_ui_data != NULL);
    
    if(app_ui_data->source != NULL)
    {
    g_free(app_ui_data->source);
    app_ui_data->source = NULL;
    }
        
    if(app_ui_data->filter != NULL)
    {
        g_free((gchar *)app_ui_data->filter);
        app_ui_data->filter = NULL;
    }
    
    switch (response_id) {
    case GTK_RESPONSE_OK:
    source = ui_feed_dialog_decode_source(ui_data);

    ULOG_DEBUG("Trying to add source:\n%s", source);
    
    if(strcmp(source, "") == 0) {
        ULOG_DEBUG("Error case 1");
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_no_web_address_for_feed"));
        g_free(source);
        source = NULL;
    } else if (strcasecmp(source, "http:///") == 0) {
        ULOG_DEBUG("Error case 2");
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_no_web_address_for_feed"));
        g_free(source);
        source = NULL;
    } else if ((strncasecmp(source, "http://", 7) != 0) &&
           (strncasecmp(source, "https://", 8) != 0) &&
           (strncasecmp(source, "file://", 7) != 0) ) {
            ULOG_DEBUG("Error case 3. Only takes http,https,file protocol");
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_adding_failed"));
        g_free(source);
        source = NULL;
    } else if (strchr(source,'.') == NULL) {
        ULOG_DEBUG("Error case 4");
            fprintf(stderr,"invalid new feed address\n");
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_adding_failed"));
        g_free(source);
        source = NULL;
    } else {
        //something like this http://qewr./ is not valid. Should be http://abc.xyz/
        char *tmp = strchr(source,'.');
        fprintf(stderr, "tmp = %s\n", tmp);
        if (strncmp(tmp, "./", 2) == 0) {
            ULOG_DEBUG("Error case 5");
        fprintf(stderr,"invalid new feed address\n");
            gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_adding_failed"));
            g_free(source);
            source = NULL;
        }
        else {
            ULOG_DEBUG("New subscription: %s", source);                               
            //tvh :TODO: Why not modal here?????                     
            gtk_window_set_modal(GTK_WINDOW(app_ui_data->add_feed_dialog), TRUE);
            /* tvh: Moved from dbus iap connected handling here */
               gtk_widget_set_sensitive(app_data->app_ui_data->newfeedbtn, FALSE);
            gtk_widget_set_sensitive(app_data->app_ui_data->changefolderbtn, FALSE);
            // set the receiving here
            /*
            if (!app_data->app_ui_data->receiving_feeds )
                app_data->app_ui_data->receiving_feeds = TRUE;
            */
            ui_feedlist_new_subscription(source, NULL,
                     FEED_REQ_RESET_TITLE |
                     FEED_REQ_RESET_UPDATE_INT |
                     FEED_REQ_AUTO_DISCOVER |
                     FEED_REQ_DOWNLOAD_FAVICON,
                     ui_data->parent_folder);
        }
        }
    break;

    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_DELETE_EVENT:
    app_ui_data->dialog = NULL;
    app_ui_data->add_feed_dialog = NULL;
        
        if (SFM_DOWNLOAD == app_ui_data->search_mode) {
        app_ui_data->search_mode = SFM_INTERRUPTED;
        download_cancel_all();
        }
    gtk_widget_destroy(GTK_WIDGET(dialog));
    g_free(ui_data);
    break;
    }
}


/** By clicking on one of the buttons to close the feed details dialog
  * this function is called
  *
  * @param dialog the feed details dialog
  * @param response_id  the id sent by hitting one of the dialog buttons
  * @param user_data pointer to an ui data structure
  */

static void on_importopmldialog_response(GtkDialog * dialog, gint response_id,
                  gpointer user_data)
{
    ULOG_DEBUG ("on_importopmldialog_response........\n");
    struct fp_prop_ui_data *ui_data = NULL;
    folderPtr ptr = NULL;

    ui_data = (struct fp_prop_ui_data *) user_data;
    
    switch  (response_id)
    {
    case GTK_RESPONSE_OK:
        {
        fprintf(stderr, "on_import_dialog_response: GTK_RESPONSE_OK \n");
        /* get the parent feed pointer to pass here so that suscribed request
            can come into the right folder
        */
        ptr = ui_data->parent_folder;

        GtkWidget * opmlAddressEntry = lookup_widget (GTK_WIDGET (dialog), "opmlAddressEntry");
        const gchar *fd_source = gtk_entry_get_text(GTK_ENTRY (opmlAddressEntry) );
        
        ui_feed_directory_button_clicked(NULL,NULL,fd_source, ptr);
        }

    case GTK_RESPONSE_CANCEL:
        {
        //or a separate function like:
        importopml_dialog_destroy ();
       
        }
    default: 
        {
            ULOG_DEBUG("\ndefault case for opml dialog response");
        }
    }
}


/** By clicking on one of the buttons to close the feed details dialog
  * this function is called
  *
  * @param dialog the feed details dialog
  * @param response_id  the id sent by hitting one of the dialog buttons
  * @param user_data pointer to an ui data structure
  */
static void on_propdialog_response(GtkDialog * dialog, gint response_id,
                   gpointer user_data)
{
    struct fp_prop_ui_data *ui_data = (struct fp_prop_ui_data *) user_data;
    AppUIData *app_ui_data;
    gboolean close_dialog = TRUE;
    const gchar *title = NULL;
    const gchar *source = NULL;
    gchar *tmp = NULL;
    
    ULOG_DEBUG(__FUNCTION__);

    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;
    
    //g_assert(app_ui_data != NULL);
    
    switch (response_id) {
    case GTK_RESPONSE_OK:
    /* General */
        //g_assert(ui_data != NULL);
        //g_assert(ui_data->feedNameEntry != NULL);
        g_return_if_fail(ui_data->feedNameEntry != NULL);
    
        if (ui_data->parent_folder != NULL) {
           ULOG_INFO("%s: changing folder", __FUNCTION__);
           ULOG_DEBUG("fp: %p", ui_data->fp);
           ui_feedlist_remove((nodePtr) ui_data->fp);
           //g_assert(ui_data->fp->ui_data == NULL);
           g_return_if_fail(ui_data->fp->ui_data == NULL);
           ui_feedlist_add(ui_data->parent_folder,
                  (nodePtr) ui_data->fp, -1);
           ULOG_INFO("%s: changed folder", __FUNCTION__);
        } else {
           ULOG_INFO("%s: folder was not changed", __FUNCTION__);
        }

        title = gtk_entry_get_text(GTK_ENTRY
                      (ui_data->feedNameEntry));
        /*Make sure title is not NULL*/
        if( title == NULL)
        {
           gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_no_title_for_feed"));
           close_dialog = FALSE;
           break;
        }
        /* make sure title not only contains of white spaces*/
        else {
           tmp = g_strdup(title);
            //remove white spaces
           g_strstrip(tmp);
           if (strcmp (tmp, "") == 0)
           {
               gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_no_title_for_feed"));
               close_dialog = FALSE;
               break;
           }
 
        }

        source = gtk_entry_get_text(GTK_ENTRY
                      (ui_data->feedLocationEntry));

        /*Make sure source is not NULL*/
        if( source == NULL)
        {
           gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_no_source_for_feed"));
           close_dialog = FALSE;
           break;
        }
        else {
            if (NULL != tmp)
                g_free(tmp);
            tmp = g_strdup(source);
            g_strstrip(tmp);
            //if(strcmp(tmp, "") == 0)
            if (strlen(tmp) == 0 )
            {  //cannot be string of white_spaces
               gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ni_no_source_for_feed"));
               close_dialog = FALSE;
               break;
            }
            else
              {                              
               ULOG_DEBUG("%s: setting title", __FUNCTION__);
               feed_set_title(ui_data->fp, title);
               feed_set_source(ui_data->fp, source);
        
               ULOG_DEBUG("%s: call feedlist update ...", __FUNCTION__);
               ui_feedlist_update();
               ULOG_DEBUG("%s: call feedlist save ...", __FUNCTION__);
               conf_feedlist_schedule_save();
            }
    
        }    
       break;
    case RSS_RESPONSE_FEED_PROP_UNSUBSCRIBE:

       ULOG_INFO
        ("Unsubscribe was selected ... unsubscribe the feed now.");
    
        //g_assert(ui_data != NULL);
        g_return_if_fail(ui_data != NULL);
        //g_assert(ui_data->fp != NULL);
        g_return_if_fail(ui_data->fp != NULL);
        
        if((nodePtr)ui_data->fp != ui_feedlist_get_selected())
            ui_feedlist_select((nodePtr)ui_data->fp);
    
        close_dialog = ui_feedlist_delete((nodePtr)ui_data->fp);

        break;
    }

    ULOG_DEBUG("on_propdialog_response: done");

    if(close_dialog)
    {
        g_free(ui_data);
        app_ui_data->dialog = NULL;
        gtk_widget_destroy(GTK_WIDGET(dialog));
    }
    /*g_free*/
    if (NULL != tmp)
        g_free(tmp);
}

/** Determines the feeds favicon or default icon 
  *
  * @param fp the feed to get the icon for
  * @return the icon
  */
static GdkPixbuf* ui_feed_get_icon(feedPtr fp) 
{
    gpointer    favicon;
    
    //g_assert(FST_FOLDER != fp->type);
    g_return_val_if_fail ((FST_FOLDER != fp->type), NULL);

    if(NULL != (favicon = feed_get_favicon(fp)))
        return favicon;
    if (feed_get_available(fp))
        return icons[ICON_RSS_NEWS_FEED];
    else 
        return icons[ICON_RSS_NEWS_INVALID_FEED];
}


/* Remove the hidden and new attribute from that items, what it has these*/
void ui_feed_remove_hidden_attribute(feedPtr fp)
{
    GSList *item = NULL;
    item = fp->items;
    int i=0;
//    ULOG_DEBUG("Entering ui_feed_remove_hidden_attribute\n");
    /*We have to analyse, if we have hidden items. If we don't, we don't do anything. */
    while ((NULL != item) && !i) {
        if (item_get_hidden((itemPtr) item->data))
        {
            i++;
            ULOG_DEBUG("Hidden item found %d\n",i);
        }
        item = g_slist_next(item);
    }
    item = fp->items;
    /*If we don't have items, what have hidden status, than there are only new 
    items, so we don't need to do anything*/
    if (i)
    {
        while (NULL != item) 
        {
            if (item_get_new_status((itemPtr) item->data))
            {
                if (item_get_hidden((itemPtr) item->data))
                {
                    ULOG_DEBUG("Removing Hidden and New attribute\n");
                    item_set_hidden((itemPtr) item->data,FALSE);
                    item_set_new_status((itemPtr) item->data,FALSE);
                }
            }
            item = g_slist_next(item);
        }        
    }
}

/* Delete the old items. It will delet only, if we have hidden items
(setted by the ui_feed_set_new_items_hidden function, before the beginning of
the refresh)*/
void ui_feed_delete_old_items(feedPtr fp)
{
    GSList *item = NULL;
    item = fp->items;
    int i=0;
//    ULOG_DEBUG("Entering ui_feed_delete_old_items\n");
    /*We have to analyse, if we have hidden items. If we don't, we don't do anything. */
    while ((NULL != item) && !i) {
        if (item_get_hidden((itemPtr) item->data))
        {
            i++;
            ULOG_DEBUG("Hidden item found %d\n",i);
        }
        item = g_slist_next(item);
    }
    item = fp->items;
    /*If we don't have items, what have hidden status, than there are only new 
    items, so we don't need to do anything*/
    if (i)
    {
        while (NULL != item) 
        {
            if (FALSE==item_get_new_status((itemPtr) item->data))
            {
                if(FALSE==item_get_mark(item->data))
                {
                    ULOG_DEBUG("Deleting unneded item\n");
                    feed_remove_item(fp,item->data);
                    item=NULL;
                }
            }

            if (item) 
                item = g_slist_next(item);
            else
                item=fp->items;
        }        
    }
}

/*Temporarly mark the New items as hidden. So we will know for example after 
refreshing, what items are realy new, and what was new just before refreshing*/
void ui_feed_set_new_items_hidden(feedPtr fp)
{
    GSList *item = NULL;
    itemPtr curitem;
//    ULOG_DEBUG("Entering ui_feed_set_new_items_hidden\n");
    //g_assert(fp != NULL);
    g_return_if_fail(fp != NULL);
    item = fp->items;

    while (NULL != item) {
        
        curitem=item->data;
        if (fp->newCount)
        {
        if (TRUE==curitem->newStatus)
            {
                ULOG_DEBUG("Hiding an item\n");
                curitem->hidden=TRUE;
            }
    }
    item = g_slist_next(item);
    }
//    ULOG_DEBUG("Leaving ui_feed_set_new_items_hidden\n");
}

static gboolean favicon_idle_cb(gpointer fp)
{
    favicon_download((feedPtr) fp);
    return FALSE;
}
/************************************************************************/
/*                        PUBLIC FUNCTIONS                              */
/************************************************************************/

/** handles completed feed update requests */
void ui_feed_process_update_result(struct request *request)
{
    feedPtr fp = (feedPtr) request->user_data;
    feedHandlerPtr fhp = NULL;
    gchar *old_title = NULL, *old_source = NULL;
    gint old_update_interval = 0;

    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    debug_enter("ui_feed_process_update_result");
    ULOG_DEBUG("ui_feed_process_update_result");
    
    /* NOTE: Removed this */
//    ui_lock();
    /*Temporarly mark the "old" new items, as hidden*/
    ui_feed_set_new_items_hidden(fp);
    
    //g_assert(NULL != request);
    g_return_if_fail(NULL != request);

    feed_load(fp);
    
    feed_reset_update_counter(fp);

    ULOG_DEBUG("\nret from download: request->source: %s",request->source);
    ULOG_DEBUG("request->data: %d",(request->data==NULL?0:1));
    //ULOG_DEBUG("request->data: %s", (gchar * ) request->data);
   
    /* no matter what the result of the update is we need to save update
       status and the last update time to cache */
    fp->needsCacheSave = TRUE;

    feed_set_available(fp, TRUE);

    if (401 == request->httpstatus) {    /* unauthorized */
        ULOG_DEBUG("Returned 401: Unauthorized request");
    feed_set_available(fp, FALSE);
    app_data->app_ui_data->errors_in_refreshing = TRUE;
    } else if (410 == request->httpstatus) {    /* gone */
        ULOG_DEBUG("Returned 410");
    feed_set_available(fp, FALSE);
    feed_set_discontinued(fp, TRUE);
    ULOG_DEBUG
        ("\"%s\" is discontinued. Liferea won't updated it anymore!",
         feed_get_title(fp));
    app_data->app_ui_data->errors_in_refreshing = TRUE;
    } else if (304 == request->httpstatus) {
        ULOG_DEBUG("Returned 304");
        /*feed_remove_unmarked(fp, TRUE);*/
    } else if (NULL != request->data) {
        ULOG_DEBUG("Returned OK. Starting to parse the feed");
        
    feed_set_lastmodified(fp, request->lastmodified);
    feed_set_etag(fp, request->etag);

    /* note this is to update the feed URL on permanent redirects */
    if (0 != strcmp(request->source, feed_get_source(fp))) {
        feed_set_source(fp, request->source);
        ULOG_DEBUG
        ("The URL of \"%s\" has changed permanently and was updated",
         feed_get_title(fp));
    }

    /* we save all properties that should not be overwritten in all cases */
    old_update_interval = feed_get_update_interval(fp);
    old_title = g_strdup(feed_get_title(fp));
    old_source = g_strdup(feed_get_source(fp));

    ULOG_DEBUG("Old title: %s",old_title);

        feed_set_parse_errors(fp, FALSE);

    /* parse the new downloaded feed into new_fp */
    fhp =
        feed_parse(fp, request->data, request->size,
               request->flags & FEED_REQ_AUTO_DISCOVER);

    if (fhp == NULL || feed_get_parse_errors(fp)) {
        ULOG_DEBUG("Could not detect the type of this feed");
        feed_set_available(fp, FALSE);
        app_data->app_ui_data->errors_in_refreshing = TRUE;
    }
    else if (strcmp(fhp->typeStr, "opml") == 0)                         
    {
        /* If opml gotten from add feed dialog. Just returns, because feed directory
        dialog will be popped up by now */

       //free stuff
       g_free(old_title);
       g_free(old_source);
       fp = NULL;
        return;
    }    
     
     else if(feed_get_new_subscription(fp) == TRUE &&
            feed_get_feed_directory(fp) == TRUE &&
            app_data->app_ui_data->add_feed_dialog != NULL) {
        ULOG_DEBUG("Unnecessary feed directory found. Not adding to the feed tree");
        
        if(!app_data->app_ui_data->feed_directory_cancelled)
            newdialog_destroy();
        else
            newdialog_return_to_normal();
            
        app_data->app_ui_data->feed_directory_cancelled = FALSE;            
        
        feed_set_new_subscription(fp, FALSE);
        feed_set_available(fp, FALSE);
    } else {
        fp->fhp = fhp;
            
        ULOG_DEBUG("Feed type found %s", fhp->typeStr);
        
        /* Feed has been added from the add feed dialog and has been found so
           it is added to the feed tree */
        if(feed_get_new_subscription(fp) &&
                app_data->app_ui_data->add_feed_dialog != NULL)
        {
            folderPtr selected_parent = app_data->app_ui_data->ui_data->parent_folder;
            folderPtr parent = NULL;
        
            parent = (selected_parent != NULL)
                ? selected_parent : ui_feedlist_get_root_folder();
         
            ui_feedlist_add(parent, (nodePtr) fp, -1);
            ui_feedlist_update();
            ui_feedlist_select((nodePtr) fp);
            
            newdialog_destroy();
            feed_set_new_subscription(fp, FALSE);
        }
        
        /* restore user defined properties if necessary */
        if (!(request->flags & FEED_REQ_RESET_TITLE))
        {
            ULOG_DEBUG("Resetting the title");
            feed_set_title(fp, old_title);
        }

        if (!(request->flags & FEED_REQ_AUTO_DISCOVER))
            feed_set_source(fp, old_source);
        
        ui_feed_delete_old_items(fp);
        ui_feed_remove_hidden_attribute(fp);

        if (!(request->flags & FEED_REQ_RESET_UPDATE_INT))
            feed_set_update_interval(fp, old_update_interval);
        

        ULOG_DEBUG("\"%s\" updated...",
                     feed_get_title(fp));

        if ((feedPtr) displayed_node == fp)
            ui_itemlist_load((nodePtr) fp);

        if (request->flags & FEED_REQ_SHOW_PROPDIALOG)
        {
            //g_assert(app_data != NULL);
            //g_assert(app_data->app_ui_data != NULL);
        
            ui_feed_propdialog_new(GTK_WINDOW
                       (app_data->app_ui_data->app), fp);
        }
    }
    g_free(old_title);
    g_free(old_source);
    } else {
        ULOG_DEBUG("Feed not available");
        ULOG_DEBUG("\"%s\" is not available",
                     feed_get_title(fp));
        feed_set_available(fp, FALSE);
        app_data->app_ui_data->errors_in_refreshing = TRUE;
    }

    ULOG_DEBUG("Download processing ended");
    
    ui_feed_set_error_description(fp, request->httpstatus,
                  request->returncode);

    fp->request = NULL;
    ui_notification_update(fp);
    ui_feedlist_update();
    
    if(feed_get_available(fp))
        if(request->flags & FEED_REQ_DOWNLOAD_FAVICON)
            /* 2006June14 should add this in idle fashion otherwise
            the call add_counter (NUM_DOWNLOADED, -1) and this     g_mutex_unlock(num_downloaded_check_mutex);
            in process_downloaded_idle_cb() will not be right coz another request has been pushed
            while processing one new feeds.*/
            //favicon_download(fp);
            //g_idle_add( favicon_idle_cb, fp);    
            g_idle_add_full(G_PRIORITY_HIGH_IDLE, favicon_idle_cb, fp, NULL);            
            /*or we can do this to reduce the downloading every time. This is hackish. But acceptable
            if (!feed_get_favicon(fp))
                //then download it
                g_idle_add_full(G_PRIORITY_HIGH_IDLE, favicon_idle_cb, fp, NULL);                        
            */
        
    feed_unload(fp);

    /* If the add feed dialog is displayed and the feed is invalid
       or it is a feed directory the feed is freed */
    if((app_data->app_ui_data->add_feed_dialog != NULL &&
       feed_get_available(fp) == FALSE) ||
       (feed_get_feed_directory(fp) == TRUE && 
       feed_get_available(fp) == FALSE))
    {
       ULOG_DEBUG("Deleting an invalid feed");
       ui_feedlist_delete_((nodePtr)fp);
    }
//    ui_unlock();

    debug_exit("ui_feed_process_update_result");
}

/** updating of a single feed list entry */
void ui_feed_update(feedPtr fp)
{
    GtkTreeModel *model = NULL;
    GtkTreeIter *iter = NULL;
    gchar *label = NULL, *tmp = NULL;
    int unread = 0, marked = 0;

    if (fp->ui_data == NULL)
    return;

    iter = &((ui_data *) fp->ui_data)->row;
    model = GTK_TREE_MODEL(feedstore);

    //g_assert(FST_FOLDER != fp->type);
    g_return_if_fail(FST_FOLDER != fp->type);

    if (FST_FEED == fp->type)
    rss_max_feed = rss_max_feed + 1;
    unread = feed_get_unread_counter(fp);
    marked = feed_get_mark_counter(fp);
    /*
    label = unhtmlize(g_strdup(feed_get_title(fp)));

    // FIXME: Unescape text here! 
    tmp = g_markup_escape_text(label, -1);
    tmp = g_strstrip(tmp);
    
    g_free(label);
    */

    //let's be simple. why need to unhtmlize?    

    //need to escape though, as textrenderer is using markup


    label = g_strdup(feed_get_title(fp));
    label= utf8_fix(label);
    tmp = g_markup_escape_text(label, -1);    
    tmp = g_strstrip(tmp);
//    ULOG_DEBUG("ui_feed_update: tmp = %s", tmp);
    g_free(label);
    
    if (unread > 0 && marked > 0)
    label =
        g_strdup_printf
        ("<span weight=\"bold\" foreground=\"%s\">%s (%d)</span>",
         "red"/*get_logical_color(EMPTEXTCOLOR)*/, tmp, unread);
    else if (marked > 0)
    label = g_strdup_printf("<span foreground=\"%s\">%s</span>",
                "red"/*get_logical_color(EMPTEXTCOLOR)*/, tmp);
    else if (unread > 0)
    label =
        g_strdup_printf
        ("<span foreground=\"%s\" weight=\"bold\">%s (%d)</span>",
         "black" /*get_logical_color(DEFAULTTEXTCOLOR)*/, tmp, unread);
    else
    label = g_strdup_printf("<span foreground=\"%s\">%s</span>",
                "black" /*get_logical_color(DEFAULTTEXTCOLOR)*/, tmp);
    g_free(tmp);
    
    /* CHECK: valgrind says internal row-changed event from this leaks memory */
    gtk_tree_store_set(feedstore, iter, FS_LABEL, label,
               FS_UNREAD, unread,
               FS_ICON, ui_feed_get_icon(fp), -1);
    g_free(label);
}

/********************************************************************
 * newdialog                                                       *
 *******************************************************************/

GtkWidget *ui_feed_newdialog_new(GtkWindow * parent, const gchar *source)
{
    GtkWidget *newdialog;
    struct fp_prop_ui_data *ui_data = NULL;
    AppUIData *app_ui_data;
    folderPtr ptr = NULL;
    
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    app_ui_data->ui_data = ui_data = g_new0(struct fp_prop_ui_data, 1);

    //g_assert(app_ui_data != NULL);
    
    if(ui_data == NULL) {
        ULOG_DEBUG("Not enough memory for add feed dialog");
        return NULL;
    }
    
    /* Create the dialog */
    app_ui_data->add_feed_dialog = app_ui_data->dialog = ui_data->dialog = newdialog = create_newdialog();
    
    ui_data->label_widget = lookup_widget(newdialog, "folderName");
    
    ptr = ui_feedlist_get_target_folder();

    if(ptr != NULL && ptr != ui_feedlist_get_root_folder()) {
        gtk_label_set_text(GTK_LABEL(ui_data->label_widget),
               ptr->title);
    
        ui_data->parent_folder = ptr;
    } else {
        ui_data->parent_folder = ui_feedlist_get_root_folder();
    }

    GtkWidget *changefolder = lookup_widget(newdialog, "changefolder");
    g_signal_connect(changefolder, "clicked",
             G_CALLBACK(display_change_folder_dialog), ui_data);

    gtk_window_set_transient_for(GTK_WINDOW(newdialog),
                 GTK_WINDOW(parent));
    gtk_window_set_modal(GTK_WINDOW(newdialog), TRUE);

    /***********************************************************************
     * Source                                                              *
     **********************************************************************/

    /* Setup source entry */
    ui_data->sourceEntry = lookup_widget(newdialog, "sourceEntry");
    if (source != NULL)
    {
        ULOG_DEBUG("ui_feed_newdialog_new: setting default sourcEntry to: %s", source);
        //set it here
         gtk_entry_set_text (GTK_ENTRY(ui_data->sourceEntry), (const gchar *) source);
    }    

    gtk_widget_grab_focus(GTK_WIDGET(ui_data->sourceEntry));
    gtk_entry_set_activates_default(GTK_ENTRY(ui_data->sourceEntry), TRUE);

    gtk_widget_grab_default(lookup_widget(newdialog, "newfeedbtn"));
    g_signal_connect(G_OBJECT(newdialog), "response",
             G_CALLBACK(on_newdialog_response), ui_data);
    g_signal_connect(G_OBJECT(newdialog), "key-press-event",
             G_CALLBACK(key_press_for_cancelling_dialog), NULL);
    gtk_widget_show_all(newdialog);

    return newdialog;
}

/*
    importopml file dialog (interface.c)
    Connect it to actions and signals
*/
GtkWidget *ui_feed_importopmldialog(GtkWindow * parent, const gchar * opml_source)
{
    GtkWidget *importopmldialog;
    struct fp_prop_ui_data *ui_data = NULL;
    AppUIData *app_ui_data;
    folderPtr ptr = NULL;
    
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    app_ui_data->ui_data = ui_data = g_new0(struct fp_prop_ui_data, 1);

    //g_assert(app_ui_data != NULL);
    
    if(ui_data == NULL) {
        ULOG_DEBUG("Not enough memory to import opml");
        return NULL;
    }
    
    /* Create the dialog */
    app_ui_data->importopmldialog = app_ui_data->dialog = ui_data->dialog = importopmldialog = create_importopmldialog();
    fprintf(stderr, "ui_feed_importopmldialog: tvh: created opml dialog \n");

    ui_data->label_widget = lookup_widget(importopmldialog, "folderName");
    
    ptr = ui_feedlist_get_target_folder();

    if(ptr != NULL && ptr != ui_feedlist_get_root_folder()) {
        ui_data->parent_folder = ptr;
    } else {
        ui_data->parent_folder = ui_feedlist_get_root_folder();
    }

    gtk_label_set_text(GTK_LABEL(ui_data->label_widget),
               ui_data->parent_folder->title);


    /*signal for "browse"
        Will popup a Hildon File Chooser with filter to *.opml and *.xml
    */
    GtkWidget *browser = lookup_widget(importopmldialog, "importopml_browse");
    g_signal_connect(browser, "clicked",
             G_CALLBACK(display_file_chooser_dialog), ui_data);

    /*signal for "changefolder"*/
    GtkWidget *changefolder = lookup_widget(importopmldialog, "changefolder");
    g_signal_connect(changefolder, "clicked",
             G_CALLBACK(display_change_folder_dialog), ui_data);

    gtk_window_set_transient_for(GTK_WINDOW(importopmldialog),
                 GTK_WINDOW(parent));
    gtk_window_set_modal(GTK_WINDOW(importopmldialog), TRUE);

    /***********************************************************************
     * Source                                                              *
     **********************************************************************/

    /* Setup opml source file entry */
    ui_data->sourceEntry = lookup_widget(importopmldialog, "opmlAddressEntry");
    gtk_widget_grab_focus(GTK_WIDGET(ui_data->sourceEntry));
    gtk_entry_set_activates_default(GTK_ENTRY(ui_data->sourceEntry), TRUE);

    if (opml_source != NULL)
    {
        ULOG_DEBUG("ui_feed_importopmldialog: setting default sourcEntry to: %s", opml_source);
        //set it here
         gtk_entry_set_text (GTK_ENTRY(ui_data->sourceEntry), (const gchar *) opml_source);
    }    


    gtk_widget_grab_default(lookup_widget(importopmldialog, "importopml_ok"));

    /* signal to the OK and CANCEL buttons' */
    g_signal_connect(G_OBJECT(importopmldialog), "response",
             G_CALLBACK(on_importopmldialog_response), ui_data);

    /*ESC key event */
    g_signal_connect(G_OBJECT(importopmldialog), "key-press-event",
             G_CALLBACK(key_press_for_cancelling_dialog), NULL);
    gtk_widget_show_all(importopmldialog);

    return importopmldialog;
}

GtkWidget *ui_feed_propdialog_new(GtkWindow * parent, feedPtr fp)
{
    GtkWidget *propdialog;
    struct fp_prop_ui_data *ui_data = NULL;
    AppUIData *app_ui_data;
     
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    ui_data = g_new0(struct fp_prop_ui_data, 1);
    ui_data->fp = fp;

    //g_assert(app_ui_data != NULL);
    
    /* Create the dialog */
    app_ui_data->dialog = ui_data->dialog = propdialog =
    create_propdialog();

    GtkWidget *changefolder = lookup_widget(propdialog, "changefolder");
    g_signal_connect(changefolder, "clicked",
             G_CALLBACK(display_change_folder_dialog), ui_data);

    gtk_window_set_transient_for(GTK_WINDOW(propdialog),
                 GTK_WINDOW(parent));
    
    gtk_window_set_modal(GTK_WINDOW(propdialog),TRUE);                 
    /***********************************************************************
     * General                                                             *
     **********************************************************************/

    /* If we're not modal, ui_data may be overwritten */
    ui_data->feedNameEntry = lookup_widget(propdialog, "feedNameEntry");
    gtk_entry_set_text(GTK_ENTRY(ui_data->feedNameEntry),
               feed_get_title(fp));
    ui_data->feedLocationEntry = lookup_widget(propdialog, "feedLocationEntry");
    gtk_entry_set_text(GTK_ENTRY(ui_data->feedLocationEntry),
               feed_get_source(fp));

    GtkWidget *subscribedData =
    lookup_widget(propdialog, "subscribedData");
    gchar *time_str = ui_itemlist_format_date(feed_get_added(fp));
    gtk_label_set_text(GTK_LABEL(subscribedData), time_str);
    g_free(time_str);

    GtkWidget *folderName = lookup_widget(propdialog, "folderName");
    folderPtr parentPtr = (folderPtr) ui_feedlist_get_parent((nodePtr) fp);
    if (parentPtr != NULL) {
     gtk_label_set_text(GTK_LABEL(folderName), parentPtr->title);
    }

    /***********************************************************************
     * Source                                                              *
     **********************************************************************/

    g_signal_connect(G_OBJECT(propdialog), "response",
             G_CALLBACK(on_propdialog_response), ui_data);
    g_signal_connect(G_OBJECT(propdialog), "key-press-event",
             G_CALLBACK(key_press_for_cancelling_dialog), NULL);
    ui_data->label_widget = lookup_widget(propdialog, "folderName");

    gtk_widget_show_all(propdialog);

    return propdialog;
}

void newdialog_return_to_normal()
{
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    gtk_window_set_modal(GTK_WINDOW(app_data->app_ui_data->add_feed_dialog), TRUE);
    gtk_widget_set_sensitive(app_data->app_ui_data->newfeedbtn, TRUE);
    gtk_widget_set_sensitive(app_data->app_ui_data->changefolderbtn, TRUE);
//     gtk_widget_set_sensitive(app_data->app_ui_data->cancelbtn, TRUE);
}

void newdialog_destroy()
{
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);    

    gtk_widget_destroy(GTK_WIDGET(app_data->app_ui_data->add_feed_dialog));
        
    app_data->app_ui_data->dialog = NULL;
    app_data->app_ui_data->add_feed_dialog = NULL;
     
    g_free(app_data->app_ui_data->ui_data);
        
    app_data->app_ui_data->ui_data = NULL;
}

void  importopml_dialog_destroy (void)
{
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);    

    gtk_widget_destroy(GTK_WIDGET(app_data->app_ui_data->importopmldialog));
        
    app_data->app_ui_data->dialog = NULL;
    app_data->app_ui_data->importopmldialog = NULL;
    app_data->app_ui_data->dialog_closed = TRUE; 
    
    g_free(app_data->app_ui_data->ui_data);
        
    app_data->app_ui_data->ui_data = NULL;


}
