/**
 * @file ui_feedlist.c GUI feed list handling
 *
 * Copyright (C) 2004 Lars Lindner <lars.lindner@gmx.net>
 * Copyright (C) 2004 Nathan J. Conrad <t98502@users.sourceforge.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

/* for strsep */
#define _BSD_SOURCE

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <osso-log.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <hildon-widgets/gtk-infoprint.h>
#include <osso-ic-gconf.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <osso-ic.h>

//#include "guitreemodelfilter.h"
#include "support.h"
#include "interface.h"
#include "callbacks.h"
#include "feed.h"
#include "folder.h"
#include "conf.h"
#include "ui_feedlist.h"
#include "ui_mainwindow.h"
#include "ui_feed.h"
#include "update.h"
#include "htmlview.h"
#include "favicon.h"
#include "debug.h"
#include "ui_notification.h"
#include "dbus.h"
//#include "gtkhtml/gtkhtml_plugin.h"

#define AUTOMATIC_DELETION_TIMEOUT 24 * 60

typedef struct {
    gpointer ptr;
    gboolean found;
    GtkTreeIter iter;
    GtkTreePath *path;
} SearchData;

extern GtkWidget *mainwindow;
extern GHashTable *feedHandler;
extern AppData *app_data;
extern guint rss_max_feed;
extern gulong rss_unread;
extern time_t rss_updated;

GtkTreeModel *feedmodel = NULL;
GtkTreeStore *feedstore = NULL;

/* flag to enable/disable the GtkTreeModel filter */
gboolean filter_feeds_without_unread_headlines = FALSE;

static GtkTreeModel *filter = NULL;
static folderPtr root_folder_ptr = NULL;

/* Lock used to control the updating of gtkhtml. Used when
 1. Switching too fast between 2 feeds. while loop in gtkhtml_write_html, without this
    will crash
 2. Deleting a being-displayed feed
*/
G_LOCK_DEFINE_STATIC (load_selected_mutex);
G_LOCK_DEFINE_STATIC (last_node_mutex);

gboolean ui_feedlist_auto_update(void *data);
void rss_feed_tree_model_init(GtkTreeModelIface * iface,
                  gpointer user_data);

/************************************************************************/
/*                        PRIVATE FUNCTIONS                             */
/************************************************************************/

/**
 * Helper function to search a tree model by a node.
 * To be used with gtk_tree_model_foreach.
 *
 * @param model GtkTreeModel (not NULL)
 * @param path current path (not NULL)
 * @param iter current iterator (not NULL)
 * @param data pointer to a SearchData object (not NULL)
 * @return TRUE if the searched item was found, FALSE otherwise
 */
static
 gboolean
search_func(GtkTreeModel * model,
        GtkTreePath * path, GtkTreeIter * iter, gpointer data)
{
    SearchData *data1 = NULL;
    gpointer ptr = NULL;

    //g_assert(model != NULL && path != NULL && iter != NULL);
    g_return_val_if_fail((model != NULL && path != NULL && iter != NULL), FALSE);
    data1 = (SearchData *) data;
    //g_assert(data1 != NULL);
    g_return_val_if_fail((data1 != NULL), FALSE);
    gtk_tree_model_get(model, iter, FS_PTR, &ptr, -1);
    if (data1->ptr == ptr) {
    //g_assert(data1->path == NULL);
    g_return_val_if_fail ((data1->path == NULL), FALSE);
    data1->found = TRUE;
    data1->iter = *iter;
    data1->path = gtk_tree_path_copy(path);
    return TRUE;
    }
    return FALSE;
}


/**
 * Helper function to count the feeds in a folder hierarchy.
 *
 * @param node current node
 * @param data pointer to an integer (gint), which is incremented
 *   (not NULL)
 */
static
void count_feeds_func(nodePtr node, gpointer data)
{
    gint *count_ptr = NULL;

    /* We don't check the type of the node here.
       The caller of ui_feedlist_do_for_all_full
       has to use an appropriate filter.
     */
    count_ptr = (gint *) data;
    //g_assert(count_ptr != NULL);
    g_return_if_fail(count_ptr != NULL);
    (*count_ptr)++;
}

/** Updates the feedlist
  *
  * @param iter current location in the feed tree
  */
static void ui_feedlist_update_(GtkTreeIter * iter)
{
    GtkTreeIter childiter;
    gboolean valid = FALSE;
    nodePtr ptr = NULL;

    //ULOG_DEBUG("ui_feedlist_update");
    if (iter != NULL) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), iter, FS_PTR, &ptr,
               -1);
    valid =
        gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore),
                     &childiter, iter);
    } else {
    valid =
        gtk_tree_model_get_iter_first(GTK_TREE_MODEL(feedstore),
                      &childiter);
    }

    if (ptr != NULL)
    ((ui_data *) (ptr->ui_data))->row = *iter;

    while (valid) {
    ui_feedlist_update_(&childiter);
    valid =
        gtk_tree_model_iter_next(GTK_TREE_MODEL(feedstore),
                     &childiter);
    }

    if (ptr != NULL) {
    if (FST_FOLDER == ptr->type)
        ui_folder_update((folderPtr) ptr);
    else
        ui_feed_update((feedPtr) ptr);
    }
}

/** Handles feedlist key presses
  *
  * @param widget not used
  * @param event not used
  * @param data not used
  * @return always FALSE
  */
static gboolean ui_feedlist_key_press_cb(GtkWidget * widget,
                     GdkEventKey * event,
                     gpointer data)
{
    return FALSE;
}

/** Function for deciding whether unread headlines should be shown or not
  *
  * @param model the tree model to use
  * @param iter current location in the tree
  * @param data not used
  */
static gboolean filter_visible_function(GtkTreeModel * model,
                    GtkTreeIter * iter, gpointer data)
{
    gint count = 0;

    if (!filter_feeds_without_unread_headlines)
    return TRUE;

    gtk_tree_model_get(model, iter, FS_UNREAD, &count, -1);

    if (0 != count)
    return TRUE;
    else
    return FALSE;
}

G_DEFINE_TYPE_WITH_CODE(RssFeedTree, rss_feed_tree, G_TYPE_OBJECT,
            G_IMPLEMENT_INTERFACE(GTK_TYPE_TREE_MODEL,
                          rss_feed_tree_model_init)
            G_IMPLEMENT_INTERFACE(GTK_TYPE_TREE_DRAG_DEST,
                          rss_feed_tree_drag_init))

static void rss_feed_tree_init(RssFeedTree * self)
{
}

static void rss_feed_tree_class_init(RssFeedTreeClass * klass)
{
}

/** This function is used by the auto update mechanism. It simply
  * checks if an internet connection exists and if so, does a
  * refresh
  *
  * @param fp a pointer to the feed to update
  */ 
  /*TODO: Remove this */
static void ui_feedlist_check_update_counter(feedPtr fp)
{
    ULOG_DEBUG("Auto update");
    if(!app_data->app_ui_data->start_on_background)
    {
#ifdef WITHOUT_CONNECTIVITY    
        feed_schedule_update(fp, FEED_REQ_DOWNLOAD_FAVICON);
#else
    //    if(*getStringConfValue(ICD_GCONF_CURRENT))
    //        feed_schedule_update(fp,0);
    gchar *icd_conf = NULL;
    icd_conf =getStringConfValue(ICD_GCONF_CURRENT);
    if (strlen(icd_conf) > 0)
        feed_schedule_update(fp, FEED_REQ_DOWNLOAD_FAVICON);
    g_free(icd_conf);
#endif        
    }
}

/************************************************************************/
/*                        PUBLIC FUNCTIONS                              */
/************************************************************************/

void ui_feedlist_delete_(nodePtr ptr)
{
    //g_assert(ptr != NULL);
    g_return_if_fail(ptr != NULL);

    if (ptr == (nodePtr) root_folder_ptr) {
    root_folder_ptr = NULL;
    }
    if ((FST_FEED == ptr->type) || (FST_VFOLDER == ptr->type)) {
        ULOG_DEBUG("Deleting feed");
        ui_notification_remove_feed((feedPtr) ptr);    /* removes an existing notification for this feed */
       ui_folder_remove_node(ptr);
       feed_free((feedPtr) ptr);
    } else {
    ui_feedlist_do_for_all(ptr,
                   ACTION_FILTER_CHILDREN | ACTION_FILTER_ANY,
                   ui_feedlist_delete_);
    folder_free((folderPtr) ptr);
    }
    ui_feedlist_update();    /* because vfolder unread counts may have changed */
}

gboolean ui_feedlist_is_ancestor(nodePtr ptr, GtkTreeIter *folder)
{
       GtkTreeIter *iter = NULL;
       GtkTreeModel *fm = NULL;
       GtkTreePath *node_path = NULL;
       GtkTreePath *folder_path = NULL;
       gboolean is_ancestor = FALSE;                                   
       gchar *folder_string = NULL;
       gchar *node_string = NULL;

       iter = &((ui_data *) (ptr->ui_data))->row;

       if(iter == NULL)
           return FALSE;

       fm = GTK_TREE_MODEL(feedstore);

       node_path = gtk_tree_model_get_path(fm, iter);
       folder_path = gtk_tree_model_get_path(fm, folder);
       
       folder_string = gtk_tree_path_to_string(folder_path);
       node_string = gtk_tree_path_to_string(node_path);
       
       g_free(folder_string);
       g_free(node_string);

       is_ancestor = gtk_tree_path_is_ancestor(folder_path, node_path);

       gtk_tree_path_free(node_path);
       gtk_tree_path_free(folder_path);

       return is_ancestor;
}

folderPtr ui_feedlist_get_parent(nodePtr ptr)
{
    GtkTreeIter *iter = &((ui_data *) (ptr->ui_data))->row;
    GtkTreeIter parent;
    folderPtr parentPtr = NULL;

    if (gtk_tree_model_iter_parent
    (GTK_TREE_MODEL(feedstore), &parent, iter)) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &parent, FS_PTR,
               &parentPtr, -1);
    return parentPtr;
    }

    return NULL;
}

nodePtr ui_feedlist_get_selected(void)
{
    GtkTreeSelection *select = NULL;
    GtkTreeModel *model = NULL;
    GtkTreeIter iter;
    nodePtr ptr = NULL;

    /*ULOG_DEBUG(__FUNCTION__);*/

    select = gtk_tree_view_get_selection(GTK_TREE_VIEW(feedlist));
    if (gtk_tree_selection_get_selected(select, &model, &iter)) {
    gtk_tree_model_get(model, &iter, FS_PTR, &ptr, -1);
    return ptr;
    }

    return NULL;
}

gboolean ui_feedlist_display_news(gchar * feed, guint32 nr)
{

    ULOG_DEBUG("%s feed %s item %d", __FUNCTION__, feed, nr);
    //g_assert(NULL != feed);
    g_return_val_if_fail ((NULL != feed), FALSE);

    nodePtr ptr = ui_feedlist_find_feed(NULL, feed);
    if (NULL == ptr) {
    ULOG_WARN("Could not find feed (%s)", feed);
    return FALSE;
    }
    ui_feedlist_select(ptr);
    if (displayed_node != ptr) {
        ui_feedlist_load_selected(ptr);
    
        app_data->app_ui_data->scroll_to_feed = ptr;
        app_data->app_ui_data->scroll_to_item = nr;
    } else
        gtkhtml_scroll_to_item(ptr, nr);
    return TRUE;
}

nodePtr ui_feedlist_find_feed(GtkTreeIter * from, gchar * feed)
{
    GtkTreeIter iter;
    gboolean valid = FALSE;
    nodePtr node = NULL;

    //g_assert(NULL != feedstore);
    g_return_val_if_fail ((NULL != feedstore), NULL);

    if (from != NULL) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), from, FS_PTR, &node,
               -1);
    valid =
        gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore), &iter,
                     from);
    } else {
    valid =
        gtk_tree_model_get_iter_first(GTK_TREE_MODEL(feedstore),
                      &iter);
    }

    while (valid) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &iter,
               FS_PTR, &node, -1);
    if (NULL != node) {
        if (FST_FEED == feed_get_type((feedPtr) node)) {
        if (strcmp(((feedPtr) node)->source, feed) == 0) {
            return node;
        }
        }
        if (FST_FOLDER == feed_get_type((feedPtr) node)) {
        node = ui_feedlist_find_feed(&iter, feed);
        if (NULL != node) {
            return node;
        }
        }
    } else {
        /* CHECK: Why would we have a NULL value in the tree model? */
    }

    valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(feedstore), &iter);
    }

    return NULL;
}

void ui_feedlist_find_mainwindow_data(GtkTreeIter * from, gint * nr_feeds,
                      gint * nr_unread,
                      gint * nr_unread_feeds,
                      gint * nr_marked,
                      glong * last_poll_time)
{
    GtkTreeIter iter;
    gboolean valid = FALSE;
    nodePtr node = NULL;
    gint nr_unread_s = 0;

    //g_assert(NULL != feedstore);
    g_return_if_fail(NULL != feedstore);
    //ULOG_DEBUG("ui_feedlist_find_mainwindow_data: traversing through the feedlist now ");
    if (from != NULL) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), from, FS_PTR, &node,
               -1);
    valid =
        gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore), &iter,
                     from);
    } else {
    valid =
        gtk_tree_model_get_iter_first(GTK_TREE_MODEL(feedstore),
                      &iter);
    }

    while (valid) {
    nr_unread_s = 0;
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &iter,
               FS_PTR, &node, -1);
    if (NULL != node) {
        if (FST_FEED == feed_get_type((feedPtr) node)) {
        //GSList *iter;
        //itemPtr ptr;
        /* Count the number of feeds */
        (*nr_feeds)++;

        /* Find the latest refresh time */
        if (((feedPtr) node)->lastPoll.tv_sec > *last_poll_time)
            *last_poll_time = ((feedPtr) node)->lastPoll.tv_sec;

        /* Find the number of unread posts */
        if (((feedPtr) node)->unreadCount > 0) {
            (*nr_unread) += ((feedPtr) node)->unreadCount;
            (*nr_unread_feeds)++;
        }

        /* Find the number of items saved for later reference */
        /* tvh: Why need to load feeds just for getting this? 
            Don't we already have fp->markCount? 
        */
        /*
        feed_load((feedPtr) node);
        iter = feed_get_item_list(((feedPtr) node));
        while (NULL != iter) {
            ptr = iter->data;
            if (ptr->marked)
            (*nr_marked)++;
            iter = g_slist_next(iter);
        }
        feed_unload((feedPtr) node);
        */
        /* lets use this in stead */
        (*nr_marked) += ((feedPtr) node)->markCount;

        } else if (FST_FOLDER == feed_get_type((feedPtr) node)) {
        ui_feedlist_find_mainwindow_data(&iter, nr_feeds,
                         nr_unread,
                         nr_unread_feeds,
                         nr_marked,
                         last_poll_time);
        }
    } else {
        /* CHECK: Why would we have a NULL value in the tree model? */
    }
    valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(feedstore), &iter);
    }
}

void ui_feedlist_delete_read_items(GtkTreeIter *from)
{
    GtkTreeIter iter;
    gboolean valid = FALSE;
    nodePtr node = NULL;
    feedPtr fp = NULL;

    //g_assert(NULL != feedstore);
    g_return_if_fail(NULL != feedstore);

    if (from != NULL) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), from, FS_PTR, &node,
               -1);
    valid =
        gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore), &iter,
                     from);
    } else {
    valid =
        gtk_tree_model_get_iter_first(GTK_TREE_MODEL(feedstore),
                      &iter);
    }

    while (valid) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &iter,
               FS_PTR, &node, -1);
    if (NULL != node) {
        fp = (feedPtr)node;
        
        if (FST_FEED == feed_get_type(fp)) {
                if(feed_get_unread_counter(fp) == 0 &&
           feed_get_feed_read(fp) != 0)
        {
            GTimeVal time;
            
            g_get_current_time(&time);
            
            if(time.tv_sec - feed_get_feed_read(fp) > AUTOMATIC_DELETION_TIMEOUT * 60)
            {
                feed_load(fp);
                feed_remove_unmarked(fp, FALSE);
            feed_unload(fp);
            }
        }
        } else if (FST_FOLDER == feed_get_type(fp)) {
        ui_feedlist_delete_read_items(&iter);
        }
    }
    valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(feedstore), &iter);
    }
}

nodePtr ui_feedlist_get_selected_from_tree(GtkTreeView * treeview)
{
    GtkTreeSelection *select = NULL;
    GtkTreeModel *model = NULL;
    GtkTreeIter iter;
    nodePtr ptr = NULL;

    select = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    if (gtk_tree_selection_get_selected(select, &model, &iter)) {
    gtk_tree_model_get(model, &iter, FS_PTR, &ptr, -1);
    return ptr;
    }

    return NULL;
}

folderPtr ui_feedlist_get_target_folder()
{
    nodePtr ptr = NULL;
    GtkTreeIter iter;
    GtkTreeIter parent;
    GtkTreePath *path = NULL;
    GtkTreeSelection *select = NULL;
    GtkTreeModel *model = NULL;
    
    if (NULL == (ptr = ui_feedlist_get_selected())) {
    select = gtk_tree_view_get_selection(GTK_TREE_VIEW(feedlist));
        if(!gtk_tree_selection_get_selected(select, &model, &iter)) {
            return ui_feedlist_get_root_folder();
        }
    
    if(model == NULL)
        return NULL;
    
    path = gtk_tree_model_get_path(model, &iter);
    
    if(gtk_tree_path_get_depth(path) == 1) {
        gtk_tree_path_free(path);
        return ui_feedlist_get_root_folder();
    }
    
    gtk_tree_model_iter_parent(model, &parent, &iter);
        gtk_tree_model_get(model, &parent, FS_PTR, &ptr, -1);
    
    gtk_tree_path_free(path);
    
    if(ptr == NULL)
        return NULL;
    
    if(FST_FOLDER == ptr->type)
        return (folderPtr)ptr;
    else
        return NULL;
    }

    ui_feedlist_convert_store_iter_to_visible_iter
    (&iter, &((ui_data *) (ptr->ui_data))->row);

    if (FST_FOLDER == ptr->type)
    return (folderPtr) ptr;
    else
    return ui_feedlist_get_parent(ptr);
}

void ui_feedlist_update_iter(GtkTreeIter * iter)
{
    rss_max_feed = 0;

    ui_feedlist_update_(iter);

    if (filter_feeds_without_unread_headlines)
    gtk_tree_model_filter_refilter(GTK_TREE_MODEL_FILTER(filter));

    gtk_widget_queue_draw(feedlist);
}

void rss_feed_tree_model_init(GtkTreeModelIface * iface,
                  gpointer user_data)
{
    GtkWidget *widget =
    GTK_WIDGET(gtk_tree_model_filter_new
           (GTK_TREE_MODEL(gtk_tree_store_new(1, G_TYPE_INT)),
            NULL));
    GtkTreeModelIface *originaliface =
    g_type_interface_peek(((GTypeInstance *) widget)->g_class,
                  GTK_TYPE_TREE_MODEL);
    iface->get_flags = originaliface->get_flags;
    iface->get_n_columns = originaliface->get_n_columns;
    iface->get_column_type = originaliface->get_column_type;
    iface->get_iter = originaliface->get_iter;
    iface->get_path = originaliface->get_path;
    iface->get_value = originaliface->get_value;
    iface->iter_next = originaliface->iter_next;
    iface->iter_children = originaliface->iter_children;
    iface->iter_has_child = originaliface->iter_has_child;
    iface->iter_n_children = originaliface->iter_n_children;
    iface->iter_nth_child = originaliface->iter_nth_child;
    iface->iter_parent = originaliface->iter_parent;
    iface->ref_node = originaliface->ref_node;
    iface->unref_node = originaliface->unref_node;
}

/* Sets either the unread feeds filter model or the standard
   GTK tree model. This is necessary because only the standard
   model supports drag and drop. */
void ui_feedlist_set_model(GtkTreeView * feedview,
               GtkTreeStore * feedstore, gboolean filtered)
{
    GtkTreeModel *model_without_root = NULL;
    GtkTreeModel *model = NULL;
    GtkTreePath *path = NULL;
    gboolean search_result = FALSE;

    ULOG_DEBUG(__FUNCTION__);

    search_result = ui_feedlist_search_model_by_ptr
    (GTK_TREE_MODEL(feedstore),
     (nodePtr) root_folder_ptr, &path, NULL);

    if (search_result && path != NULL) {
    model_without_root = gtk_tree_model_filter_new
        (GTK_TREE_MODEL(feedstore), path);
    } else {
    ULOG_DEBUG("root node not found");
    model_without_root = GTK_TREE_MODEL(feedstore);
    }
    if (path != NULL) {
    gtk_tree_path_free(path);
    path = NULL;
    }
    if (filtered) {
    ULOG_DEBUG("unread feed only filter is on");
    filter = gtk_tree_model_filter_new(model_without_root, NULL);

    gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER
                           (filter),
                           filter_visible_function,
                           NULL, NULL);
    model = GTK_TREE_MODEL(filter);
    } else {
    ULOG_DEBUG("unread feed only filter is off");
    model = model_without_root;
    }

    feedmodel = model;

    gtk_tree_view_set_model(GTK_TREE_VIEW(feedview), model);
}

gboolean ui_feedlist_event_cb(GtkWidget * widget, GdkEvent * event,
                  gpointer user_data)
{
    GdkEventButton *eb = NULL;
    GtkTreeView *treeview = NULL;
    GtkTreeModel *model = NULL;
    GtkTreePath *path = NULL;
    GtkTreeIter iter;
    nodePtr at_cursor = NULL;
    nodePtr selected = NULL;

    treeview = GTK_TREE_VIEW(widget);
    //g_assert(treeview != NULL);
    g_return_val_if_fail ((treeview != NULL), FALSE);

    if (event->type == GDK_BUTTON_PRESS) {
    /*ULOG_DEBUG("%s GDK_BUTTON_PRESS", __FUNCTION__);*/

    eb = (GdkEventButton *) event;
    if (eb->button == 1) {
        if (gtk_tree_view_get_path_at_pos(treeview,
                          eb->x,
                          eb->y,
                          &path, NULL, NULL, NULL)) {
        model = gtk_tree_view_get_model(treeview);
        //g_assert(model != NULL);
        g_return_val_if_fail ((model != NULL), FALSE);
        gtk_tree_model_get_iter(model, &iter, path);
        gtk_tree_model_get(model, &iter, FS_PTR, &at_cursor, -1);
        gtk_tree_path_free(path);
        path = NULL;
        selected = (nodePtr)
            ui_feedlist_get_selected();

        if (at_cursor != NULL &&
            at_cursor == selected && at_cursor->type == FST_FEED) {
            ui_feedlist_load_selected(selected);
            /*return TRUE;*/
        }
        }
    }
    }
    return FALSE;
}

gboolean ui_feedlist_cursor_changed_cb(GtkWidget * widget,
                       GdkEvent * event,
                       gpointer user_data)
{
    user_data = NULL;

    AppUIData *app_ui_data;
     
    //g_assert(app_data != NULL);
    
    
    app_ui_data = app_data->app_ui_data;

    nodePtr ptr = (nodePtr) ui_feedlist_get_selected();

    if (NULL != ptr && ptr->type == FST_FEED && app_ui_data->search_mode != SFM_SEARCH) {
        if (app_ui_data->search_mode == SFM_NORMAL)
        {
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->unsubscribe_button), TRUE);
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->feed_properties_button),
                     TRUE);
            gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_open),
                     TRUE);
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_unsubscribe),
                     TRUE);
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_properties), TRUE);
        }
        else if (get_counter(RSS_MAX) > 0)// it should be SFM_DOWNLOAD now
        {
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->unsubscribe_button), FALSE);
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->feed_properties_button),
                     TRUE);
            gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_open),
                     TRUE);
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_unsubscribe),
                     FALSE);
            gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_properties), TRUE);
        }
                     
    } else { //folder or search mode
        gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->unsubscribe_button), FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->feed_properties_button),
                     FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET(app_ui_data->menuitem_open),
                     FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_unsubscribe),
                     FALSE);
        gtk_widget_set_sensitive(GTK_WIDGET
                     (app_ui_data->menuitem_properties),
                     FALSE);
    }
    return FALSE;
}

/* sets up the entry list store and connects it to the entry list
   view in the main window */
void ui_feedlist_init(GtkWidget * feedview)
{
    GtkCellRenderer *textRenderer = NULL;
    GtkCellRenderer *iconRenderer = NULL;    
    GtkTreeViewColumn *column = NULL;
    GtkTreeSelection *select = NULL;

    //g_assert(mainwindow != NULL);
    g_return_if_fail(mainwindow != NULL);
    //g_assert(feedview != NULL);
    g_return_if_fail(feedview != NULL);
    /* Set up store */
    feedstore = gtk_tree_store_new(FS_LEN,
                   G_TYPE_STRING,
                   GDK_TYPE_PIXBUF,
                   G_TYPE_POINTER, G_TYPE_INT);

    /* Add root node */
    root_folder_ptr = create_root_folder(_("rss_ia_news_feed_folders"));
    ui_feedlist_add_new_folder_to_root((nodePtr) root_folder_ptr);

    ui_feedlist_update();

    ui_feedlist_set_model(GTK_TREE_VIEW(feedview), feedstore, FALSE);

    /*tvh: 2006May24: unref the model???*/
    g_object_unref(G_OBJECT(feedstore));
    /* we only render the state and title */
    iconRenderer = gtk_cell_renderer_pixbuf_new();
    textRenderer = gtk_cell_renderer_text_new();
    //numberRenderer = gtk_cell_renderer_text_new();
    
    /* Make the folder able to ellipzise (trunkcate with ...) the text */
    g_object_set(G_OBJECT(textRenderer), "ellipsize", PANGO_ELLIPSIZE_MIDDLE, NULL);
    gtk_cell_renderer_set_fixed_size(textRenderer, CELL_RENDERER_WIDTH, -1);

    column = gtk_tree_view_column_new();

    gtk_tree_view_column_pack_start(column, iconRenderer, FALSE);
    gtk_tree_view_column_pack_start(column, textRenderer, TRUE);
    //gtk_tree_view_column_pack_start(column, numberRenderer, FALSE);

    gtk_tree_view_column_add_attribute(column, iconRenderer, "pixbuf",
                       FS_ICON);
    gtk_tree_view_column_add_attribute(column, textRenderer, "markup",
                       FS_LABEL);
    //gtk_tree_view_column_add_attribute(column, numberRenderer, "text",
    //                   FS_UNREAD);

    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_append_column(GTK_TREE_VIEW(feedview), column);
    // g_object_set(G_OBJECT(column), "sizing", GTK_TREE_VIEW_COLUMN_FIXED, NULL);
    //gtk_tree_view_column_set_spacing (column,0);
    //gtk_tree_view_column_set_fixed_width(column, 350);

    /* And connect signals */
    /* row_activated signal doesn't get emitted under Hildon theme for some reason,
       we need to catch it at much lower level.
       g_signal_connect(G_OBJECT(feedview), "row-activated", G_CALLBACK(ui_feedlist_row_activated_cb), NULL); */
    g_signal_connect(G_OBJECT(feedview), "event",
             G_CALLBACK(ui_feedlist_event_cb), NULL);
    g_signal_connect(G_OBJECT(feedview), "row-expanded",
             G_CALLBACK(ui_folder_collapse_expand_cb), NULL);
    g_signal_connect(G_OBJECT(feedview), "row-collapsed",
             G_CALLBACK(ui_folder_collapse_expand_cb), NULL);
    g_signal_connect(G_OBJECT(feedview), "key-press-event",
             G_CALLBACK(ui_feedlist_key_press_cb), NULL);
    g_signal_connect(G_OBJECT(feedview), "cursor-changed",
             G_CALLBACK(ui_feedlist_cursor_changed_cb), NULL);

    /* Setup the selection handler for the main view */
    select = gtk_tree_view_get_selection(GTK_TREE_VIEW(feedview));
    gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);

    ui_dnd_init();
}

void ui_feedlist_select(nodePtr np)
{
    GtkTreeIter iter;
    GtkTreeView *treeview;
    GtkTreeSelection *selection = NULL;
    GtkTreePath *path = NULL;
    gint count = 0;

    ULOG_DEBUG(__FUNCTION__);

    /* some comfort: select the created iter */
    treeview = GTK_TREE_VIEW(feedlist);

    selection = gtk_tree_view_get_selection(treeview);

    if (NULL != np) {
    if (filter_feeds_without_unread_headlines) {
        /* check if the node has unread items, if not it is not in 
           the filtered model and cannot be selected */
        gtk_tree_model_get(GTK_TREE_MODEL(feedstore),
                   &((ui_data *) (np->ui_data))->row,
                   FS_UNREAD, &count, -1);
        if (0 == count)
        return;
    }

    ui_feedlist_convert_store_iter_to_visible_iter
        (&iter, &((ui_data *) (np->ui_data))->row);
    path = gtk_tree_model_get_path(feedmodel, &iter);

    gtk_tree_view_scroll_to_cell(treeview, path, NULL,
                     FALSE, 0.0, 0.0);
    gtk_tree_view_expand_to_path(treeview, path);
    gtk_tree_selection_select_path(selection, path);
    gtk_tree_path_free(path);
    } else {
    gtk_tree_selection_unselect_all(selection);
    }
}

/* function for finding next unread item */
feedPtr ui_feedlist_find_unread_feed(nodePtr folder)
{
    feedPtr fp = NULL;
    nodePtr ptr = NULL;
    GtkTreeModel *model = NULL;
    GtkTreeIter iter, iter2, *parent = NULL;
    gboolean valid = FALSE;
    gint count = 0;

    if (folder != NULL) {
    ui_feedlist_convert_store_iter_to_visible_iter
        (&iter2, &((ui_data *) (folder->ui_data))->row);
    parent = &iter2;
    }

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(feedlist));
    valid = gtk_tree_model_iter_children(model, &iter, parent);
    while (valid) {
    gtk_tree_model_get(model, &iter,
               FS_PTR, &ptr, FS_UNREAD, &count, -1);
    if (count > 0) {
        if ((FST_FEED == ptr->type) || (FST_VFOLDER == ptr->type)) {
        return (feedPtr) ptr;
        } else if (FST_FOLDER == ptr->type) {
        if (NULL != (fp = ui_feedlist_find_unread_feed(ptr)))
            return fp;
        }            /* Directories are never checked */
    }
    valid = gtk_tree_model_iter_next(model, &iter);
    }
    return NULL;
}

/*------------------------------------------------------------------------------*/
/* delete entry callbacks                             */
/*------------------------------------------------------------------------------*/

void ui_feedlist_delete_folder(folderPtr folder)
{
    ULOG_DEBUG("ui_feedlist_delete_folder...");
    ui_feedlist_delete_((nodePtr) folder);
    rss_dbus_refresh_completed();
    
    if(!app_data->app_ui_data->feed_displayed)
        ui_mainwindow_finish();
}

gboolean delete_feed_needs_sync = FALSE;
/* delete the feed, thread-safely
 * @param data: feedpointer
*/
static
gboolean delete_selected_feed_timedout_cb(gpointer data)
{
    GtkTreeSelection *selection = NULL;
    GtkTreeModel *model = NULL;
    GtkTreeIter selected_iter;
    GtkTreePath *path = NULL;

    //if the page is not yet finished. Another timeout.
    if (!app_data->app_ui_data->html_displayed )
    {
        ULOG_DEBUG("delete_selected_feed_timedout_cb");
        return TRUE;

    }
    // let's try to lock the mutex. That is. wait until the 
    // page is stopped displaying. Then we can delete it. Otherwise, a crash
    if (delete_feed_needs_sync)
        G_LOCK(load_selected_mutex);
    ULOG_DEBUG("delete_selected_feed_timeout_cb : Got a lock");

    //also a thread safe is needed if    
    if (delete_feed_needs_sync)
        gdk_threads_enter();
    
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(feedlist));

    if(gtk_tree_selection_get_selected (selection, &model, &selected_iter)) {
        /* We are going to select the next or the previous feed in the tree, so
        that the cursor doesnt jump to the top of the tree after deleting a feed */
        if(gtk_tree_model_iter_next(model, &selected_iter)) {
            gtk_tree_selection_select_iter(selection, &selected_iter);
        }
        else {
            if(gtk_tree_selection_get_selected (selection, &model, &selected_iter)) {
                path = gtk_tree_model_get_path(model, &selected_iter);
                if(gtk_tree_path_prev(path)) {
                    gtk_tree_selection_select_path(selection, path);
                } /* Checking the depth here is just paranoia */
                else if(gtk_tree_path_get_depth(path) > 1 && gtk_tree_path_up(path)) {
                    gtk_tree_selection_select_path(selection, path);
                }
                gtk_tree_path_free(path);
            }
        }
    }
    ULOG_DEBUG("ui_feedlist_delete:  Now deleting it 2 "); 
    //ui_feedlist_delete_((nodePtr) fp);
    ui_feedlist_delete_((nodePtr) data);    
    rss_dbus_refresh_completed();
    ULOG_DEBUG("ui_feedlist_delete: writing updated stats "); 
    if(!app_data->app_ui_data->feed_displayed)
        ui_mainwindow_finish();

    if (delete_feed_needs_sync)
        G_UNLOCK(load_selected_mutex);
    ULOG_DEBUG("delete_selected_feed_timeout_cb : RELEASED the  lock");

    //also a thread safe is needed if    
    if (delete_feed_needs_sync)
        gdk_threads_leave();
    
    // Now remove the timeout GSource
    return FALSE;
}

/* delete the selected node 
   Only if it's a feed
*/
gboolean ui_feedlist_delete(nodePtr ptr)
{
    /*
    GtkTreeSelection *selection = NULL;
    GtkTreeModel *model = NULL;
    GtkTreeIter selected_iter;
    GtkTreePath *path = NULL;
    */
    //g_assert(ptr != NULL);
    //g_assert(ptr->ui_data != NULL);
    //g_assert(ptr == ui_feedlist_get_selected());
    g_return_val_if_fail((ptr != NULL), FALSE);
    g_return_val_if_fail((ptr->ui_data != NULL), FALSE);
    g_return_val_if_fail((ptr == ui_feedlist_get_selected()), FALSE);

    if (ptr->type == FST_FEED) 
    {    /* Unsubscribe will only delete feeds, not folders. */
        feedPtr fp = (feedPtr) ptr;
        if (feed_get_nonremovable(fp)) 
        {
            gtk_infoprint
            (NULL, _("rss_ib_unsubscribing_feed_not_allowed"));
        } 
        else 
        {
            //g_assert(app_data != NULL);
            g_return_val_if_fail((app_data != NULL), FALSE);
            ConfirmationResponse resp = ui_show_feed_delete_confirmation
            (app_data->app_ui_data, fp->title);
            
            if (resp == CONFRESP_OK)
            {
                if (ptr == displayed_node)
                {
                    delete_feed_needs_sync = TRUE;
                    ULOG_DEBUG("ui_feedlist_delete:  Stopping the page display ");
                    gtkhtml_stop();
                    // timeout so that the page is sured finished....Look at comment for 
                    // ui_feedlist_load_selected()
                    g_timeout_add(10, delete_selected_feed_timedout_cb, ptr);
                }
                else // no need for a timeout
                {
                    delete_feed_needs_sync = FALSE;
                    delete_selected_feed_timedout_cb(ptr);
                    /* 
                    //and we can update searchview here if necessary
                   if (!displayed_node )
                        return FALSE;
                    if (displayed_node->type == FST_VFOLDER)
                    {
                        on_searchresults_menuitem_selected(NULL,NULL);
                    }
                    */
                    
                }
            }
        }
    }
//    return FALSE;
    //all feeds are delete-able
    return TRUE;
}

/*------------------------------------------------------------------------------*/
/* property dialog callbacks                             */
/*------------------------------------------------------------------------------*/

void on_popup_prop_selected(gpointer callback_data, guint callback_action,
                GtkWidget * widget)
{
    feedPtr fp = (feedPtr) callback_data;
    
    //g_assert(NULL != fp);
    g_return_if_fail(NULL != fp);
    if (NULL != fp) {
    if (FST_FEED == feed_get_type(fp)) {
        //g_assert(app_data != NULL);
        //g_assert(app_data->app_ui_data != NULL);
        ui_feed_propdialog_new(GTK_WINDOW(app_data->app_ui_data->app),
                   fp);
        return;
    }
    }

    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_no_feed_selected"));
}

/*------------------------------------------------------------------------------*/
/* new entry dialog callbacks                             */
/*------------------------------------------------------------------------------*/

void ui_feedlist_add(folderPtr parent, nodePtr node, gint position)
{
    GtkTreeIter *iter = NULL, *parentIter = NULL;

    //g_assert(node->ui_data == NULL);
    g_return_if_fail(node->ui_data == NULL);

    //ULOG_DEBUG("-----------ui_feedlist_add");
    /* if parent is NULL we have the root folder and don't create a new row! */
    node->ui_data = (gpointer) g_new0(struct ui_data, 1);
    iter = &(((ui_data *) (node->ui_data))->row);

    if (parent != NULL) {
    //g_assert(parent->ui_data != NULL);
    g_return_if_fail(parent->ui_data != NULL);
    parentIter = &(((ui_data *) (parent->ui_data))->row);
    }
    //NOTE: for now we should have always position = -1
    if (position < 0)
    gtk_tree_store_append(feedstore, iter, parentIter);
    else
    gtk_tree_store_insert(feedstore, iter, parentIter, position);

    gtk_tree_store_set(feedstore, iter, FS_PTR, node, -1);

    ui_folder_check_if_empty();
    ui_feedlist_update();
}

void ui_feedlist_remove(nodePtr ptr)
{
    //g_assert(ptr != NULL);
    g_return_if_fail(ptr != NULL);

    if ((FST_FEED == ptr->type) || (FST_VFOLDER == ptr->type)) {
    ui_notification_remove_feed((feedPtr) ptr);    /* removes an existing notification for this feed */
    ui_folder_remove_node(ptr);
    }
    ui_feedlist_update();    /* because vfolder unread counts may have changed *///does this apply here
}

void ui_feedlist_add_new_folder_to_root(nodePtr node)
{
    GtkTreeIter *iter = NULL;

    ULOG_DEBUG(__FUNCTION__);

    //g_assert(node->ui_data == NULL);
    g_return_if_fail(node->ui_data == NULL);

    /* if parent is NULL we have the root folder and don't create a new row! */
    node->ui_data = (gpointer) g_new0(struct ui_data, 1);
    iter = &(((ui_data *) (node->ui_data))->row);

    gtk_tree_store_append(feedstore, iter, NULL);

    gtk_tree_store_set(feedstore, iter, FS_PTR, node, -1);
}

void ui_feedlist_new_subscription(gchar * source,
                  const gchar * filter, gint flags,
                  folderPtr selected_parent)
{
    debug_enter("ui_feedlist_new_subscription");

    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    
    app_data->app_ui_data->iap_action = OSSO_IAP_NEW_SUBSCRIPTION;
    app_data->app_ui_data->new_subscription_flags = flags;
    app_data->app_ui_data->source = source;
    app_data->app_ui_data->filter = filter;
    
    ULOG_DEBUG("new subscription iap action: %d", app_data->app_ui_data->iap_action);
    osso_iap_connect(OSSO_IAP_ANY, OSSO_IAP_REQUESTED_CONNECT,
             app_data->app_ui_data);

    debug_exit("ui_feedlist_new_subscription");
}

void ui_feedlist_new_subscription_for_fd(const gchar * source,
                         const gchar * filter, gint flags,
                         folderPtr selected_parent, gboolean refresh, gboolean feedlist_update)
{
    feedPtr fp = NULL;
    gchar *tmp = NULL;
    int pos = -1;
    folderPtr parent = NULL;

    //g_assert(app_data != NULL);

    ULOG_DEBUG("###ui_feedlist_new_subscription_for_fd###");
    ULOG_DEBUG("\n##########RECEIVING A FILE#################");
    
    fp = feed_new();
    tmp = conf_new_id();
    feed_set_id(fp, tmp);
    g_free(tmp);

    feed_set_source(fp, source);
    feed_set_title(fp, source);
    feed_set_filter(fp, filter);
    feed_set_added(fp, time(NULL));

    parent = (selected_parent != NULL)
    ? selected_parent : ui_feedlist_get_root_folder();

    /*feed_set_available(fp, TRUE);*/    /* To prevent the big red X from being next to the new feed */
    ui_feedlist_add(parent, (nodePtr) fp, pos);
    
    if(feedlist_update)
    {
        ui_feedlist_update();
        ui_feedlist_select((nodePtr) fp);
    }

    if(refresh)
    {    
#ifdef WITHOUT_CONNECTIVITY
        feed_schedule_update(fp,
                             flags | FEED_REQ_PRIORITY_HIGH |
                 FEED_REQ_DOWNLOAD_FAVICON | FEED_REQ_AUTH_DIALOG );
#else
        gconf_client_clear_cache(app_data->gconf_client);
        if(*getStringConfValue(ICD_GCONF_CURRENT))
            feed_schedule_update(fp,
                     flags | FEED_REQ_PRIORITY_HIGH |
                     FEED_REQ_DOWNLOAD_FAVICON | FEED_REQ_AUTH_DIALOG );
#endif
    }

    ULOG_DEBUG("ui_feedlist_new_subscription_for_fd. Finished");
}

void on_newbtn_clicked_cb(GtkButton * button, gpointer user_data)
{
    GtkWidget *newdialog;
    AppUIData *app_ui_data;
     
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    if (SFM_SEARCH == app_ui_data->search_mode){
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_add_feed_searching"));
        return;
    }

    /* NOTE: downloading images also uses SFM_DOWNLOAD mode */
    if ((SFM_DOWNLOAD == app_ui_data->search_mode) &&
        (get_counter(RSS_MAX) > 0 )){
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_unable_add_feed"));
        return;
    }
    
    //g_assert(app_ui_data != NULL);
    
    newdialog = ui_feed_newdialog_new(GTK_WINDOW(app_ui_data->app), NULL);

    if(newdialog != NULL)
        gtk_widget_show(newdialog);
}

/* Used in handling "mime_open" dbus signal
   

*/
//void on_newbtn_clicked(GtkButton * button, gpointer user_data, const gchar *source)
//void  on_mime_open_add_feed_requested(AppData * app_data, const gchar * fd_source)
void  on_newbtn_clicked(AppData * appdata, const gchar * fd_source)
{
    GtkWidget *newdialog;
    AppUIData *app_ui_data;
     
    //g_assert(app_data != NULL);
    
    app_ui_data = app_data->app_ui_data;

    //g_assert(app_ui_data != NULL);
    
    newdialog = ui_feed_newdialog_new(GTK_WINDOW(app_ui_data->app), fd_source);

    if(newdialog != NULL)
        gtk_widget_show(newdialog);
    return;
}

/* recursively calls func for every feed in the feed list */
void ui_feedlist_do_for_all_full(nodePtr ptr, gint filter,
                 gpointer func, gint params,
                 gpointer user_data)
{
    GtkTreeIter childiter;
    gboolean valid = FALSE, apply = FALSE, descend = FALSE;
    nodePtr child = NULL;
    //ULOG_DEBUG("\n\nui_feedlist_do_for_all_full\n\n");
    if (NULL == ptr) {
    valid =
        gtk_tree_model_get_iter_root(GTK_TREE_MODEL(feedstore),
                     &childiter);
    } else {
    //g_assert(ptr->ui_data);
    g_return_if_fail(ptr->ui_data);
    valid =
        gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore),
                     &childiter,
                     &((ui_data *) ptr->ui_data)->row);
    }

    while (valid) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &childiter, FS_PTR,
               &child, -1);
    /* Must update counter here because the current node may be deleted! */
    valid =
        gtk_tree_model_iter_next(GTK_TREE_MODEL(feedstore),
                     &childiter);
    /* If child == NULL, this is an empty node. */
    if (child != NULL) {
        gboolean directory = (FST_FEED == child->type) &&
        (((feedPtr) child)->fhp != NULL)
        && ((feedPtr) child)->fhp->directory;
        gboolean is_root =
        (FST_FOLDER == child->type) && ((folderPtr) child)->root;
        apply = (filter & ACTION_FILTER_CHILDREN) ||
        ((filter & ACTION_FILTER_FEED)
         && (FST_FEED == child->type) &&
         !directory) || ((filter & ACTION_FILTER_FEED)
                 && (FST_VFOLDER == child->type) &&
                 !directory)
        || ((filter & ACTION_FILTER_DIRECTORY)
            && (FST_FEED == child->type) &&
            directory) || ((filter & ACTION_FILTER_FOLDER)
                   && (FST_FOLDER == child->type)
                   && !is_root) ||
        ((filter & ACTION_FILTER_ROOT)
         && is_root);
        descend = !(filter & ACTION_FILTER_CHILDREN);

        if (TRUE == apply) {
        if (params == 0)
            ((nodeActionFunc) func) (child);
        else
            ((nodeActionDataFunc) func) (child, user_data);
        }

        /* if the iter has children and we are descending, iterate over the children. */
        if (descend)
        ui_feedlist_do_for_all_full(child, filter, func, params,
                        user_data);
    }
    }
}

/* recursively calls func for every feed in the feed list */
gboolean ui_feedlist_do_for_all_full_with_int(nodePtr ptr, gint filter,
                          gpointer func, gint params,
                          gpointer user_data)
{
    GtkTreeIter childiter;
    gboolean valid = FALSE, apply = FALSE, descend = FALSE;
    nodePtr child = NULL;
    gboolean cont = FALSE;
    gboolean result = TRUE;

    //ULOG_DEBUG("ui_feedlist_do_for_all_full_with_int");
    if (NULL == ptr) {
    valid =
        gtk_tree_model_get_iter_root(GTK_TREE_MODEL(feedstore),
                     &childiter);
    } else {
    //g_assert(ptr->ui_data);
    g_return_val_if_fail((ptr->ui_data), FALSE);
    valid =
        gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore),
                     &childiter,
                     &((ui_data *) ptr->ui_data)->row);
    }

    while (valid) {
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &childiter, FS_PTR,
               &child, -1);
    /* Must update counter here because the current node may be deleted! */
    valid =
        gtk_tree_model_iter_next(GTK_TREE_MODEL(feedstore),
                     &childiter);
    /* If child == NULL, this is an empty node. */
    if (child != NULL) {
        gboolean directory = (FST_FEED == child->type)
        && (((feedPtr) child)->fhp != NULL)
        && ((feedPtr) child)->fhp->directory;
        gboolean is_root = (FST_FOLDER == child->type)
        && ((folderPtr) child)->root;
        apply = (filter & ACTION_FILTER_CHILDREN)
        || ((filter & ACTION_FILTER_FEED)
            && (FST_FEED == child->type) && !directory)
        || ((filter & ACTION_FILTER_FEED)
            && (FST_VFOLDER == child->type) && !directory)
        || ((filter & ACTION_FILTER_DIRECTORY)
            && (FST_FEED == child->type) &&
            directory) || ((filter & ACTION_FILTER_FOLDER)
                   && (FST_FOLDER == child->type)
                   && !is_root) ||
        ((filter & ACTION_FILTER_ROOT)
         && is_root);
        descend = !(filter & ACTION_FILTER_CHILDREN);
        if (TRUE == apply) {
        if (params == 0) {
            cont = ((nodeActionFuncWithResult) func)
            (child);
        } else {
            cont = ((nodeActionDataFuncWithResult)
                func)
            (child, user_data);
        }
        } else {
        cont = TRUE;
        }

        if (!cont) {
        result = FALSE;
        break;
        }

        /* if the iter has children and we are descending, iterate over the children. */
        if (descend) {
        result = ui_feedlist_do_for_all_full_with_int
            (child, filter, func, params, user_data);
        if (!result) {
            break;
        }
        }

    }
    }
    return result;
}

void set_update_timer(gboolean constant_interval)
{
    glong last_refresh = 0;    
    GTimeVal current_time;
    int interval = 0;
     
    ULOG_DEBUG("Setting update timer");

    if(!getBooleanConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES))
        return;
    
    interval = getNumericConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES_INTERVAL);

    ULOG_DEBUG("Setting update timer itnerval = %d", interval);
    
    if(interval < MIN_INTERVAL)
        interval = MIN_INTERVAL;
    
    //g_get_current_time(&current_time);    
    get_corrected_time (&current_time);
    
    last_refresh = current_time.tv_sec - rss_updated;
    ULOG_DEBUG("curent = %d rss_updated = %d", current_time.tv_sec , rss_updated);        
    interval *= 60;
    ULOG_DEBUG("last_refresh = %d, interval = %d", last_refresh, interval);        
    
    if(!constant_interval)
    {
        interval -= last_refresh;
    }
    if(interval < 0)
    {
        ULOG_DEBUG("interval = %d", interval);
        interval = 0;
    }                    
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    ULOG_DEBUG("interval finally = %d", interval);
            
    app_data->app_ui_data->automatic_update_source =
        g_timeout_add(interval * 1000, ui_feedlist_auto_update, NULL);
}

gboolean ui_feedlist_auto_update(void *data)
{

    ULOG_DEBUG("ui_feedlist_auto_update: starting due to timed interval" );
    if(!getBooleanConfValue(RSS_SETTINGS_AUTOMATIC_UPDATES) ||
       strlen(getStringConfValue(ICD_GCONF_CURRENT)) == 0)
    {
        ULOG_DEBUG("ui_feedlist_auto_update: conditions not meet, exiting..." );
        return FALSE;
    }
 
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
   
    ULOG_DEBUG("Search mode: %d SFM_NORMAL: %d", app_data->app_ui_data->search_mode, SFM_NORMAL);    
    
    //if(app_data->app_ui_data->search_mode == SFM_NORMAL) {
    if ((SFM_NORMAL == app_data->app_ui_data->search_mode) ||
        ((SFM_DOWNLOAD == app_data->app_ui_data->search_mode) &&
                (get_counter (RSS_MAX) == 0) ) ) {

        /*TODO: why clear cache here? */
        gconf_client_clear_cache(app_data->gconf_client);
/*        ui_feedlist_do_for_all(NULL, ACTION_FILTER_FEED, (gpointer)
                   ui_feedlist_check_update_counter);*/
        app_data->app_ui_data->iap_action = OSSO_IAP_REFRESH_ALL_NO_DIALOG;
        osso_iap_connect(OSSO_IAP_ANY, OSSO_IAP_REQUESTED_CONNECT,
            app_data->app_ui_data);
    }                  

    return FALSE;
}


/* Moved to top of file
G_LOCK_DEFINE_STATIC (load_selected_mutex);
G_LOCK_DEFINE_STATIC (last_node_mutex);
*/
static nodePtr _lastNode = NULL;

static
gboolean load_selected_cb(gpointer data)
{
    // if page is not yet finished. Let's wait for another timeout
    // tvh: 2006May16: This actually causes 2 timedout can get in more easily
    // hence the deadlock happens more
    /*
    if (!app_data->app_ui_data->html_displayed)
        return TRUE;
    */
    ULOG_DEBUG("Trying lock load_selected_cb: %p", g_thread_self());
    G_LOCK(load_selected_mutex);
    ULOG_DEBUG("Locked load_selected_cb: %p", g_thread_self());
    app_data->app_ui_data->html_displayed = FALSE;
    nodePtr ptr = (nodePtr) data;

    G_LOCK(last_node_mutex);
    nodePtr lastNode = _lastNode;
    G_UNLOCK(last_node_mutex);

    if (lastNode == ptr) {
        displayed_node =ptr;
    
        gdk_threads_enter();
        ui_mainwindow_set_search(FALSE);
//        gdk_threads_leave();
        
        if ((FST_FEED == ptr->type) || (FST_VFOLDER == ptr->type)) {
            //g_assert(app_data != NULL);
            //g_assert(app_data->app_ui_data != NULL);
                
            ULOG_DEBUG("\n------------ Loading feed: -------------");
            ULOG_DEBUG("---- %s ----", feed_get_title((feedPtr)ptr));
            /* Set title */
//            gdk_threads_enter();
            hildon_appview_set_title(app_data->app_ui_data->main_view,
                         feed_get_title((feedPtr) ptr));
            hildon_app_set_two_part_title(app_data->app_ui_data->app, TRUE);
//            gdk_threads_leave();
        
            ULOG_DEBUG("Get available: %d", feed_get_available((feedPtr) ptr));
            ULOG_DEBUG("Unread counter: %d", feed_get_unread_counter((feedPtr)ptr));
            
            if (!feed_get_available((feedPtr) ptr))
            {
                app_data->app_ui_data->iap_action = OSSO_IAP_REFRESH_FEED;
                app_data->app_ui_data->new_subscription = (feedPtr)ptr;
                app_data->app_ui_data->new_subscription_flags = FEED_REQ_RESET_TITLE |
                                            FEED_REQ_RESET_UPDATE_INT |
                                            FEED_REQ_AUTO_DISCOVER;
                osso_iap_connect(OSSO_IAP_ANY, OSSO_IAP_REQUESTED_CONNECT,
                  app_data->app_ui_data);
                  
                feed_load((feedPtr)ptr);
                feed_set_feed_read((feedPtr)ptr, -1);
                feed_unload((feedPtr)ptr);              
            }
            else if(feed_get_unread_counter((feedPtr)ptr) > 0)
            {
                feed_load((feedPtr)ptr);
                feed_set_feed_read((feedPtr)ptr, -1);
                feed_unload((feedPtr)ptr);
            }
//            gdk_threads_enter();
            /* workaround to ensure the feedlist is focussed when we click it
               (Mozilla might prevent this, ui_itemlist_display() depends on this */
            gtk_widget_grab_focus(feedlist);
                
            /* Set up the item list */
            ui_itemlist_load((nodePtr) ptr);
            feed_mark_all_items_read((feedPtr /*nodePtr*/) ptr);
            
            ui_feedlist_update();
            /*
            if (app_data->app_ui_data->scroll_to_feed == displayed_node &&
                app_data->app_ui_data->scroll_to_item) 
            {
                gtkhtml_scroll_to_item(app_data->app_ui_data->scroll_to_feed, 
                    app_data->app_ui_data->scroll_to_item);
                app_data->app_ui_data->scroll_to_feed = NULL;
                app_data->app_ui_data->scroll_to_item = 0;
            }
            */
//            gdk_threads_leave();
        }
        gdk_threads_leave();
    }
    app_data->app_ui_data->html_displayed = TRUE;    
    G_UNLOCK(load_selected_mutex);
    ULOG_DEBUG("Unlocked load_selected_cb: %p", g_thread_self());
    
    return FALSE;
}


//nodePtr oldptr;
/** Load the selected feed on the ui feed list on the left
  * This function will load discard the search view if search 
  * is currently going on

*/

/**
 * Thus spoke the GTK FAQ: "Callbacks require a bit of attention.
 * Callbacks from GTK+ (signals) are made within the GTK+ lock. However
 * callbacks from GLib (timeouts, IO callbacks, and idle functions) are
 * made outside of the GTK+ lock. So, within a signal handler you do not
 * need to call gdk_threads_enter(), but within the other types of
 * callbacks, you do."
 *
 * We need to set a timeout g_timeout_add to handle the loading if a selected feed.
 * This timeout lets the execution come back to the while loop in gtkhtml_write_html()
 * and exit out of the gtk_main_iteration () there.
 * within 100 milliseconds, it should be enough for that while loop to finish
 * and unlock the load_selected_mutex(). 
 * TODO: Now this might be dangerous if the while loop in gtkhtml is somehow too busy
 * and return within less than 100 milliseconds. Then we have a deadlock.!!!!
 * 
 */
 guint i = 0;

void ui_feedlist_load_selected(nodePtr ptr)
{
    ULOG_DEBUG(__FUNCTION__);

    /* If we're searching, we shouldn't let users click a feed
    as this will require
        1. Display the normal search entry with sensitive add_feed
            refresh....buttons. This is not such a good idea
      
    */
    if (app_data->app_ui_data->search_mode == SFM_SEARCH)
        return;

    if(ptr == NULL)
        return;

    //feed already displayed
    if (displayed_node==ptr)
    {
        return;
    }

    // use the timeout to avoid the threads racing with ui
    else
    {
        displayed_node = NULL;
        gtkhtml_stop();
        ULOG_DEBUG("ui_feedlist_load_selected: locking last_node_mutex");
        G_LOCK(last_node_mutex);
        _lastNode = ptr;
        G_UNLOCK(last_node_mutex);
        ULOG_DEBUG("ui_feedlist_load_selected: unlocked last_node_mutex");

        //stop first
        //gtkhtml_stop();
        
        if (i>0)
        {
            ULOG_DEBUG("TTTTTTTTTTTTTTTTTTTTTTTTTremove timeoutTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT");
            g_source_remove(i);
         }   
        ULOG_DEBUG("ui_feedlist_load_selected: into timeout now");
        i = g_timeout_add(100, load_selected_cb, ptr);
    }
    ULOG_DEBUG("ui_feedlist_load_selected: RETURNING....");
    return;
}
/**
 * ReLoad and display in itemlist the feed currently selected by treeview.
 * Used when user changes the sortOrder
 * Basically just a copy of the else part of ui_feedlist_load_selected()
 */

void ui_feedlist_reload_selected(nodePtr ptr)
{
    ULOG_DEBUG(__FUNCTION__);

    // use the timeout to avoid the threads racing with ui
    //displayed_node = NULL;
    gtkhtml_stop();
    G_LOCK(last_node_mutex);
    _lastNode = ptr;
    G_UNLOCK(last_node_mutex);
    if (i>0)
    {
        g_source_remove(i);
    }   
    i = g_timeout_add(100, load_selected_cb, ptr);
    return;
}


folderPtr ui_feedlist_get_root_folder(void)
{
    return root_folder_ptr;
}

void ui_feedlist_reset_root_folder(void)
{
    /* If root_folder_ptr is not NULL it has to be freed, too. */
    root_folder_ptr = NULL;
}


gboolean ui_feedlist_search_model_by_ptr(GtkTreeModel * model,
                     nodePtr node,
                     GtkTreePath ** path,
                     GtkTreeIter * iter)
{
    SearchData data;

    data.ptr = node;
    data.found = FALSE;
    data.path = NULL;
    gtk_tree_model_foreach(model, search_func, &data);
    if (path != NULL) {
    *path = data.path;
    }
    if (iter != NULL && data.found) {
    *iter = data.iter;
    }
    return data.found;
}


void ui_feedlist_convert_store_iter_to_visible_iter(GtkTreeIter *
                            iter_target,
                            GtkTreeIter *
                            iter_source)
{
    GtkTreeModel *child_model = NULL;
    GtkTreeIter tmp_iter;

    if (filter_feeds_without_unread_headlines) {
    child_model = gtk_tree_model_filter_get_model
        (GTK_TREE_MODEL_FILTER(feedmodel));
    gtk_tree_model_filter_convert_child_iter_to_iter
        (GTK_TREE_MODEL_FILTER(child_model), &tmp_iter, iter_source);
    gtk_tree_model_filter_convert_child_iter_to_iter
        (GTK_TREE_MODEL_FILTER(feedmodel), iter_target, &tmp_iter);
    } else {
    gtk_tree_model_filter_convert_child_iter_to_iter
        (GTK_TREE_MODEL_FILTER(feedmodel), iter_target, iter_source);
    }
}


void ui_feedlist_convert_visible_iter_to_store_iter(GtkTreeIter *
                            iter_target,
                            GtkTreeIter *
                            iter_source)
{
    GtkTreeModel *child_model = NULL;
    GtkTreeIter tmp_iter;

    if (filter_feeds_without_unread_headlines) {
    child_model = gtk_tree_model_filter_get_model
        (GTK_TREE_MODEL_FILTER(feedmodel));
    gtk_tree_model_filter_convert_iter_to_child_iter
        (GTK_TREE_MODEL_FILTER(feedmodel), &tmp_iter, iter_source);
    gtk_tree_model_filter_convert_iter_to_child_iter
        (GTK_TREE_MODEL_FILTER(child_model), iter_target, &tmp_iter);
    } else {
    gtk_tree_model_filter_convert_iter_to_child_iter
        (GTK_TREE_MODEL_FILTER(feedmodel), iter_target, iter_source);
    }
}


gint ui_feedlist_count_feeds(folderPtr folder)
{
    gint count = 0;
    //g_assert(folder != NULL);
    g_return_val_if_fail ((folder != NULL), 0);

    ui_feedlist_do_for_all_full((nodePtr) folder,
                ACTION_FILTER_FEED,
                count_feeds_func, 1, &count);
    return count;
}


gboolean ui_feedlist_confirm_and_delete_folder(folderPtr folder)
{
    GtkTreeIter *iter = NULL;
    GtkTreeIter child_iter;
    nodePtr child_node = NULL;
    gint child_count = 0;
    gint n_feeds = 0;
    gboolean do_delete = FALSE;
    ConfirmationResponse resp = CONFRESP_ERROR;
    gboolean result = FALSE;

    //g_assert(folder != NULL);
    g_return_val_if_fail ((folder != NULL), FALSE);

    if (folder_get_nonremovable(folder)) {
    /*gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), _("rss_ib_deleting_folder_not_allowed"));*/
    gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app), dgettext("hildon-common-strings", "ckct_ib_unable_to_delete"));
    } else {
    iter = &((ui_data *) (folder->ui_data))->row;
    //g_assert(iter != NULL);
    g_return_val_if_fail((iter != NULL), FALSE);
    child_count = gtk_tree_model_iter_n_children
        (GTK_TREE_MODEL(feedstore), iter);
    gtk_tree_model_iter_children(GTK_TREE_MODEL(feedstore),
                     &child_iter, iter);
    gtk_tree_model_get(GTK_TREE_MODEL(feedstore), &child_iter,
               FS_PTR, &child_node, -1);
    /* An empty folder should have the "No feeds" element. */
    if (child_count >= 2 || (child_count == 1 && child_node != NULL)) {
        //g_assert(app_data != NULL);
        n_feeds = ui_feedlist_count_feeds(folder);
        ULOG_DEBUG("Feed count: %d", n_feeds);
        resp = ui_show_folder_delete_confirmation
        (app_data->app_ui_data, n_feeds);
        do_delete = (resp == CONFRESP_OK);
    } else {
        do_delete = TRUE;
    }
    if (do_delete) {
        ULOG_INFO("Selected folder for delete: %s", folder->title);
        ui_feedlist_delete_folder(folder);
        result = TRUE;
    }
    }
    return result;
}

void ui_feedlist_update_applet()
{
    gint nr_feeds = 0;
    gint nr_unread = 0;
    gint nr_unread_feeds = 0;
    gint nr_marked = 0;
    glong last_poll_time = 0;
    
    ui_feedlist_find_mainwindow_data(NULL, &nr_feeds, &nr_unread,
                     &nr_unread_feeds, &nr_marked,
                     &last_poll_time);

    rss_unread = (long)nr_unread;
    
    if(rss_updated < (time_t)last_poll_time)
         rss_updated = (time_t)last_poll_time;

    general_save();
    rss_dbus_signal_applet(OSSO_RSS_FEED_READER_FOLDER_READ_SIGNAL);
}
