/**
 * @file update.c feed update request processing
 *
 * Copyright (C) 2003, 2004 Lars Lindner <lars.lindner@gmx.net>
 * Copyright (C) 2004 Nathan J. Conrad <t98502@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <sys/wait.h>
#include <osso-log.h>
#include "support.h"
#include "debug.h"
#include "update.h"
#include "conf.h"
#include "ui_queue.h"
#include "ui_mainwindow.h"
#include "net/downloadlib.h"
//#include <gtkhtml/gtkhtml-stream.h>

#define DEFAULT_UPDATE_THREAD_CONCURRENCY    3 

/* communication queues for requesting updates and sending the results */
GAsyncQueue *requests_high_prio = NULL;
GAsyncQueue *requests_normal_prio = NULL;
GAsyncQueue *requests_image = NULL;
GAsyncQueue *results = NULL;

static gboolean results_thread_done = FALSE;
/* keep those requests that are being processed so that download_cancel_all
 * can deal with it properly.
 */
 /*TODO: thread_is_active can be removed */
/* We can use struct request * active_threads[] in stead of those 2*/
static gboolean thread_is_active[DEFAULT_UPDATE_THREAD_CONCURRENCY];
static gboolean active_thread_cancelled[DEFAULT_UPDATE_THREAD_CONCURRENCY];

//gint rss_threads = 0;
static gint num_pending_requests = 0;
static gint num_downloaded_requests = 0;

extern AppData *app_data;

/* condition mutex for offline mode */
//static GMutex *cond_mutex = NULL;
//static GCond *offline_cond = NULL;
//static gboolean online = TRUE;

/* Condition and mutex for if_there_is_requests signal*/
static GCond *download_cond = NULL;
static GMutex *download_cond_mutex = NULL;

/* Make sure no 2 threads enter the loop where there's only one request
  It's also used in the case of download_cancel_all
  As in results_queue_mutex below 
  */
static GMutex *pending_check_mutex = NULL;
/* Control the proper checking of number of download requests */
static GMutex *num_downloaded_check_mutex = NULL;

/* Used in the case where user wants to stop
 * This will be used to stop :
     1. results thread from processing more queue
     2. download thread from pushing active requests to results queue
 * so that download_cancel_all() can work without the worry that it has cancelled 
 * all, but later on, download threads still push the active requests to results queue
*/
static GMutex *results_queue_mutex = NULL;

/* Only one thread should access this list at a time
in case we want to delete the image requests in the middle
This is needed when we refresh all the feeds ,yet we have
 image feed displayed, hence image requests are pending
 behind feed requests.

 TODO: actually if this is used, requests_image can be a normal
    queue or list, not a GAsyncQueue
*/
static GMutex *image_list_mutex = NULL;

/* results queue. Proceeds only if there's signal from download threads that
 * there is some requests downloaded */
static GCond *results_queue_cond = NULL;
static GMutex *results_queue_cond_mutex = NULL;

/* Mutex for updating counters, such as rss_max_download */
//static GMutex *counter_mutex = NULL;

static gboolean program_quitting = FALSE;
//static gboolean downloading = FALSE;
/* Let's force threads to join */
static GThread *update_threads[DEFAULT_UPDATE_THREAD_CONCURRENCY + 1];
extern GThread *mainThread;
/* prototypes */
static void *download_thread(void *data);
static void on_download_finished(void);


/************************************************************************/
/*                        UTILITY FUNCTIONS                             */
/************************************************************************/
/* read the counter . It needs to be atomic :( */
gint get_counter(int type)
{
    switch (type) {
    case IMG_CURRENT:
        return g_atomic_int_get( (volatile gint *)
            &(app_data->app_ui_data->images_current_download));
    case IMG_MAX:
        return g_atomic_int_get( (volatile gint *)
            &(app_data->app_ui_data->images_max_download));
    case RSS_CURRENT:
        return g_atomic_int_get( (volatile gint *)
            &(app_data->app_ui_data->rss_current_download));
    case RSS_MAX:
        return g_atomic_int_get( (volatile gint *)
            &(app_data->app_ui_data->rss_max_download));
    case NUM_PENDING:
        return g_atomic_int_get( (volatile gint *) &num_pending_requests);
    case NUM_DOWNLOADED:
        return g_atomic_int_get( (volatile gint *) &num_downloaded_requests);
       
    default:
        ULOG_DEBUG("oooooooooooooooooooooERRORooooooooooooooooooooooooooooo");            
        return 0;
    }
    return 0;    
}
/* add the counter by val
 * val can be positive/negative
*/

static void add_counter(int type, gint val)
{
    switch (type) {
    case IMG_CURRENT:
        g_atomic_int_add( (volatile gint *)
            &(app_data->app_ui_data->images_current_download), val);return;
    case IMG_MAX:
        g_atomic_int_add( (volatile gint *)
            &(app_data->app_ui_data->images_max_download), val);return;
    case RSS_CURRENT:
        g_atomic_int_add( (volatile gint *)
            &(app_data->app_ui_data->rss_current_download), val);return;
    case RSS_MAX:
        g_atomic_int_add( (volatile gint *)
            &(app_data->app_ui_data->rss_max_download), val);return;
    case NUM_PENDING:
        g_atomic_int_add( (volatile gint *)
            & (num_pending_requests), val);return;
    case NUM_DOWNLOADED:
        g_atomic_int_add((volatile gint *)
            & (num_downloaded_requests), val);return;
    default:
        ULOG_DEBUG("oooooooooooooooooooooERRORooooooooooooooooooooooooooooo");            
        return;
    }
    return;
}

/* set the counter to val
*/
/*
static void set_counter(int type, gint val)
{
    switch (type) {
    case IMG_CURRENT:
        g_atomic_int_set( (volatile gint *)
            &(app_data->app_ui_data->images_current_download), val);return;
    case IMG_MAX:
        g_atomic_int_set( (volatile gint *)
            &(app_data->app_ui_data->images_max_download), val);return;
    case RSS_CURRENT:
        g_atomic_int_set( (volatile gint *)
            &(app_data->app_ui_data->rss_current_download), val);return;
    case RSS_MAX:
        g_atomic_int_set( (volatile gint *)
            &(app_data->app_ui_data->rss_max_download), val);return;
    case NUM_PENDING:
        g_atomic_int_set( (volatile gint *)
            & (num_pending_requests), val);return;
        
    case NUM_DOWNLOADED:
        g_atomic_int_set( (volatile gint *)
            & (num_downloaded_requests), val);return;

    default:
        ULOG_DEBUG("oooooooooooooooooooooERRORooooooooooooooooooooooooooooo");  
        return;          
    }
    return;
}
*/

/************************************************************************/
/*                        PRIVATE FUNCTIONS                             */
/************************************************************************/

/** This is needed during long downloads to handle
  * user input and to draw stuff to the screen
  */
static void download_wait_for_pending_events()
{
    gdk_threads_enter();
    while (gtk_events_pending ())
    gtk_main_iteration ();
    gdk_threads_leave();
}

static gboolean process_progressbar_idle_cb(gpointer data)
{


    gboolean refresh_finished = FALSE;
    gboolean receiving_images_finished = FALSE;
    //ULOG_DEBUG("''''''''''''''''''''''PROGRESSBAR IDLE'''''''''''''''''''''''''''''''''");
    //ULOG_DEBUG("testing the g_atmoic get_couter RSS_CURRENT/RSS_MAX = %d : %d",
    //             get_counter(RSS_CURRENT), get_counter(RSS_MAX));
    
    if ( get_counter(RSS_MAX)  == 0)
    {
        refresh_finished = TRUE;
    }
   
    //don't have to check 
    if (get_counter(IMG_MAX) ==0)
        receiving_images_finished = TRUE;
    
    // If there are feeds downloaded.
    if( (get_counter(RSS_MAX) > 0) && (app_data->app_ui_data->search_mode == SFM_DOWNLOAD))
    {
        update_progress_bar(PROGRESSBAR_REFRESH_MODE);
        //If displaying the statistics, then refresh stat window too
        if ( displayed_node == NULL )
            ui_mainwindow_finish();
    
        /* TODO: tvh 2006May09: this might happen even if there are feeds to download
                if downloading is SOOOO fast that main thread hasn't been 
                able to increase rss_max_download yet
        */
        if (get_counter(RSS_CURRENT) == get_counter (RSS_MAX) )
        {
            refresh_finished = TRUE;
        }
    }

    else if (app_data->app_ui_data->search_mode == SFM_DOWNLOAD) { 
        // no feeds pending, we can move to IMAGE_MODE for the progress bar
        //also update the downloading_images boolean value here
        if (show_receiving_images_progressbar_button() == TRUE)
        {
            update_progress_bar(PROGRESSBAR_IMAGE_MODE);
        }
        if (get_counter (IMG_CURRENT) == get_counter (IMG_MAX))
        {
            receiving_images_finished = TRUE;
        }

    }
    //sleep(1);
    // Ok done with with the progress bar 
    // check for interrupt signal
    // This can be done in the handle_stopped_by_user
    
    if (app_data->app_ui_data->search_mode == SFM_INTERRUPTED)
       { 
        //ULOG_DEBUG("results_thread__________________SFM_INTERRUPTED_________________results_thread");
        //should signal applet here first
        if (get_counter(RSS_MAX) > 0 )
            {
                //ULOG_DEBUG("updating system is signalling applet now");
                on_refresh_finished(FALSE, FALSE); //banner & signal applet
            }
         
        download_cancel_all();
        return FALSE;
        //TODO: ???
        //refresh_finished = TRUE;
        //receiving_images_finished = TRUE;
    } 
    //having this here might invoke other results_thread process. 
    //which will call another process_download_idle_cb, making it as if this func 
    //were being  kinda executed in parallel
    //download_wait_for_pending_events();

    //only either refresh_finished or receiving_images_finished is called
     // at a time, IF it's not a SFM_INTERRUPTED.
    if ((refresh_finished) && get_counter(RSS_MAX) > 0){
        //signal the applet
        on_refresh_finished(TRUE, FALSE);
    }
    if (receiving_images_finished){
        // same as on_refresh_finished (TRUE, TRUE) here
        on_receiving_images_finished();
    }
    if ((get_counter(NUM_PENDING) == 0) && (get_counter(NUM_DOWNLOADED) == 0) &&
                    refresh_finished && receiving_images_finished)
    {
        //here we need to ui_update so that favicons can be added.
        //download_wait_for_pending_events();
        on_download_finished();
    }
    if(app_data->app_ui_data->network_disconnected)
    {
        if (get_counter(RSS_MAX) > 0)
            on_refresh_finished(TRUE, FALSE);
        else //TODO: another banner id for "download images cancelled" or something
            on_refresh_finished(FALSE, FALSE);

        app_data->app_ui_data->network_disconnected = FALSE;
        download_cancel_all();
    }
    return FALSE;
}

/* This func is needed so that it will be executed 
 * in the main loop. Therefore it's threads-safe
 */
static
gboolean process_downloaded_idle_cb(gpointer data)
{
    //ULOG_DEBUG("mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmprocess_download_idle_cbmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm");
    //ULOG_DEBUG("process_downloaded_idle_cb: %p", g_thread_self());
    gboolean refresh_finished = FALSE;
    gboolean receiving_images_finished = FALSE;
    struct request * request = (struct request *) data;
    //g_assert(request != NULL);
    g_return_val_if_fail((request != NULL), FALSE);
    
    if (app_data->app_ui_data->download_stopped)
    {
    
        g_mutex_unlock(num_downloaded_check_mutex);
        
        download_request_free(request, TRUE); /* request Processed */
        return FALSE;
    }
    if (NULL != request->callback)
        (request->callback) (request);
    
//    download_request_free(request, TRUE); /* request Processed */


    /** This is needed during long downloads to handle
        user input and to draw stuff to the screen
    */
    //having this here would leave the progressbar almost stopped. Shouldn't do this
    //download_wait_for_pending_events();

   //update counters
    //ULOG_DEBUG("........................................1...................................");
    //ULOG_DEBUG("NUM_DOWNLOADED is %d being reduced by 1", get_counter(NUM_DOWNLOADED));
    add_counter (NUM_DOWNLOADED, -1);
    //ULOG_DEBUG("NUM_DOWNLOADED is now %d ", get_counter(NUM_DOWNLOADED));
    
    //now that we 've decreased it properly, we can release the lock
    // so results thread can go on
    g_mutex_unlock(num_downloaded_check_mutex);

    if (request->type_image)
        add_counter(IMG_CURRENT,1);
    else 
        add_counter(RSS_CURRENT,1);

    download_request_free(request, TRUE); /* request Processed */

    //now if a new request, the favicon request has been put into a g_idle , HIGH PRIO
    //if we continue executing here, we will kinda ignore that new favicon requesst
    //when checking for condition to finish refresh or not
    //so we put in g_idle with DEFAULT PRIO again, so that favicon is in deed added before 
    //we update the progressbar
    //ULOG_DEBUG("....................... into idle for progressbar now............................");
    g_idle_add_full(G_PRIORITY_DEFAULT_IDLE, process_progressbar_idle_cb, data, NULL);     
    return FALSE;                   

/*
    ULOG_DEBUG("testing the g_atmoic get_couter RSS_MAX = %d", get_counter(RSS_MAX));
    if ( get_counter(RSS_MAX)  == 0)
    {
        ULOG_DEBUG("........................................3...................................");
        refresh_finished = TRUE;
    
    }
        ULOG_DEBUG("........................................4...................................");
    
    //don't have to check 
    if (get_counter(IMG_MAX) ==0)
        receiving_images_finished = TRUE;
    
    // If there are feeds downloaded.
    if( (get_counter(RSS_MAX) > 0) && (app_data->app_ui_data->search_mode == SFM_DOWNLOAD))
    {
            ULOG_DEBUG("........................................5...................................");

        update_progress_bar(PROGRESSBAR_REFRESH_MODE);
                ULOG_DEBUG("........................................6...................................");

        //If displaying the statistics, then refresh stat window too
        if ( displayed_node == NULL )
            ui_mainwindow_finish();
    
        // TODO: tvh 2006May09: this might happen even if there are feeds to download
        //        if downloading is SOOOO fast that main thread hasn't been 
        //        able to increase rss_max_download yet
        
        if (get_counter(RSS_CURRENT) == get_counter (RSS_MAX) )
        {
            refresh_finished = TRUE;
                    ULOG_DEBUG("........................................7...................................");

        }
    }

    else if (app_data->app_ui_data->search_mode == SFM_DOWNLOAD) { 
        // no feeds pending, we can move to IMAGE_MODE for the progress bar
        //also update the downloading_images boolean value here
        if (show_receiving_images_progressbar_button() == TRUE)
        {
            update_progress_bar(PROGRESSBAR_IMAGE_MODE);
                    ULOG_DEBUG("........................................8...................................");
        }
        if (get_counter (IMG_CURRENT) == get_counter (IMG_MAX))
        {
                ULOG_DEBUG("........................................9...................................");

            receiving_images_finished = TRUE;
        }

    }
    //sleep(1);
    // Ok done with with the progress bar 
    // check for interrupt signal
    // This can be done in the handle_stopped_by_user
    
    if (app_data->app_ui_data->search_mode == SFM_INTERRUPTED) { 
        ULOG_DEBUG("results_thread__________________SFM_INTERRUPTED_________________results_thread");
        //should signal applet here first
        if (get_counter(RSS_MAX) > 0 )
            {
                ULOG_DEBUG("updating system is signalling applet now");
                on_refresh_finished(FALSE, FALSE); //banner & signal applet
            }
         
        download_cancel_all();
        return FALSE;
        //TODO: ???
        //refresh_finished = TRUE;
        //receiving_images_finished = TRUE;
    } 
    //having this here might invoke other results_thread process. 
    //which will call another process_download_idle_cb, making it as if this func 
    //were being  kinda executed in parallel
    //download_wait_for_pending_events();

    //only either refresh_finished or receiving_images_finished is called
     // at a time, IF it's not a SFM_INTERRUPTED.
    if (refresh_finished) {
        ULOG_DEBUG("........................................10...................................");
    
        //signal the applet
        on_refresh_finished(TRUE, FALSE);
    }
    if (receiving_images_finished){
        ULOG_DEBUG("........................................11...................................");
        // same as on_refresh_finished (TRUE, TRUE) here
        on_receiving_images_finished();
    }
        ULOG_DEBUG("........................................12...................................");    
    if ((get_counter(NUM_PENDING) == 0) && (get_counter(NUM_DOWNLOADED) == 0) &&
                    refresh_finished && receiving_images_finished)
    {
                ULOG_DEBUG("........................................13...................................");
                //here we need to ui_update so that favicons can be added.
                download_wait_for_pending_events();
        
        on_download_finished();
    }
            ULOG_DEBUG("........................................14...................................");
    if(app_data->app_ui_data->network_disconnected)
    {
            ULOG_DEBUG("........................................15...................................");
        if (get_counter(RSS_MAX) > 0)
            on_refresh_finished(TRUE, FALSE);
        else //TODO: another banner id for "download images cancelled" or something
            on_refresh_finished(FALSE, FALSE);
        ULOG_DEBUG("........................................16...................................");        
        app_data->app_ui_data->network_disconnected = FALSE;
        download_cancel_all();
        ULOG_DEBUG("........................................17...................................");        
    }
            ULOG_DEBUG("........................................18...................................");
    return FALSE;

*/
}

/* tvh: 2006May09: TODO: need this? */
/** Filter idea (and some of the code) was taken from Snownews 
  *
  * @param cmd
  * @param data
  * @param size
  * @return 
  */
static char *filter(gchar * cmd, gchar * data, size_t * size)
{
    int fd = 0;
    gchar *command = NULL;
    const gchar *tmpdir = g_get_tmp_dir();
    char *tmpfilename = NULL;
    char *out = NULL;
    FILE *file = NULL, *p = NULL;

    tmpfilename =
    g_strdup_printf("%s" G_DIR_SEPARATOR_S "liferea-XXXXXX", tmpdir);

    fd = g_mkstemp(tmpfilename);

    if (fd == -1) {
    odebug1("Error opening temp file %s to use for filtering!",
        tmpfilename);
    g_free(tmpfilename);
    return NULL;
    }


    file = fdopen(fd, "w");
    fwrite(data, strlen(data), 1, file);
    fclose(file);

    *size = 0;
    command = g_strdup_printf("%s < %s", cmd, tmpfilename);
    p = popen(command, "r");
    g_free(command);
    if (NULL != p) {
    while (!feof(p) && !ferror(p)) {
        size_t len;
        out = g_realloc(out, *size + 1025);
        len = fread(&out[*size], 1, 1024, p);
        if (len > 0)
        *size += len;
    }
    pclose(p);
    out[*size] = '\0';
    }
    /* Clean up. */
    unlink(tmpfilename);
    g_free(tmpfilename);
    return out;
}


/* Once all the requests are downloaded & processed, we now update the ui
*/
static void on_download_finished(void)
{
    ULOG_DEBUG("on_download_finished");
    //finish all the progressbar idle_cb first
    //ui_update();
    //download_wait_for_pending_events();
    restore_search_button();
    hide_progressbar_if_needed();
    //TODO: Set the counters here 
    //app_data->app_ui_data->receiving_feeds = FALSE;

    /* Signal applet if nec */
    
}

/* Get a request out of the three queues.
 * Conditions must have been assured so that there must be some requests out there
 * in one of the three queues 
 * @param data : The thread number, only used for debugging
*/
static struct request * get_request_to_download(void * data)
{
    struct request * request;

    /* Error case */
    if (num_pending_requests == 0)
        return NULL;
    request = g_async_queue_try_pop(requests_high_prio);
    if (request == NULL)
    {
        request = g_async_queue_try_pop(requests_normal_prio);
        if (request == NULL)
        {
            // **NOTE** If page switched, we have cancelled the image requests
            // so this might return NULL
            g_mutex_lock(image_list_mutex);
            request = g_async_queue_try_pop(requests_image);
            g_mutex_unlock(image_list_mutex);
        }
    }
    // now that we have the request, let's decrease the counter, atomically
    if (request != NULL)
        add_counter(NUM_PENDING, -1);
    return request;

}

/* Handle the request by
 * Check if it should be downloaded.
 * If not, free it, if yes, download it and wait until the download is done
 * If successfully downloaded, move it to results async queue
 * @param request : the request popped out of the three queues
 * @param data : The thread number, only used for debugging
 */

static void process_and_download_request(struct request * request, void * data)
{
    //g_assert(NULL!= request);
    g_return_if_fail(NULL!= request);
    gint i = GPOINTER_TO_INT(data);
    thread_is_active[i] = TRUE;

    if (request->callback == NULL){
        //ULOG_DEBUG("process_and_download_request: THREAD NUMBER %d freeing cancelled request (%s)",
        //       GPOINTER_TO_INT(data), request->source);
        download_request_free(request, FALSE);
    }
    else {

        //download and wait till it comes back. This will block
        download_process(request);
        //ULOG_DEBUG("process_and_download_request: Downloaded! THREAD %d request->source = %s",GPOINTER_TO_INT(data),  request->source);
        //sleep(4);
        //push download to the results queue for after-download processing
        
        /*this can block in 2 cases, mostly useful in case 2: 
            1. while results_thread is checking it. Very rarely happens
            2. When user has cancelled download. And download_cancel_all is ongoing
        */
        g_mutex_lock(results_queue_mutex);
        if (active_thread_cancelled[i])
        {
            // it must have been cancelled by download_cancel_all. So just remove it.
            // no need to set counters here. As it only happens in the case
            // of cancelling all requests. TODO: is it really so?
            g_mutex_unlock(results_queue_mutex);
            download_request_free(request, FALSE);
            thread_is_active[i] = FALSE;
            active_thread_cancelled[i] = FALSE;
            return;
        }    
        if(app_data != NULL &&
               app_data->app_ui_data != NULL /* &&
               app_data->app_ui_data->search_mode == SFM_DOWNLOAD*/)
        {
            g_async_queue_push(results, (gpointer) request);
            /* increase the number of requests in results queue */
            //g_mutex_lock(num_downloaded_check_mutex);
            add_counter(NUM_DOWNLOADED,1);
            //pushed. No reason why should keep this lock any more.
            g_mutex_unlock(results_queue_mutex);

            //g_cond_signal the threads here for available to process_results
            g_mutex_lock(results_queue_cond_mutex);
            g_cond_signal(results_queue_cond);
            g_mutex_unlock(results_queue_cond_mutex);
        }
        else if (request != NULL) 
        {
            //ULOG_DEBUG("process_and_download_request: THREAD %d freeing request due to search_mode is not SFM_DOWNLOAD", GPOINTER_TO_INT(data));
            //handle_download_stopped_by_user will do this
            g_mutex_unlock(results_queue_mutex);            
            download_request_free(request, FALSE);
        }
    }
    thread_is_active[i] = FALSE;
    active_thread_cancelled[i] = FALSE;
    return;
}

/* Process the downloaded request gotten from results queue
 * Does 2 things: call the nec callbacks & update progressbar
 * @param data : The thread number, only used for debugging
*/
//guint j = 0;
static void process_downloaded_request(struct request * request, void * data)
{


    //ULOG_DEBUG("process_download_request: thread Number %d",  GPOINTER_TO_INT(data));    
    //j = 
    //g_idle_add_full(G_PRIORITY_HIGH_IDLE, process_downloaded_idle_cb, request,NULL);
    g_idle_add_full(G_PRIORITY_DEFAULT_IDLE, process_downloaded_idle_cb, request,NULL);

    return ;
}

/** Download requests are processed here. High priority requests are
  * processed first
  * 
  * @param data : The thread number, only used for debugging
  */
static void *download_thread(void *data)
{
    struct request *request = NULL;
    //gboolean high_priority = (gboolean) data;

    for (;!program_quitting;) {
        //ULOG_DEBUG("download_thread: Thread NUMBER %d entering the for loop", GPOINTER_TO_INT(data));
        /* only one thread should have access to num_pending_requests checking */
        g_mutex_lock(pending_check_mutex);
        //ULOG_DEBUG("download_thread: Thread NUMBER %d got a lock on pending_check_mutex", GPOINTER_TO_INT(data));
        if (program_quitting)
        {   //unlock so others threads can exit too
            //ULOG_DEBUG("download_thread: Thread NUMBER %d EXITING the for loop", GPOINTER_TO_INT(data));
            g_mutex_unlock(pending_check_mutex);
            return NULL;
        }
        /* Block if there is nothing to download 
            Only proceed if there is 
        */
        //if (num_pending_requests > 0){
        if (get_counter(NUM_PENDING) > 0){
            //ULOG_DEBUG("download_thread: Thread NUMBER %d trying to get a request", GPOINTER_TO_INT(data));
            request = get_request_to_download(data);
            // commented out due to image requests cancelling case
            // TODO: This is bad. Should assert it here
            //g_assert(request!= NULL);
            //ULOG_DEBUG("download_thread: Thread NUMBER %d GOT another request %s",
            //                                         GPOINTER_TO_INT(data), request->source);

            //unlock pending_check_mutex , so that other threads can go on checking|waiting g_cond
            //because now the num_pending_request has been decreased properly
            // so it can never be the case that there's only 1 request, but 2 threads can enter here
            g_mutex_unlock(pending_check_mutex);
            // Now PROCESS the request
            if (request != NULL)
                process_and_download_request(request, data);
        }
        else {
            //also unlock the pending_check_mutex so that other threads
            // can enter the checking...
            g_mutex_unlock(pending_check_mutex);
            //NUM_PENDING is 0, so we will just wait...
            // Then wait here until there is something to download
            //what if the g_cond_signal in download_queue() was lost??????
            g_mutex_lock(download_cond_mutex);
            //ULOG_DEBUG("\n\ndownload_thread:Thread NUMBER %d \
<<<<<<WAITING FOR THE download_cond\n\n", GPOINTER_TO_INT(data));
            g_cond_wait(download_cond, download_cond_mutex);
            g_mutex_unlock(download_cond_mutex);

            //ULOG_DEBUG("\n\ndownload_thread:Thread NUMBER %d >>>>>>done WAITING FOR THE download_cond\n\n", GPOINTER_TO_INT(data));
        }

    } //end for(;;)
    //ULOG_DEBUG("download_thread: Thread NUMBER %d  RETURNING...................................", GPOINTER_TO_INT(data));
    return NULL;
}


/*  Thread update_threads[DEFAULT_UPDATE_THREAD_CONCURRENCY] will loop in this func
    in order to update the ui. Only one thread is accessing this func
    Will wait until num_downloaded_requests is greater than 0
*/
void *results_thread(void *data)
{
    //gboolean refresh_finished = FALSE;
    //gboolean receiving_images_finished = FALSE;
    GTimeVal cur_time;
    
    for (;!program_quitting;)
    {
        //ULOG_DEBUG("results_thread %d",  GPOINTER_TO_INT(data));
        //if (num_downloaded_requests > 0)
        //lock the checking of this varialbe, as now we use g_idle to process
        // the downloaded request so this results thread might be checking this
        // for multiple times

        g_mutex_lock(num_downloaded_check_mutex);
        //ULOG_DEBUG("results_thread got the lock on num_downloaded_check_mutex");
        if (get_counter (NUM_DOWNLOADED) > 0)
        {

            //get a lock first before trying it. Because cancelling downloads
            //might have locked this, preventing this results thread to proceed.
            //ULOG_DEBUG("2");
            g_mutex_lock(results_queue_mutex);
            //ULOG_DEBUG("results_thread got the lock on results_queue_mutex");
            //ULOG_DEBUG("3");
            struct request *request = NULL;
            //ULOG_DEBUG("4");
            //Will not block
            request = g_async_queue_try_pop(results);
            g_mutex_unlock(results_queue_mutex);
            //ULOG_DEBUG("results_thread unlocked results_queue_mutex");
                        
            //ULOG_DEBUG("results_thread: ***RESULTS THREAD**** %d num_downloaded = %d",  GPOINTER_TO_INT(data), get_counter(NUM_DOWNLOADED));
            if (request == NULL)
            {
                // this shouldn't happen at all
                g_mutex_unlock(num_downloaded_check_mutex);
                //ULOG_DEBUG("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
                //ULOG_DEBUG("this thread must have tried to pop a request from results\n\
                //          where the results queue has already been cancelled, yet num_downloaded_requests\n\
                //        hasn't been set to 0 yet.\n");
                //ULOG_DEBUG("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
                //arghh...hackish
                //num_downloaded_requests = 0;
            }
            else
            {
                //num_downloaded mutex will be unlocked inside here
                //ULOG_DEBUG("results_thread:  ***RESULTS THREAD**** %d let's process the request", GPOINTER_TO_INT(data));
                process_downloaded_request(request, data);
                //ULOG_DEBUG("results_thread : processed the request");
            }
        }

        else {
            //unlock this so it can go to the next loop
            g_mutex_unlock(num_downloaded_check_mutex);
            //or we can just sleep here and loop forever
            //sleep(2)
            g_mutex_lock(results_queue_cond_mutex);
            //should only timed wait, coz signal from download threads might be lost
            //right after results thread has checked for the NUM_DOWNLOADED condition
            g_get_current_time(&cur_time);
            cur_time.tv_sec = cur_time.tv_sec + 10;
            g_cond_timed_wait(results_queue_cond, results_queue_cond_mutex, &cur_time);
            //g_cond_wait(results_queue_cond, results_queue_cond_mutex);
            g_mutex_unlock(results_queue_cond_mutex);
        }
    }//end for

    //ULOG_DEBUG("results_thread:   ***RESULTS THREAD****  %d RETURNING...............................", GPOINTER_TO_INT(data));
    results_thread_done = TRUE;
    return NULL;
}


/************************************************************************/
/*                        PUBLIC FUNCTIONS                              */
/************************************************************************/

void download_process(struct request *request)
{
    FILE *f;
    int status = 0;
    //g_assert(request != NULL);
    g_return_if_fail(request != NULL);
    
    request->data = NULL;
    /*tvh 2006May10 fake it here for sbox faking of requests.to be removed */
    //sleep(1);
    //return;

    if (request->source[0] == '|') {
    /* if the first char is a | we have a pipe else a file */
    request->size = 0;
    f = popen(&request->source[1], "r");
    if (NULL != f) {
        while (!feof(f)) {
        size_t len;
        request->data =
            g_realloc(request->data, request->size + 1025);
        len = fread(&request->data[request->size], 1, 1024, f);
        if (len > 0)
            request->size += len;
        }
        status = pclose(f);
        if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
        request->httpstatus = 200;
        else
        request->httpstatus = 404;
        request->data[request->size] = '\0';
    } else {
        ULOG_DEBUG
        ("Error: Could not open pipe \"%s\"", &request->source[1]);
    }
/*
    } else if  ( (NULL != strstr(request->source, "http://")) || 
                (NULL != strstr(request->source, "https://"))  ) 
*/
    //allow HtTp://
     } else if  ( (strncasecmp(request->source, "http://", 7) >= 0 ) || 
                ( strncasecmp(request->source, "https://", 8) >= 0)  ) 
   {

        /*tvh: IMPORTANT here we need to process local files as well 
         Added more compareing to process http:// or https:// only 
            should not invoke netio stuff when processing file:///....
        */
        //ULOG_DEBUG("download_process: URL found, starting to process %s",request->source);
        downloadlib_process_url(request);
        //ULOG_DEBUG("download_process: Received http status %d",request->httpstatus);
        //ULOG_DEBUG("download_process: After process_url, request->data: \n%s",request->data);
        //ULOG_DEBUG("download_process: for %s",request->source);
        
        if (request->httpstatus >= 400) {
            //ULOG_DEBUG("download_process: Received http status >= 400");
        /*    
        if(request->data != NULL)
            ULOG_DEBUG("download_process: %s",request->data);
        */
        g_free(request->data);
        request->data = NULL;
        request->size = 0;
    }
    } else if  (NULL != strstr(request->source, "file://")) {
        /* g_file is not using prefix 'file://'*/ 
        gchar *file_source = NULL;

        file_source = g_strdup (request->source + 7);
        //ULOG_DEBUG("download_process: processing FILES now for \nfile = %s\n",file_source);

        if (g_file_test(file_source, G_FILE_TEST_EXISTS)) {
            /* we have a file... */
           if ((!g_file_get_contents
                    (file_source, &(request->data), &(request->size),
                     NULL)) || (request->data[0] == '\0')) {
               request->httpstatus = 403;
               ULOG_DEBUG
                ("Error: Could not open file \"%s\"\n", request->source);
           } else {
               //g_assert(NULL != request->data);
               g_return_if_fail(NULL != request->data);
               request->httpstatus = 200;
           }
           /* free this potential mem leak. */
           g_free(file_source);

        } else {
           ULOG_DEBUG("Error: There is no such file \"%s\"n",request->source);
           request->httpstatus = 404;
        }

    } else {
         ULOG_DEBUG("Wrong source");
    }
    
    /* And execute the postfilter */
    if (request->data != NULL && request->filtercmd != NULL) {
    size_t size;
    gchar *tmp = filter(request->filtercmd, request->data, &size);
    if (tmp != NULL) {
        g_free(request->data);
        request->data = tmp;
        request->size = size;
    }
    }

}

gpointer download_request_new()
{
    struct request *request;

    debug_enter("update_request_new");

    request = g_new0(struct request, 1);

    debug_exit("update_request_new");

    return (gpointer) request;
}

/**
  * @param: processed indicates whether or not the request->callback has been 
  * executed or not. It's ONLY true in ONE case when process_downloaded_idle_cb()
  * is called 
  * This is needed as some requests might need the callback to be called,
  * even though the requests haven't been downloaded, or have been cancelled midway though
  * the download. 
  * At least, currently, gtkhtml stream needs to close the stream for image requests!!
*/
void download_request_free(struct request *request, gboolean processed)
{

    ULOG_DEBUG("\n\n**********************download_request_free**********************\n");
    if (request == NULL)
        return;

    if (NULL != request) {
        //The gtkhtml lib needs to close the stream properly first.
        // only if it's not yet processed
        if ( (request->type_image == TRUE) && (!processed))
            request->callback (request);
       
        if (NULL != (gchar * )request->source)
        {
           //ULOG_DEBUG("Freeing request: %s", (gchar *) request->source);
           g_free(request->source);
           request->source = NULL;
        }
        if (request->filtercmd)
        {
            g_free(request->filtercmd);
            request->filtercmd = NULL;
        }
        if (request->etag)
        {
            g_free(request->etag);
            request->etag = NULL;
        }
        if (request->data)
        {
          g_free(request->data);
            request->data = NULL;
        }
        /*
        if ((request->type_image == TRUE) && (request->user_data != NULL))
        {
            gtk_html_stream_close((GtkHTMLStream *)request->user_data, GTK_HTML_STREAM_OK);
        }
        */
        /*FIXME: What about the other fields?*/

        /* IMPORTANT:Might be dangersous here to free some of this
        because we will destroy the feeds, the folder or even the ui
        if freeing them...Needs coming back
        */
    
    if (request->user_data)
       request->user_data = NULL;
    /*
    
    if (request->lastmodified)
       g_free(request->lastmodified);
    if (request->folder)
       g_free(request->folder);
    */
    // and request->callback, request->feed and request->folder */

    g_free(request);
    request = NULL;
    }
    //ULOG_DEBUG("--------------------------------------------");
}


void download_init(void)
{
    gint i = 0;
    //int count = 0;

    downloadlib_init();

    /* Use GAsync as they are gthread safe */
    requests_high_prio = g_async_queue_new();
    requests_normal_prio = g_async_queue_new();
    requests_image = g_async_queue_new();
    results = g_async_queue_new();

    /* Let the threads run */
    program_quitting = FALSE;

    /* counters */
    app_data->app_ui_data->rss_current_download = 0;
    app_data->app_ui_data->rss_max_download = 0;
    app_data->app_ui_data->images_current_download = 0;
    app_data->app_ui_data->images_max_download = 0;
    num_pending_requests = 0;
    num_downloaded_requests = 0;
    
    /* Threads should wait till connected */
    //online = FALSE;
    //offline_cond = g_cond_new();
    //cond_mutex = g_mutex_new();

    download_cond = g_cond_new();
    download_cond_mutex = g_mutex_new();

    results_queue_cond = g_cond_new();
    results_queue_cond_mutex = g_mutex_new();

    /* Counter mutex */
    //counter_mutex = g_mutex_new();
    /* Pending requests mutex to control accessing num_pending_request
    */
    pending_check_mutex = g_mutex_new(); 
    num_downloaded_check_mutex = g_mutex_new(); 
    results_queue_mutex = g_mutex_new(); 
    image_list_mutex = g_mutex_new();  
    /*if (1 >=
    (rss_threads = getNumericConfValue(UPDATE_THREAD_CONCURRENCY)))*/
    //rss_threads = DEFAULT_UPDATE_THREAD_CONCURRENCY;

    for (i = 0; i < DEFAULT_UPDATE_THREAD_CONCURRENCY; i++)
    {
           ULOG_DEBUG("download_init: creating thread nubmer %d", i);
           update_threads[i] = g_thread_create(download_thread, GINT_TO_POINTER (i), TRUE,
                  NULL);
    }
    /* One more thread to process the results queue, concurrently */

    ULOG_DEBUG("download_init: creating the update ui thread with nubmer %d",
         DEFAULT_UPDATE_THREAD_CONCURRENCY);
    update_threads[DEFAULT_UPDATE_THREAD_CONCURRENCY] = 
        g_thread_create(results_thread, GINT_TO_POINTER (DEFAULT_UPDATE_THREAD_CONCURRENCY), TRUE, NULL);
                                
    return;
}

/* De-init all the stuff init'ed
*/
void download_deinit(void)
{
    ULOG_DEBUG("....download_deinit....");
    gint i;
    program_quitting = TRUE;
    //Broadcast for other threads to quit
    ULOG_DEBUG("....download_deinit....1");
    g_mutex_lock(download_cond_mutex);
        ULOG_DEBUG("....download_deinit....2");
    g_cond_broadcast(download_cond);
        ULOG_DEBUG("....download_deinit....3");
    g_mutex_unlock(download_cond_mutex);
        ULOG_DEBUG("....download_deinit....4");
    
    for (i = 0; i < DEFAULT_UPDATE_THREAD_CONCURRENCY ; i++)
    {
        ULOG_DEBUG("joing thread number %d", i);
        //another one coz some threads might have been busy?
        g_mutex_lock(download_cond_mutex);
        g_cond_broadcast(download_cond);
        g_mutex_unlock(download_cond_mutex);

        g_thread_join(update_threads[i]);
    
    }

    //so that results thread can exit
    // results thread
    // unlock this so that results thread can proceed
    g_mutex_unlock(num_downloaded_check_mutex);
    
    /*
    g_mutex_lock(results_queue_cond_mutex);    
        ULOG_DEBUG("....download_deinit....5");
    g_cond_signal(results_queue_cond);
        ULOG_DEBUG("....download_deinit....6");
    g_mutex_unlock(results_queue_cond_mutex);    
        ULOG_DEBUG("....download_deinit....7");

    g_thread_join(update_threads[DEFAULT_UPDATE_THREAD_CONCURRENCY]);
    */
    /*this is dirty hack around the quitting. When download is stopped
    * but somehow results thread is still waiting on the g_cond. Perhaps download
    threads haven't signalled properly when exiting? or somthing else
    This way it will always quit :D
    TODO: Fix it properly*/
    while (!results_thread_done)
    {
        g_mutex_lock(results_queue_cond_mutex);    
        ULOG_DEBUG("....download_deinit....5");
        g_cond_signal(results_queue_cond);
        ULOG_DEBUG("....download_deinit....6");
        g_mutex_unlock(results_queue_cond_mutex);    

    }    
    /* free the queues */
    ULOG_DEBUG("....download_deinit....8");
    //sleep(1);
    g_async_queue_unref(requests_high_prio);
    requests_high_prio = NULL;
    g_async_queue_unref(requests_normal_prio); 
    requests_high_prio = NULL;
    g_async_queue_unref(requests_image); 
    requests_image = NULL;
    g_async_queue_unref(results);
    results = NULL;
    ULOG_DEBUG("....download_deinit....9 num_pending_requests = %d", num_pending_requests);
    /* Offline cond */
    //g_cond_free(offline_cond); g_mutex_free(cond_mutex);
    /* There is request cond */ 
    
    g_cond_free(download_cond);
    g_mutex_free(download_cond_mutex);
    ULOG_DEBUG("....download_deinit....10");

    g_cond_free(results_queue_cond); g_mutex_free(results_queue_cond_mutex);
    ULOG_DEBUG("....download_deinit....11");
    //sleep (2);
    g_mutex_free(pending_check_mutex);
    ULOG_DEBUG("....download_deinit....12");    
    g_mutex_free(num_downloaded_check_mutex);
    ULOG_DEBUG("....download_deinit....13");
        
    g_mutex_free(results_queue_mutex);
    ULOG_DEBUG("....download_deinit....14");    
    g_mutex_free(image_list_mutex);
    ULOG_DEBUG("....download_deinit....DONE");    
    //g_mutex_free(counter_mutex);
    //download_request_free(fakerequest);
}
/*We have 2 download queue modes: downloading feeds, or downloading images
  * for gtkhtml to display
  * rss_max_download is uncertain so it can be increased ++ here
  * but images_max_download is set in gtkhtml.c so it shouldn't be bothered here

*/

void download_queue(struct request *new_request)
{
    //g_assert(NULL != new_request);
    g_return_if_fail(NULL != new_request);
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    g_return_if_fail(app_data != NULL);
    g_return_if_fail(app_data->app_ui_data != NULL);
    
    /* no need to get connected here
        all the requests must have required connections during  
        the refresh. If connection is lost during downloading, just
        ignore it
    */
    ULOG_DEBUG("\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^NEW REQUEST ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
    if (!app_data->app_ui_data->connected)
    {
        ULOG_DEBUG("download_queue: freeeing request due to network not yet connected");
        download_request_free(new_request, FALSE);
        return;
    }

   ULOG_DEBUG("download_queue: request->source = %s", new_request->source);
    //Now we will definitely push so let's SHOW progress bar first
    // jUST show IT
    // if feed_directory_loading... No need to. It's handled in feed_directory.c already
    if (!new_request->type_image)// && !app_data->app_ui_data->feed_directory_loading)
    {
        ULOG_DEBUG("download_queue: updateing progress bar : REFRESH");
        show_refresh_progressbar_button();
    }
    else
    {
        ULOG_DEBUG("download_queue: updateing progress bar : IMAGE");
        show_receiving_images_progressbar_button();
    }

    // Push request to appropriate queue
    if (new_request->priority == 1)
        g_async_queue_push(requests_high_prio, new_request);
    else if (!new_request->type_image)
        g_async_queue_push(requests_normal_prio, new_request);
    else
    {
                g_mutex_lock(image_list_mutex);
        g_async_queue_push(requests_image, new_request);
            g_mutex_unlock(image_list_mutex);
    }
    
    if (new_request->type_image)
        //TODO: moved to inside gtkhtml. so we won't have the progressbar : "Receiving 0/0"?
        add_counter(IMG_MAX,1);
    else
        add_counter(RSS_MAX,1);
    /*
    ULOG_DEBUG("download_queue: num_pending_requests =%d", num_pending_requests);
    ULOG_DEBUG("download_queue: num_downloaded_requests NOT YET processed =%d", num_downloaded_requests);
    ULOG_DEBUG("download_queue: images_max =%d", app_data->app_ui_data->images_max_download);
    ULOG_DEBUG("download_queue: rss_max =%d", app_data->app_ui_data->rss_max_download);
    */

    // increase the counters FIRST, before signalling the cond
    add_counter(NUM_PENDING, 1);

    // wake up ONE thread - there is something to download

    g_mutex_lock(download_cond_mutex);
    g_cond_signal(download_cond);
    g_mutex_unlock(download_cond_mutex);

    ULOG_DEBUG("download_queue: request pushed: request->source = %s", new_request->source);
    ULOG_DEBUG("\n^^^^^^^^^^^^^^ REQUEST PUSHED^^^^^^^^^^^^^^\n");
    app_data->app_ui_data->download_stopped = FALSE;
    return;
}

void download_set_online(gboolean mode)
{
    //g_assert(app_data->app_ui_data != NULL);
    g_return_if_fail(app_data->app_ui_data != NULL);
    app_data->app_ui_data->connected = mode;
}

/* Empties a queue.
 * @param: queue the queue to free
   @results_queue : Whether it is results queue
 * When , later on, we want to cancel image requests for example, we need to
 * distinguish this, so that we can choose to decrease num_pending_requests or not
*/
void download_empty_queue(GAsyncQueue * queue, gboolean results_queue)
{
    gpointer data = NULL;
    struct request *req = NULL;

    //g_assert(queue != NULL);
    if (queue == NULL)
        return;
        
    ULOG_DEBUG("download_empty_queue");
    while (TRUE) {
       data = g_async_queue_try_pop(queue);
       /* Stop the function when the queue is empty. */
       if (data == NULL) {
           break;
       }
       
       req = (struct request *) data;
       if (req->feed != NULL) {
           req->feed->request = NULL;
       }
        /*
       // image stream will be freed by gtkhtml model
       if (req->type_image == TRUE)
            req->callback (req);
        */
       download_request_free(req, FALSE); /* not yet processed*/
       /* increase counters properly */
       if (results_queue)
       {
            if (req->type_image)
                add_counter(IMG_CURRENT, 1); //regard as 1 more image downloaded & processed
            else
                add_counter(RSS_CURRENT, 1);
       }
       else //it's the pending queues, be it feed or image queue
       {
            add_counter(NUM_PENDING, -1);
       }

       req = NULL;
    }
}

static void cancel_active_threads(void)
{
    //app_data->app_ui_data->download_cancelled = TRUE;
    
    gint i;
    //TODO: accessing the array here might be dangerous as the download threads
    // are accessing them...
    for (i = 0; i < DEFAULT_UPDATE_THREAD_CONCURRENCY ;i ++)
    {
        if (thread_is_active[i])
            active_thread_cancelled[i] = TRUE;    
    }

    return;
    
}
void download_cancel_all(void)
{
    ULOG_DEBUG("\n\n\n\n_____________________download_cancel_all_________________________\n\n\n\n");
    //stop downloading first
    downloadlib_cancel_all();
     
    //now empty the queues. Stop download threads from getting any more pending requests first
    ULOG_DEBUG("locking pending_check_mutex");
    g_mutex_lock(pending_check_mutex);
    ULOG_DEBUG("LOCKED pending_check_mutex. Now empty download queues");
    download_empty_queue(requests_high_prio, FALSE);
    download_empty_queue(requests_normal_prio, FALSE);
    download_empty_queue(requests_image, FALSE);
    
    //lock 2 things: 1. results thread from further processing
    //               2. download threads from pushing to results queue
    g_mutex_lock(results_queue_mutex);
    // why this? Let the thread finish updating result
    download_empty_queue(results, TRUE);
    cancel_active_threads();

    //unlock counters
    g_mutex_unlock(pending_check_mutex);   
    g_mutex_unlock(results_queue_mutex);
    
    on_download_finished();

    /* other chores */
    //after this, queues emptied, few active requests are maybe still hanging around
    //but we can always safely reset the figures here
    /* requests counters??? */
    num_pending_requests = 0;
    num_downloaded_requests = 0;
    app_data->app_ui_data->rss_max_download = 0;
    app_data->app_ui_data->rss_current_download = 0;
    app_data->app_ui_data->images_max_download = 0;
    app_data->app_ui_data->images_current_download = 0;
    return;
}


/** Handle the download cancelling when user presses stop button 
 * while it's refreshing or receiving images
*/
/* tvh: 2006May10:Just copied from on_searchentry_activate for now
TODO: Improvement needed: 
For example, if both feeds and images are being downloaded. Pressing once should
    stop the feeds ,but continue with the images
    Press another time will stop the images too
*/
void handle_download_stopped_by_user(void)
{
    ULOG_DEBUG("#####################USER PRESSED STOP DOWNLOADING######################");
    app_data->app_ui_data->search_mode = SFM_INTERRUPTED;

    //flush the ui events first
    //while (gtk_events_pending ())
    //    gtk_main_iteration ();
    //if  (!app_data->app_ui_data->start_on_background)
    if (get_counter(RSS_MAX) > 0 )    
        gtk_infoprint(GTK_WINDOW(app_data->app_ui_data->app),
                         _("rss_ib_refresh_interrupted"));
  
    if (get_counter(RSS_MAX) > 0 )
    {
        download_cancel_all();
        //ULOG_DEBUG("main thread is signalling applet now");
        on_refresh_finished(FALSE, FALSE); //banner & signal applet
    }
    
    // this will be done by _idle calls??
    else
        download_cancel_all();

    app_data->app_ui_data->download_stopped = TRUE;
}

void download_cancel_image_requests_when_feeds_swiched(void)
{
    //can have a MUTEX here i think.....
    if (app_data->app_ui_data->progressbar_mode == PROGRESSBAR_IMAGE_MODE)
    {
        //definitely we're only downloading images
        //ULOG_DEBUG("####switching page while still downloading images ###########");
        handle_download_stopped_by_user();
    }
    else
    {
        // image requests are behind the feeds requests
        // then block the threads from accessing this queue first
        //ULOG_DEBUG("####switching page while still downloading BOTH FEEDS and IMAGES ###########");
        g_mutex_lock(image_list_mutex);
        struct request * request;
        // Now remove the list or mark them as invalid.
        while (g_async_queue_length (requests_image) > 0)
        {
            request = g_async_queue_pop (requests_image);
            add_counter(NUM_PENDING, -1);
            add_counter(IMG_MAX, -1);
            download_request_free(request, FALSE);
            request = NULL;
        }
        // Then unlock it. There shouldn't be any image requests in the results queue
        //yet, as the threads have been trying to download only feeds. That's why we're in
        // this 'else' case. Otherwise we could have been in 'if' case
        g_mutex_unlock(image_list_mutex);

    }
    //download_empty_queue(requests_image, FALSE);
}

gboolean is_image_queue_empty(void)
{
    if (requests_image == NULL)
        return TRUE;
    return (g_async_queue_length (requests_image) > 0) ? FALSE:TRUE;
}
void download_remove_dequeue_source()
{
    //g_assert(app_data != NULL);
    //g_assert(app_data->app_ui_data != NULL);
    g_return_if_fail(app_data!=NULL);
    g_return_if_fail(app_data->app_ui_data != NULL);
    
    if(app_data->app_ui_data->download_dequeue_source != 0 &&
       app_data->app_ui_data->search_mode != SFM_DOWNLOAD)
    {
        g_source_remove(app_data->app_ui_data->download_dequeue_source);
    app_data->app_ui_data->download_dequeue_source = 0;
    }
}
